package com.qcg.comp.sdk;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Service;
import javax.xml.ws.WebServiceFeature;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.MTOMFeature;
import javax.xml.ws.soap.SOAPBinding;

import com.qcg.comp.sdk.exceptions.DeleteOnTerminationException;
import com.qcg.comp.sdk.exceptions.FileNotFoundException;
import com.qcg.comp.sdk.exceptions.NotAuthorizedException;
import com.qcg.comp.sdk.exceptions.StagingException;
import com.qcg.comp.sdk.jsdl.JSDL;
import com.qcg.comp.sdk.types.FileInfo;
import com.qcg.comp.stubs.jsdl.DataStagingType;
import com.qcg.comp.stubs.jsdl.JobDefinitionType;
import com.qcg.comp.stubs.staging.ChModType;
import com.qcg.comp.stubs.staging.DeleteOnTerminationFault;
import com.qcg.comp.stubs.staging.FileNotFoundFault;
import com.qcg.comp.stubs.staging.MkDirType;
import com.qcg.comp.stubs.staging.NotAuthorizedFault;
import com.qcg.comp.stubs.staging.QCGComputingStaging;
import com.qcg.comp.stubs.staging.QCGComputingStagingPortType;
import com.qcg.comp.stubs.staging.StageFilesInType;
import com.qcg.comp.stubs.staging.StageFilesOutType;
import com.qcg.comp.stubs.staging.StagingFault;
import com.qcg.core.sdk.QCGCoreService;
import com.qcg.core.sdk.attachments.AttachmentHandler;
import com.qcg.core.sdk.attachments.FileStagingHandler;
import com.qcg.core.sdk.security.AuthenticationModule;
import com.sun.xml.internal.ws.developer.JAXWSProperties;
import com.sun.xml.ws.developer.StreamingAttachmentFeature;

public class QCGStaging extends QCGCoreService {

	public static final String QCG_STAGING_NS = "http://schemas.qoscosgrid.org/comp/2011/04/staging";

	protected QCGComputingStagingPortType staging;
	protected FileStagingHandlerWrapper stagingHandler;
	protected volatile IOException stagingException;
	protected AttachmentHandler attachmentHandler;

	protected class FileStagingHandlerWrapper implements FileStagingHandler {
		protected FileStagingHandler userFileHandler;
		protected Map<String, String> uuidMap = new HashMap<String, String>();
		
		public FileStagingHandlerWrapper(FileStagingHandler userFileHandler) {
			this.userFileHandler = userFileHandler;
		}

		public InputStream stageIn(String sourceFileName) throws IOException {

			return userFileHandler.stageIn(sourceFileName);
		}

		public void stageOut(String targetFileName, InputStream fileStream)
				throws IOException {
			synchronized (uuidMap) {
				userFileHandler.stageOut(uuidMap.remove(targetFileName), fileStream);
			}
		}
		
		public void addAssociation(String uuid, String targetFileName) {
			synchronized (uuidMap) {
				uuidMap.put(uuid, targetFileName);
			}
		}

	}

	protected class BinaryDataSource implements DataSource {

		protected String name;
		protected InputStream inputStream;

		public BinaryDataSource(String name, InputStream inputStream) {
			super();
			this.name = name;
			this.inputStream = inputStream;
		}

		public String getContentType() {
			return "application/octet-stream";
		}

		public InputStream getInputStream() throws IOException {
			return inputStream;
		}

		public String getName() {
			return name;
		}

		public OutputStream getOutputStream() throws IOException {
			return null;
		}

	}

	public QCGStaging(String hostname, int port, AuthenticationModule tm, AuthenticationModule mm,
			boolean useWSA, FileStagingHandler userStagingHandler)
			throws JAXBException {
		super(hostname, port, tm, mm, useWSA);

		this.stagingHandler = new FileStagingHandlerWrapper(userStagingHandler);
		this.staging = (QCGComputingStagingPortType) stubs;
	}

	
	@Override
	@SuppressWarnings("all")
	protected List<Handler> getHandlersChain() {
		attachmentHandler = new AttachmentHandler(
				new AttachmentHandler.IOExceptionListener() {

					public void exceptionThrown(IOException ex) {
						stagingException = ex;
					}
				});

		List<Handler> handlers = new ArrayList<Handler>();
		handlers.add(attachmentHandler);

		return handlers;
	}

	@Override
	protected Object getPort(Service service, WebServiceFeature... features) {
		return ((QCGComputingStaging) service).getQCGComputingStaging(features);
	}

	@Override
	protected Service getService() {
		return new QCGComputingStaging(QCGStaging.class
				.getResource("/schemas/qcg-comp-staging.wsdl"), new QName(
				QCG_STAGING_NS, "QCGComputingStaging"));
	}
	
   protected void checkForIOException() throws IOException{
   	 if (stagingException != null) {
   		 IOException ioEx = stagingException;
   		 stagingException = null;
   		 throw ioEx;
   	 }
   }

	public synchronized void stageInFiles(JSDL jsdl) throws IOException, FileNotFoundException, NotAuthorizedException, StagingException {
		JobDefinitionType jobDefinition = jsdl.toJobDefinition();

		Map<String, DataHandler> attachments = new HashMap<String, DataHandler>();

		for (DataStagingType ds : jobDefinition.getJobDescription().getDataStaging()) {
			if (ds.getSource() == null || ds.getSource().getURI() == null) {
				continue;
			}
			String sourceURI = ds.getSource().getURI();
			String sourceUUID = "cid:" + UUID.randomUUID().toString();
			ds.getSource().setURI(sourceUUID);

			attachments.put(sourceUUID, new DataHandler(new BinaryDataSource(sourceUUID, stagingHandler.stageIn(sourceURI))));
		}

		((BindingProvider) staging).getRequestContext().put(MessageContext.OUTBOUND_MESSAGE_ATTACHMENTS, attachments);

		StageFilesInType parameters = new StageFilesInType();
		parameters.setJobDefinition(jobDefinition);
		try {
			staging.stageFilesIn(parameters);
			checkForIOException() ;
		} catch (FileNotFoundFault e) {
			throw new FileNotFoundException(e);
		} catch (NotAuthorizedFault e) {
			throw new NotAuthorizedException(e);
		} catch (StagingFault e) {
			throw new StagingException(e);
		} finally {
			((BindingProvider) staging).getRequestContext().put(MessageContext.OUTBOUND_MESSAGE_ATTACHMENTS, null);
		}
	}

	public synchronized void stageOutFiles(JSDL jsdl) throws IOException, DeleteOnTerminationException, FileNotFoundException, NotAuthorizedException, StagingException {
		JobDefinitionType jobDefinition = jsdl.toJobDefinition();
		
		for (DataStagingType ds : jobDefinition.getJobDescription().getDataStaging()) {
			if (ds.getTarget() == null || ds.getTarget().getURI() == null) {
				continue;
			}
			String targetURI = ds.getTarget().getURI();
			String targetUUID = "cid:" + UUID.randomUUID().toString();
			ds.getTarget().setURI(targetUUID);
			stagingHandler.addAssociation(targetUUID, targetURI);
		}

		StageFilesOutType parameters = new StageFilesOutType();
		parameters.setJobDefinition(jobDefinition);

		this.attachmentHandler.setFileStagingHandler(this.stagingHandler);

		try {
			((BindingProvider)staging).getRequestContext().put(JAXWSProperties.HTTP_CLIENT_STREAMING_CHUNK_SIZE, 1024);

			staging.stageFilesOut(parameters);
		} catch (DeleteOnTerminationFault e) {
			throw new DeleteOnTerminationException(e);
		} catch (FileNotFoundFault e) {
			throw new FileNotFoundException(e);
		} catch (NotAuthorizedFault e) {
			throw new NotAuthorizedException(e);
		} catch (StagingFault e) {
			throw new StagingException(e);
		}
		checkForIOException(); 
	}

	public synchronized JSDL listDirectory(String directory, String userName) throws FileNotFoundFault, NotAuthorizedFault {	    
	    return new JSDL(this.staging.listDirectory(directory, userName));
	}	
	
	public synchronized FileInfo[] statFile(List<String> paths, String userName) throws FileNotFoundFault, NotAuthorizedFault, StagingFault {
		List<com.qcg.comp.stubs.staging.FileInfo> s = this.staging.statFile(paths, userName);
		FileInfo[] result = new FileInfo[s.size()];
		for (int i = 0; i < s.size(); i++) {
			com.qcg.comp.stubs.staging.FileInfo l = s.get(i);

			result[i] = new FileInfo(l.getMode(), l.getNLink(), l.getUID(),
					l.getGID(), l.getSize(), l.getATime().toGregorianCalendar()
					.getTime(), l.getCTime().toGregorianCalendar()
					.getTime(), l.getMTime().toGregorianCalendar()
					.getTime());
		}
		return result;    
	}

	public QCGStaging(String hostname, int port,
			AuthenticationModule transportLayer,
			AuthenticationModule messageLayer, FileStagingHandler userStagingHandler) throws JAXBException {
		super(hostname, port, transportLayer, messageLayer);
		
		this.stagingHandler = new FileStagingHandlerWrapper(userStagingHandler);
		this.staging = (QCGComputingStagingPortType) stubs;
	}

	public QCGStaging(String hostname, int port, AuthenticationModule am, boolean useWSA,  FileStagingHandler userStagingHandler) throws JAXBException {
		super(hostname, port, am, useWSA);
		
		this.stagingHandler = new FileStagingHandlerWrapper(userStagingHandler);
		this.staging = (QCGComputingStagingPortType) stubs;
	}

	public QCGStaging(String hostname, int port, AuthenticationModule am,  FileStagingHandler userStagingHandler)
			throws JAXBException {
		super(hostname, port, am);
		
		this.stagingHandler = new FileStagingHandlerWrapper(userStagingHandler);
		this.staging = (QCGComputingStagingPortType) stubs;
	}

	public synchronized void delete(String file, String userName) throws FileNotFoundFault, NotAuthorizedFault, StagingFault {
	    this.staging.delete(file, userName);
	    
	}

    public synchronized void chmod(String file, String mode, String userName) throws FileNotFoundFault, NotAuthorizedFault, StagingFault {
	ChModType parameters = new ChModType();

	parameters.setPath(file);
	parameters.setMode(mode);
	parameters.setUserName(userName);

	this.staging.chMod(parameters);
    }

    public synchronized void mkdir(String dir, String mode, String userName) throws FileNotFoundFault, NotAuthorizedFault, StagingFault {
	MkDirType parameters = new MkDirType();

	parameters.setPath(dir);
	parameters.setMode(mode);
	parameters.setUserName(userName);

	this.staging.mkDir(parameters);
    }
}
