package pl.psnc.qcg.rm;



import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import javax.xml.bind.JAXBException;

import com.qcg.comp.sdk.types.ReservationDocument;
import com.qcg.comp.sdk.types.ReservationEndpointReference;
import com.qcg.comp.stubs.ares.ARESReservationStatusType;
import com.qcg.comp.stubs.ares.ARESReservedHostType;


public class QCGReservationMonitor extends Thread{

	private static Logger log = Logger.getLogger("RESMON");
	
	static {
		try {
			//log.addHandler(new ConsoleHandler());
			log.setLevel(Level.ALL);
			log.getParent().getHandlers()[0].setFormatter(new Formatter() {

				@Override
				public String format(LogRecord record) {
					String tname = Thread.currentThread().getName();
					return "[" + String.format("%8s", tname) + "] - " + new Date(record.getMillis()) + " : " + record.getMessage() + "\n";
				}
			});
			
			
		} catch (SecurityException e) {
			e.printStackTrace();
		}
	}
	public static final String QCG_PA_CLUSTERS = "qcg.clusters";
	public static final String QCG_POLLING_INTERVAL = "qcg.polling.interval";
	public static final String QCG_CONTACT_STRING = "qcg.contact.string";
	
	public static final String INSERT_RES_INFO_SQL = "INSERT INTO reservations VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?);";
	public static final String INSERT_HOST_INFO_SQL = "INSERT INTO hosts VALUES (?, ?, ?, ?, ?);";
	public static final String DELETE_RES_INFO_SQL = "DELETE FROM reservations WHERE uuid = ?";
	public static final String UPDATE_RES_INFO_SQL = "UPDATE reservations SET start_time = ?, end_time = ? WHERE uuid = ?";
	public static final String TEST_SQL = "SELECT * FROM reservations WHERE uuid = 'TEST'";


	public static final int DEFAULT_POLLING_INTERVAL = 50 *1000;
	
	protected static int pollingInterval = DEFAULT_POLLING_INTERVAL;
	
	protected Cluster cluster;
	protected String contactString = System.getProperty(QCG_CONTACT_STRING, "jdbc:mysql://node1.qoscosgrid.man.poznan.pl:3306/qoscosdb?user=qoscosportal&password=test&autoReconnect=true&jdbcCompliantTruncation=false");
	protected Set<ReservationEndpointReference> activeReservations =  new HashSet<ReservationEndpointReference>();
	protected Set<ReservationEndpointReference> currentReservations =  new HashSet<ReservationEndpointReference>();
	protected Map<ReservationEndpointReference, ResInfo> resTimeWindows = new HashMap<ReservationEndpointReference, ResInfo>();

	protected Connection db;
	protected PreparedStatement resAddStmt;
	protected PreparedStatement resDelStmt;
	protected PreparedStatement resUpdateStmt;
	protected PreparedStatement hostStmt;
	protected PreparedStatement testStmt;
	
	protected class ResInfo {
		protected Date startTime;
		protected Date endTime;
		
		public ResInfo(Date startTime, Date endTime) {
			super();
			this.endTime = endTime;
			this.startTime = startTime;
		}
		
		public Date getStartTime() {
			return startTime;
		}
		public Date getEndTime() {
			return endTime;
		}
		
	}
	/**
	 * @param args
	 * @throws JAXBException 
	 */
	public static void main(String[] args) throws Exception {
		
		Class.forName("com.mysql.jdbc.Driver").newInstance();
			
		pollingInterval = Integer.parseInt(System.getProperty(QCG_POLLING_INTERVAL, DEFAULT_POLLING_INTERVAL + ""));
		
		Grid grid = new Grid();
		
		List<QCGReservationMonitor> monitorsList = new ArrayList<QCGReservationMonitor>();
		
		for (Cluster c : grid.getClusters()) {
			monitorsList.add(new QCGReservationMonitor(c));
		}
		
		for (QCGReservationMonitor m : monitorsList) {
			m.start();
		}
		
		for (QCGReservationMonitor m : monitorsList) {
			m.join();
		}
	}
	
	public QCGReservationMonitor(Cluster cluster) throws JAXBException {
		this.cluster = cluster; 
		setName(cluster.getName());
	}

	protected void createConnection() throws SQLException {
		db = DriverManager.getConnection(contactString);
		db.setAutoCommit(false);
		
		resAddStmt = db.prepareStatement(INSERT_RES_INFO_SQL);
		resDelStmt = db.prepareStatement(DELETE_RES_INFO_SQL);
		resUpdateStmt = db.prepareStatement(UPDATE_RES_INFO_SQL);
		hostStmt = db.prepareStatement(INSERT_HOST_INFO_SQL);
		testStmt = db.prepareStatement(TEST_SQL);
	}
	
	protected boolean checkConnection() {
		try {
			testStmt.execute();
			return true;
		} catch (SQLException ex) {
			log.severe("Test query failed: " + ex);
		}
		
		log.severe("Reconnecting..");
		
		try {
			resAddStmt.close();
			resDelStmt.close();
			resUpdateStmt.close();
			testStmt.close();
			db.close();
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
		try {
			createConnection();
			return true;
		} catch (SQLException ex) {
			log.severe("Unable to create connection : " + ex);
		}
		
		return false;
	}
	
	public void run() {
		log.info("contact string = " + contactString);
		log.info("polling interval = " + pollingInterval);
		
		try {
			createConnection();
		} catch (SQLException ex) {
			ex.printStackTrace();
			return;
		}
		
		while (true) {
			currentReservations.clear();
			
				List<ReservationEndpointReference> reservations = null;
				
				try {
					log.info("reservation query");
					reservations = cluster.getAres().getActiveReservations();
				} catch (Exception ex) {
					log.info("Unable to get reservation: " + ex);
					sleep();
					continue;
				}
				
				currentReservations.addAll(reservations);
				
				for (ReservationEndpointReference rEpr : reservations) {
					if (!activeReservations.contains(rEpr)) {
						addNewReservation(rEpr);
					}
				}
				
			
			Set<ReservationEndpointReference> resToRemove = new HashSet<ReservationEndpointReference>();
			for (ReservationEndpointReference rEpr : activeReservations) {
				if (!currentReservations.contains(rEpr)) {
					
					log.info("reservation terminated: " + rEpr.getReservationUUID());
					resToRemove.add(rEpr);
				}
			}
			
			activeReservations.removeAll(resToRemove);

			Date currentTime = new Date();

			resToRemove.clear();
			//Delete/update reservation info
			for (ReservationEndpointReference rEpr : resTimeWindows.keySet()) {
				ResInfo resInfo = resTimeWindows.get(rEpr);
				
				if (resInfo.getEndTime().before(currentTime)) {
					resToRemove.add(rEpr);
					log.info("left unchanged (" + resInfo.getEndTime() + "<" + currentTime + "): " + rEpr.getReservationUUID());
					continue;
				}
				
				if (!activeReservations.contains(rEpr)) {
					resToRemove.add(rEpr);
					
					if (resInfo.getStartTime().after(currentTime)) {
						deleteReservation(rEpr);
					} else if (resInfo.getEndTime().after(currentTime)) {
						updateReservation(rEpr,resInfo.getStartTime(), currentTime);
					} else {
						log.info("left unchanged ???: " + rEpr.getReservationUUID());
					}
					
				}
			}
			
			for  (ReservationEndpointReference rEpr : resToRemove) {
				resTimeWindows.remove(rEpr);
			}

			sleep();
		}
		
	}
	
	private void sleep() {
		try {
			Thread.sleep(pollingInterval);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
	private void deleteReservation(ReservationEndpointReference epr) {
		log.info("deleting reservation :" + epr.getReservationUUID());
		
		if (!checkConnection())
			return;
		
		try {
			resDelStmt.setString(1, epr.getReservationUUID());
			resDelStmt.executeUpdate();
			db.commit();
		} catch (SQLException ex) {
			
			ex.printStackTrace();
			
			try {
				db.rollback();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}
	
	private void updateReservation(ReservationEndpointReference epr,Date startTime, Date endTime) {
		log.info("updating reservation EPR :" + epr.getReservationUUID() + " -> " + endTime);
		
		if (!checkConnection())
			return;
		
		try {
			resUpdateStmt.setTimestamp(1, new Timestamp(startTime.getTime()));
			resUpdateStmt.setTimestamp(2, new Timestamp(endTime.getTime()));
			resUpdateStmt.setString(3, epr.getReservationUUID());
			resUpdateStmt.executeUpdate();
			db.commit();
		} catch (SQLException ex) {
			
			ex.printStackTrace();
			
			try {
				db.rollback();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}

	}

	private void addNewReservation(ReservationEndpointReference epr) {
		ReservationDocument resDoc = null;
		ARESReservationStatusType resStat = null;
		
		log.info("adding new reservation:" + epr.getReservationUUID());
		
		if (!checkConnection())
			return;	

		try {
			resDoc = cluster.getAres().getReservationDocument(epr);
			resStat = cluster.getAres().getReservationStatus(epr);
		} catch (Exception ex) {
			ex.printStackTrace();
			log.info("unable to get reservation stat/doc: " + ex);
			return;
		}
		
		
		String uuid = epr.getReservationUUID();
		String drmsId = resDoc.getLocalReservationId();
		String name = resDoc.getArdl().getName();
		
		//rational: second step of QCG algorithm
		if (name != null) {
			name = name.split(":")[0];
		}
		
		String username = resDoc.getArdl().getUserName();
		Date startTime = resDoc.getArdl().getStartTime();
		Date endTime = resDoc.getArdl().getEndTime();
		String description =  resDoc.getArdl().getDescription();
		
		try {
			resAddStmt.setString(1, uuid);
			resAddStmt.setString(2, cluster.getName());
			resAddStmt.setString(3, drmsId);
			resAddStmt.setString(4, name);
			resAddStmt.setTimestamp(5, new Timestamp(startTime.getTime()));
			resAddStmt.setTimestamp(6, new Timestamp(endTime.getTime()));
			resAddStmt.setString(7, username);
			resAddStmt.setString(8, "user");
			resAddStmt.setString(9, description);
			resAddStmt.executeUpdate();
			

			for (ARESReservedHostType rh : resStat.getReservedHost()){
				hostStmt.setString(1, uuid);
				hostStmt.setString(2, rh.getHostName());
				hostStmt.setInt(3, rh.getReservedSlotsCount().intValue());
				hostStmt.setInt(4, rh.getTotalSlotsCount().intValue());
				hostStmt.setInt(5, rh.getUsedSlotsCount().intValue());
				hostStmt.executeUpdate();
			}
			
			db.commit();
			activeReservations.add(epr);
			resTimeWindows.put(epr, new ResInfo(resDoc.getArdl().getStartTime(), resDoc.getArdl().getEndTime()));

		} catch (SQLException ex) {
			if (ex instanceof com.mysql.jdbc.exceptions.MySQLIntegrityConstraintViolationException) {
				activeReservations.add(epr);
				resTimeWindows.put(epr, new ResInfo(resDoc.getArdl().getStartTime(), resDoc.getArdl().getEndTime()));
			}

			log.info("unable to insert new row: " + ex);
			try {
				db.rollback();
			} catch (SQLException e) {
				log.info("unable to rollback: " + e);
			}
		}
		
	}
	

}
