/* $Id: application_mapper.c 41 2009-12-22 11:21:20Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>

#include <qcg-core/sm.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>


#include <include/jsdl_filter.h>
#include <include/namespaces.h>
#include <xmlstructs/xmlstructs.h>

__SM_ID("$Id: application_mapper.c 41 2009-12-22 11:21:20Z mamonski $");

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);
static sm_value_t *handle_factory_attributes(sm_signal_t, sm_value_t *, const sm_value_t *);


static char applications_xml[65536] = "";
static SM_BOOL truncated = SM_FALSE;

SM_GENERAL_MODULE = {

	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "application_mapper",
		.descr = "JSDL Application -> HPCPA Executable mapper",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},
};

typedef struct {
	char *path;
	char *name;
	char *version;
} *application_mapping_t;

static sm_list_t map_list;
static sm_alloc_tracked_t handle;

static void append_str(const char *str)
{
	truncated = (strlcat(applications_xml, str, sizeof(applications_xml)) >= sizeof(applications_xml));
}

static int
setup(sm_module_t this)
{
	char path[1024];
	char name[1024];
	char version[1024];
	const char *application_map_file = NULL;
	FILE *f = NULL;
	int count = 0;
	char line[1024];
	application_mapping_t map = NULL;

	SM_DEBUG_ENTER();

	SM_TRY_NOT_OK(sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:ApplicationMapFile", &application_map_file));
	sm_debug("application_map_file = %s", application_map_file);

	SM_TRY_NOT(handle = sm_alloc_tracked_create());

	if ((f = fopen(application_map_file, "r")) == NULL) {
		sm_perror("fopen");
		sm_error("Failed to open ApplicationMapFile: %s", application_map_file);
		goto fault;
	}

	append_str("<smcf:Applications xmlns:smcf='" SMC_NAMESPACE_FACTORY "'>");
	while (fgets(line, sizeof(line), f)) {

		if (line[0] == '#')
			continue;

		if (sscanf(line, "%s %s %s", name, version, path) != 3)
			continue;

		SM_TRY_NOT(map = sm_alloc_tracked_malloc(handle, sizeof(*map)));
		SM_TRY_NOT(map->name = sm_alloc_tracked_strdup(handle, name));
		SM_TRY_NOT(map->version = sm_alloc_tracked_strdup(handle, version));
		SM_TRY_NOT(map->path = sm_alloc_tracked_strdup(handle, path));

		sm_debug("Application mapping <%s, %s> -> %s", map->name, map->version, map->path);

		append_str("<smcf:Application name='");
		append_str(map->name);
		append_str("' ");
		if (strcmp(map->version, "*") != 0) {
			append_str("version='");
			append_str(map->version);
			append_str("' ");
		}
		append_str("/>");

		SM_TRY_NOT(sm_list_add(&map_list, map, 0));
		map = NULL;

		count++;
	}
	append_str("</smcf:Applications>");

	if (truncated) {
		sm_error("To many application entries");
		goto fault;
	}

	sm_debug("application XML = %s", applications_xml);

	sm_debug("Read %d application mappings", count);
	fclose(f);

	SM_TRY_FAULT(sm_signal_connect("smc_jsdl_filter", handler));
	SM_TRY_FAULT(sm_signal_connect("smc_factory_attributes", handle_factory_attributes));


	return SM_OK;

fault:
	if (f)
		fclose(f);
	if (map) {
		SM_FREE(map->name);
		SM_FREE(map->path);
		SM_FREE(map->version);
		SM_FREE(map);
	}
	sm_alloc_tracked_free(&handle);
	sm_list_free(map_list, SM_FALSE);
	return SM_FAULT;
}

static void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();

	sm_alloc_tracked_free(&handle);
	sm_list_free(map_list, SM_FALSE);
}

static sm_value_t *
handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	sm_alloc_tracked_t mh;
	_jsdl__JobDefinition jsdl;
	const char *input_jsdl;
	char *output_jsdl;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER();

	if (sm_previous_signal_failed(chained_params))
		return sm_value_clone(chained_params);
	else if (sm_signal_first(chained_params))
		input_jsdl = params[0]->vstring;
	else
		input_jsdl = chained_params[1]->vstring;

	SM_TRY_NOT(mh = sm_gsoap_buffer_get(input_jsdl, &jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", xmlstructs_namespaces));

	if (jsdl.jsdl__JobDescription->jsdl__Application &&
		jsdl.jsdl__JobDescription->jsdl__Application->jsdl__ApplicationName) {

		const char *appname = jsdl.jsdl__JobDescription->jsdl__Application->jsdl__ApplicationName;
		const char *appversion = jsdl.jsdl__JobDescription->jsdl__Application->jsdl__ApplicationVersion;
		const char *executable = NULL;
		_jsdl_hpcpa__HPCProfileApplication *hpcpa = NULL;
		sm_list_t l;

		for (l = map_list; l; l = l->next) {
			application_mapping_t map = l->data;

			if (sm_strcmp(map->name, appname))
				continue;

			sm_debug("Found application entry");

			if (appversion == NULL || !sm_strcmp(appversion, map->version) || !sm_strcmp(map->version, "*")) {
				executable = map->path;
				break;
			}
		}

		if (!executable) {
			sm_notice("No matching entry found for ApplicationName = %s, ApplicationVersion = %s", appname, SM_SAFE_STR(appversion));

			/* There is no appropriate fault in BES specification
			 * for failures like these... */
			result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_INVALID_REQUEST_MESSAGE_FAULT, "The specified JSDL Application is not known", SM_VALUE_NONE);
			goto fault;
		}

		sm_debug("ApplicationName resolved to: %s", executable);

		hpcpa = jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;

		if (!hpcpa) {
			SM_TRY_NOT(jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication = sm_alloc_tracked_malloc(mh, sizeof(struct jsdl_hpcpa__HPCProfileApplication_USCOREType)));
			hpcpa = jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;
		}

		hpcpa->jsdl_hpcpa__Executable = sm_alloc_tracked_malloc(mh, sizeof(struct jsdl_hpcpa__FileName_USCOREType));
		hpcpa->jsdl_hpcpa__Executable->__item = (char *) executable;

		SM_TRY_NOT(output_jsdl = sm_gsoap_buffer_put(&jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", xmlstructs_namespaces));

		result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_OK, SM_VALUE_STRING, output_jsdl);
		SM_FREE(output_jsdl);
		SM_TRY_NOT(result);

	} else
		SM_TRY_NOT(result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_OK, SM_VALUE_STRING, input_jsdl));

	sm_alloc_tracked_free(&mh);
	return result;

fault:
	sm_alloc_tracked_free(&mh);

	signal->stop_emission(signal);

	if (result)
		return result;
	else
		return sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_INTERNAL_FAULT, "Internal fault in application_mapper", SM_VALUE_NONE);
}

sm_value_t *
handle_factory_attributes(sm_signal_t sig, sm_value_t * drms_type, const sm_value_t *chain)
{

	return sm_value_create(1, SM_VALUE_STRING, applications_xml);
}
