/* $Id: db_sandbox.c 25 2009-10-30 17:32:01Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/odbc.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <db_sandbox.h>

__SM_ID("$Id: db_sandbox.c 25 2009-10-30 17:32:01Z mamonski $");

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "db_sandbox"
        }
};

static sm_odbc_handle_t odbc;

static const char *clean_database_txt = "DELETE FROM accounts_pool";
static const char *add_account_entry_txt = "INSERT INTO accounts_pool VALUES (? , ?, ?, 0, NULL, NULL, NULL)";
static const char *acquire_account_from_pool_txt = "SELECT account_name FROM accounts_pool WHERE ref_count = 0 LIMIT 1 FOR UPDATE";
static const char *inc_account_ref_txt = "UPDATE accounts_pool SET ref_count = ref_count + 1 WHERE account_name = ?";
static const char *dec_account_ref_txt = "UPDATE accounts_pool SET ref_count = ref_count - 1 WHERE account_name = ?";
static const char *get_ref_count_txt = "SELECT home_dir, uid, ref_count FROM accounts_pool WHERE account_name = ? FOR UPDATE";

static sm_odbc_stmt_t clean_database_stmt = NULL;
static sm_odbc_stmt_t add_account_entry_stmt = NULL;
static sm_odbc_stmt_t acquire_account_from_pool_stmt = NULL;
static sm_odbc_stmt_t inc_account_ref_stmt = NULL;
static sm_odbc_stmt_t dec_account_ref_stmt = NULL;
static sm_odbc_stmt_t get_ref_count_stmt = NULL;

sm_status_ret_code_t
smc_db_sandbox_setup(sm_conf_ctxt_t conf)
{
	const char *db = NULL;
	const char *name = NULL;
	const char *pass = NULL;

	SM_MANDATORY_P(conf);

	SM_DEBUG_ENTER();

	if (sm_conf_node_str(conf, SM_TRUE, "//conf:Database/conf:DSN", &db) != SM_OK)
		return SM_FAULT;

	if (sm_conf_node_str(conf, SM_TRUE, "//conf:Database/conf:User", &name) != SM_OK)
		return SM_FAULT;

	if (sm_conf_node_str(conf, SM_TRUE, "//conf:Database/conf:Password", &pass) == SM_FAULT)
		return SM_FAULT;

	if (!(odbc = sm_odbc_connect(db, name, pass))) {
		const char *p = NULL;
		if (pass)
			p = "...";
		sm_error("Failed to connect to database: name = \"%s\", user = \"%s\", pass = \"%s\"", SM_SAFE_STR(db), SM_SAFE_STR(name), SM_SAFE_STR(p));
		return SM_FAULT;
	}

	SM_TRY_FAULT(odbc->begin_transaction(odbc));

	/* prepare statements */
	SM_TRY_NOT(clean_database_stmt = sm_odbc_stmt_create(odbc, clean_database_txt));
	SM_TRY_NOT(add_account_entry_stmt = sm_odbc_stmt_create(odbc, add_account_entry_txt));
	SM_TRY_NOT(acquire_account_from_pool_stmt = sm_odbc_stmt_create(odbc, acquire_account_from_pool_txt));
	SM_TRY_NOT(inc_account_ref_stmt = sm_odbc_stmt_create(odbc, inc_account_ref_txt));
	SM_TRY_NOT(dec_account_ref_stmt = sm_odbc_stmt_create(odbc, dec_account_ref_txt));
	SM_TRY_NOT(get_ref_count_stmt = sm_odbc_stmt_create(odbc, get_ref_count_txt));

	SM_TRY_FAULT(odbc->commit(odbc));

	return SM_OK;
fault:
	odbc->rollback(odbc);
	sm_odbc_disconnect(&odbc);
	return SM_FAULT;
}

void
smc_db_sandbox_cleanup(void)
{
	SM_DEBUG_ENTER();

	sm_odbc_stmt_free(&clean_database_stmt);
	sm_odbc_stmt_free(&add_account_entry_stmt);
	sm_odbc_stmt_free(&acquire_account_from_pool_stmt);
	sm_odbc_stmt_free(&inc_account_ref_stmt);
	sm_odbc_stmt_free(&dec_account_ref_stmt);
	sm_odbc_stmt_free(&get_ref_count_stmt);

	sm_odbc_disconnect(&odbc);
}

int
smc_db_sandbox_clear_accounts_pool_relation()
{
	SM_DEBUG_ENTER();

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(clean_database_stmt->execute(clean_database_stmt, ""));
	SM_TRY_FAULT(odbc->commit(odbc));

	return SM_OK;
fault:
	odbc->rollback(odbc);

	return SM_FAULT;
}

int
smc_db_sandbox_add_account_entry(const char *name, const char *home_dir, int uid)
{
	SM_DEBUG_ENTER();

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(clean_database_stmt->execute(add_account_entry_stmt, "ssi", name, home_dir, &uid));
	SM_TRY_FAULT(odbc->commit(odbc));

	return SM_OK;
fault:
	odbc->rollback(odbc);

	return SM_FAULT;
}


int
smc_db_sandbox_acquire_account(char **account_name)
{
	int ret = SM_FAULT;

	SM_DEBUG_ENTER();

	*account_name = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(clean_database_stmt->execute(acquire_account_from_pool_stmt, ""));
	SM_TRY_NOT_OK(ret = acquire_account_from_pool_stmt->fetch_row(acquire_account_from_pool_stmt));
	SM_TRY_NOT_OK(ret = acquire_account_from_pool_stmt->get_string(acquire_account_from_pool_stmt, 1, account_name));
	SM_TRY_FAULT(acquire_account_from_pool_stmt->close_cursor(acquire_account_from_pool_stmt));

	sm_info("acquired account = %s", *account_name);

	SM_TRY_FAULT(inc_account_ref_stmt->execute(inc_account_ref_stmt, "s", *account_name));

	SM_TRY_FAULT(odbc->commit(odbc));

	return SM_OK;
fault:
	acquire_account_from_pool_stmt->close_cursor(acquire_account_from_pool_stmt);
	odbc->rollback(odbc);

	return ret;
}

int
smc_db_sandbox_release_account(const char *account_name, int (*purge_callback)(const char *username, const char *homedir, int uid))
{
	int ret = SM_FAULT;
	int ref_count = 0;
	char *home_dir = NULL;
	int uid = -1;

	SM_DEBUG_ENTER();

	sm_info("release account: %s", account_name);

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(get_ref_count_stmt->execute(get_ref_count_stmt, "s", account_name));
	SM_TRY_NOT_OK(ret = get_ref_count_stmt->fetch_row(get_ref_count_stmt));
	SM_TRY_NOT_OK(ret = get_ref_count_stmt->get_string(get_ref_count_stmt, 1, &home_dir));
	SM_TRY_NOT_OK(ret = get_ref_count_stmt->get_int(get_ref_count_stmt, 1, &uid));
	SM_TRY_NOT_OK(ret = get_ref_count_stmt->get_int(get_ref_count_stmt, 1, &ref_count));
	SM_TRY_FAULT(get_ref_count_stmt->close_cursor(get_ref_count_stmt));

	if (ref_count == 1 && purge_callback) {
		/* TODO time measure */
		sm_info("calling purge callback: %s %s %d", account_name, home_dir, uid);
		ret = purge_callback(account_name, home_dir, uid);

		if (ret != SM_OK) {
			sm_error("purging of account %s failed", account_name);
			goto fault;
		}
	}

	SM_TRY_FAULT(dec_account_ref_stmt->execute(dec_account_ref_stmt, "s", account_name));

	SM_TRY_FAULT(odbc->commit(odbc));

	SM_FREE(home_dir);
	return SM_OK;
fault:
	get_ref_count_stmt->close_cursor(get_ref_count_stmt);
	SM_FREE(home_dir);
	odbc->rollback(odbc);

	return ret;
}
