/* $Id: notification_wsn.c 122 2010-06-09 14:10:00Z mamonski $ */

/*
 *  Copyright 2006-2010 QCG Project Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <qcg-core/sm.h>
#include <qcg-core/conf.h>
#include <qcg-core/client.h>
#include <qcg-core/logger.h>
#include <qcg-core/string.h>
#include <qcg-core/signal.h>
#include <qcg-core/thread.h>
#include <qcg-core/uuid.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/general_module.h>
#include <qcg-core/modules/ntf_client.h>

#include <activemqcpp.h>

#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>

__SM_ID("$Id$");

#define SMC_BAT_MESSAGE_TEMPLATE    "\n<?xml version=\"1.0\" encoding=\"UTF-8\"?> <site name=\"%s\"> \n\
   <job>\n\
        \t<batch_server>%s</batch_server>\n\
        \t<job_id>%s</job_id>\n\
        \t<user>%s</user>\n\
        \t<group>%s</group>\n\
%s\
        \t<queue>%s</queue>\n\
        \t<ctime>%d</ctime>\n\
        \t<qtime>%d</qtime>\n\
        \t<etime>%d</etime>\n\
        \t<start>%d</start>\n\
        \t<end>%d</end>\n\
        \t<exec_host>\n%s\
        \t</exec_host>\n\
        \t<cputime>%d</cputime>\n\
        \t<walltime>%d</walltime>\n\
        \t<mem>%d</mem>\n\
        \t<vmem>%d</vmem>\n\
        \t<estatus>%d</estatus>\n\
        \t<infrastructure>%s</infrastructure>\n\
        \t<grid_job_id>\"%s\"</grid_job_id>\n\
        \t<userDN>\"%s\"</userDN>\n\
    </job>\n\
</site>"

#define SMC_BAT_NODE_TEMPLATE    "\t\t<node>\n\
		\t<nodename>%s</nodename>\n\
		\t<cpu>%s</cpu>\n\
		</node>\n"

#define SMC_BAT_ACCOUNT_TEMPLATE "\t\t<account>%s</account>\n"

/*
 *
example message taken from https://wiki.plgrid.pl/doku.php?id=pakiet3:bat:report-protocol
<site name="CYFRONET-LCG2">
    <job>
        <batch_server>batch.cyf-kr.edu.pl</batch_server> -> gethostname ()
        <job_id>1233</job_id> OK
        <user>plguser</user> OK
        <group>plggroup</group> from <user> -> getpwnam_r() -> getgrgid_r() 
        <account>nazwa_grantu</account> -> from JSDL (project)
        <queue>plgrid</queue> TODO rusage:queue
        <ctime>1258745240</ctime>
        <qtime>1258745240</qtime>
        <etime>1258745240</etime> ctime = qtime = etime
        <start>1258745241</start> OK
        <end>1258745364</end> OK
        <exec_host> TODO : nodes
        <node>
            <nodename>pl-wn001</nodename>
            <cpu>0</cpu>
        </node>
                <node>
            <nodename>pl-wn002</nodename>
            <cpu>1</cpu>
        </node>
        </exec_host>
        <cputime>39</cputime> CHECK:rusage
        <walltime>120</walltime> end - start TODO: add walltime in rusage (what if job was suspended)
        <mem>4000</mem> CHECK:rusage
        <vmem>20000</vmem> TODO:rusage
        <estatus>0</estatus> OK
        <infrastructure>glite</infrastructure> OK -> "qcg"
        <vo>plgrid</vo> SKIP
        <grid_job_id>"https://lb.grid.cyf-kr.edu.pl:9000/0fCVRLjzwGfO1TGQA7HX1g"</grid_job_id> OK
        <userDN>"/C=PL/O=GRID/O=Cyfronet/CN=Wojciech Ziajka"</userDN> OK
        <userFQAN>"/vo.plgrid.pl/grant1/Role=NULL/Capability=NULL"</userFQAN> SKIP
    </job>
</site>
*/

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static int connect(void);

static sm_value_t *handle_finished(sm_signal_t, sm_value_t *, const sm_value_t *);


static const char *bat_service_url = NULL;
static const char *site_name = NULL;
static const char *queue_name = NULL;
static char hostname[4096] = "";

static const char * url = NULL;

#define MAX_RETRY 8
#define SUSPEND_TIME 3600
#define CONNECTION_TIMEOUT 4000
static unsigned connect_retry = 0;
static time_t down_time = 0;
static int connected = 0;

SM_GENERAL_MODULE = {
	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "bat_updater",
		.descr = "PL-Grid accounting module",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	}
};

int
connect()
{
	time_t temp_time = time(NULL);

	if(connect_retry == MAX_RETRY && down_time != 0 && temp_time - down_time >= SUSPEND_TIME) {
		connect_retry = 0;
		down_time = 0;
		sm_info("Setting connect_retry to 0");
	}	

	if(url == NULL)
	{
		if((url = sm_asprintf("%s?soConnectTimeout=%d",bat_service_url,CONNECTION_TIMEOUT)) == NULL)
		{
			sm_error("Memory allocation wasn't possible");
			return 2;
		}
		sm_info("Connecting using URL: %s",url);
	}

	if(connect_retry < MAX_RETRY && activemqcpp_connect(url, queue_name) == 0) { 
		sm_info("ActiveMQ connected to %s, queue: %s", bat_service_url, queue_name); 
		connect_retry = 0;
		return 0;
	} else if (connect_retry < MAX_RETRY) {
		connect_retry++;
		sm_info("Connect retry: %d/%d\n",connect_retry,MAX_RETRY); //
		sm_error("Failed to connect to  %s, queue: %s", bat_service_url, queue_name);
		return 1;
	} else if (connect_retry == MAX_RETRY && down_time == 0) {
		down_time = time(NULL);
		sm_error("BAT service marked as down");
		return 2;
	} else {
		return 2;
	}	
}

int
setup(sm_module_t this)
{
	SM_DEBUG_ENTER();

	if (sm_signal_connect("smc_job_finished", handle_finished) == SM_FAULT) {
		sm_error("Failed to connect to smc_job_finished signal");
		return SM_FAULT;
	}

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:BATServiceURL", &bat_service_url) != SM_OK) {
		sm_error("Failed to read BATServiceURL");
		return SM_FAULT;
	}

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:SiteName", &site_name) != SM_OK) {
		sm_error("Failed to read SiteName");
		return SM_FAULT;
	}

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:QueueName", &queue_name) != SM_OK) {
		sm_error("Failed to read QueueName");
		return SM_FAULT;
	}

	/* hostname */
	if(gethostname(hostname, sizeof(hostname)) == -1) { 
		sm_perror("gethostname() failed/");
		return SM_FAULT;
	} else {
		sm_info("batch_server = %s",hostname);
	}

	sm_info("BAT Service URL = %s", bat_service_url);
	sm_info("Site name       = %s", site_name);
	sm_info("Queue name      = %s", queue_name);
	sm_info("Batch server    = %s", hostname);

	connected = connect();	
	return SM_OK;
}





void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();

	if( connected == 0 && activemqcpp_disconnect() == 0) { 
		sm_info("ActiveMQ: disconnected"); 
	}
	else {	
		sm_error("ActiveMQ: could not disconnect"); 
	}

	sm_free(url);
}	


sm_value_t *
handle_finished(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	static char * temp_xml = NULL;

 	struct passwd pd;
  	struct passwd *tempPwdPtr = NULL;
	char pwdbuffer[4096] = "";

	struct group grp;
 	struct group *tempGrpPtr = NULL;
  	char grpbuffer[4096] = "";
  	
	char *temp_exec_host = "";
	char *temp_exec_host_old = NULL;
	char *temp_nodename = NULL;
	char *temp_node = NULL;
	char *temp_cpu = NULL;
	char *ptr = NULL;
	char *temp_ptr = NULL;
	char *rest = NULL;
	char *token = NULL;
	char *account = NULL;

	if(connected != 0)
	{
		sm_info("Trying to reconnect before sending msg");
		if((connected = connect()) != 0)
		{
			sm_error("Can't connect to BAT server");
			return SM_VALUE_NONE_SINGLETON;
		}
	}

	SM_DEBUG_ENTER_STR(params[0]->vstring);

	sm_info("grid_job_id = %s", params[0]->vstring);
	{
		char *p = strchr(params[1]->vstring, '.');
		/* remove submission host from job id*/
		if (p)
			*p = '\0';

	}
	sm_info("job_id = %s", params[1]->vstring);
	sm_info("user_dn = %s", params[5]->vstring); 
	sm_info("user = %s", params[6]->vstring);
	sm_info("ctime=qtime=etime = %d", params[7]->vint32); 
	sm_info("start = %d", params[8]->vint32);
	sm_info("end = %d", params[9]->vint32);
	sm_info("estatus = %d", params[10]->vint32);
	sm_info("cputime = %f", params[11]->vdouble);
	sm_info("mem = %f", params[12]->vdouble); 
	sm_info("vmem = %f", params[13]->vdouble);
	sm_info("infrastructure = qcg"); 	
	sm_info("queue = %s", params[14]->vstring);
	sm_info("project = %s", params[15]->vstring);
	sm_info("walltime = %d", params[16]->vint32);
	sm_info("nodes = %s", params[17]->vstring);
	sm_info("jsdl = %s", params[4]->vstring);
	 
 	/* group */
	if ((getpwnam_r(params[6]->vstring, &pd, pwdbuffer, sizeof(pwdbuffer), &tempPwdPtr))!=0 || tempPwdPtr == NULL) {
     		sm_perror("getpwnam_r() failed.");
		goto end;
	}

	sm_debug("group id: %d",pd.pw_gid);	
	if ((getgrgid_r(pd.pw_gid,&grp,grpbuffer,sizeof(grpbuffer),&tempGrpPtr))!=0 || tempGrpPtr == NULL) {
		sm_perror("getgrgid_r() error.");
		goto end;
	} else {
		sm_info("group = %s", grp.gr_name);
 	}

	/* exec_host */
	ptr = sm_strdup(params[17]->vstring);
	do
	{
		token = strtok_r(ptr, "/", &rest);
		if (token == NULL)
		{
			sm_error("Illegal exec_host format: %s", params[17]->vstring);
			goto end;
		}

		temp_nodename = sm_strdup(token);

		if (ptr!=NULL)
		{
			temp_ptr = ptr;
			ptr = NULL;
	    	}

		token = strtok_r(NULL, "+", &rest);

		if (token == NULL)
		{
			sm_error("Illegal exec_host format: %s", params[17]->vstring);
			goto end;
		}

		temp_cpu = sm_strdup(token);

		if ((temp_node = sm_asprintf(SMC_BAT_NODE_TEMPLATE,temp_nodename,temp_cpu)) == NULL) { 
			sm_error("Memory allocation wasn't possible");
			goto end;
		}

		sm_free(temp_nodename);
		sm_free(temp_cpu); 

		temp_exec_host_old = sm_strdup(temp_exec_host);

		if (strcmp(temp_exec_host, "") != 0) {		
			sm_free(temp_exec_host);
		}

		if ((temp_exec_host = sm_asprintf("%s%s",temp_exec_host_old,temp_node)) == NULL) {
			sm_error("Memory allocation wasn't possible");
			goto end;
		}
		sm_free(temp_exec_host_old);
		
		sm_free(temp_node);
	}
	while( strcmp(rest,"") !=0 );
	
	sm_free(temp_ptr);
	
	if ( params[15]->vstring ) {
		account = sm_asprintf(SMC_BAT_ACCOUNT_TEMPLATE, params[15]->vstring);
	} else {
		account = sm_strdup("");
	}

	if ((temp_xml = sm_asprintf(SMC_BAT_MESSAGE_TEMPLATE, 
		/*site name*/ site_name,
		/*batch server*/  hostname, 
		/*job_id*/params[1]->vstring, 
		/*user*/ params[6]->vstring,
		/*group*/ grp.gr_name,  
		/*account*/ account,  
		/*queue*/ params[14]->vstring, 
		/*ctime*/ params[7]->vint32, 
		/*qtime*/ params[7]->vint32, 
		/*etime*/ params[7]->vint32,  
		/*start*/ params[8]->vint32,
		/*end*/ params[9]->vint32,
		/*exec_host*/ temp_exec_host,
		/*cpu time*/ (int) params[11]->vdouble, 
		/*wall time*/ params[16]->vint32,  
		/*mem*/ (int) params[12]->vdouble, 
		/*vmem*/ (int) params[13]->vdouble,   
		/*estatus*/ params[10]->vint32, 
		/*infrasructure*/ "qcg",  
		/*grid_job_id*/ params[0]->vstring, 
		/*userDN*/ params[5]->vstring)) == NULL) {
		sm_error("Memory allocation wasn't possible");
		goto end;
	}
	
	sm_info("Sending rusage record: %s", temp_xml);

	if (activemqcpp_send_message(temp_xml) == 0) {
		sm_info("Usage record successfully sent");
	} else {	 
		sm_info("Reconnecting before send");
		if((connected = connect()) == 0 && activemqcpp_send_message(temp_xml) == 0)
			sm_info("Usage record successfully sent");
		else
			sm_error("Can't send usage record to BAT service");
	}

	sm_free(temp_exec_host);

end:	
	sm_free(account);
	sm_free(temp_xml);
	return SM_VALUE_NONE_SINGLETON;
}


