/* $Id: monitoring_drms.c 66 2010-03-31 15:26:16Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>

#include <qcg-core/sm.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/python.h>
#include <qcg-core/utils.h>
#include <qcg-core/value.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

__SM_ID("$Id: monitoring_drms.c 66 2010-03-31 15:26:16Z mamonski $");

typedef struct reservation_python_ctxt_s {
	char *module_name;
	sm_python_object_t module;
} *reservation_python_ctxt_t;

static int setup(sm_module_t);
static void cleanup(sm_module_t);

static sm_value_t *factory_attr_handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);
static sm_value_t *contained_resources_handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);
static sm_value_t *job_info_handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);

SM_GENERAL_MODULE = {

	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "monitoring_drms",
		.descr = "Cluster monitoring module",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},
};

static int
setup(sm_module_t this)
{
	char *base = NULL, *p = NULL, *config_xml = NULL;
	const char *path = NULL;
	reservation_python_ctxt_t ctxt = NULL;	

	SM_DEBUG_ENTER();
	
	if (!(ctxt = sm_malloc(sizeof(*ctxt))))
		goto fault;

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/@path", &path) != SM_OK)
		goto fault;

	if (sm_conf_node_raw(this->conf_ctxt, SM_TRUE, "/", &config_xml) != SM_OK)
		goto fault;

	if (!(base = sm_basename(path)))
		goto fault;

	if ((p = strchr(base, '.')))
		*p = '\0';

	ctxt->module_name = base;

	sm_info("Setting up: %s", path);

	if (!(ctxt->module = sm_python_load_module(path, config_xml,
					sm_conf_ctxt_get_default_prefix(this->conf_ctxt),
					sm_conf_ctxt_get_service_prefix(this->conf_ctxt),
					sm_conf_ctxt_get_service_namespace(this->conf_ctxt)))) {
		sm_error("Failed to load module: %s", path);
		goto fault;
	}
	
	if (!sm_python_object_has_attr(ctxt->module, "go", SM_TRUE)) {
		sm_error("The module doesn't define go function");
		goto fault;
	}

	SM_TRY_FAULT(sm_signal_connect("smc_factory_attributes", factory_attr_handler));
	SM_TRY_FAULT(sm_signal_connect("smc_contained_resources", contained_resources_handler));
	SM_TRY_FAULT(sm_signal_connect("smc_job_info", job_info_handler));

	SM_FREE(config_xml);

	this->ctxt = ctxt;

	return SM_OK;

fault:

	if (ctxt->module)
		sm_python_unload_module(base);
	SM_FREE(base);
	SM_FREE(config_xml);
	SM_FREE(ctxt);
	return SM_FAULT;
}

static void
cleanup(sm_module_t this sm_unused)
{
	reservation_python_ctxt_t ctxt;
	SM_DEBUG_ENTER();

	if (!this->ctxt)
		return;

	ctxt = this->ctxt;

	sm_python_unload_module(ctxt->module_name);
	SM_FREE(this->ctxt);
}

static sm_value_t *
factory_attr_handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	sm_info("factory_attr handler called for DRMS type: %s", params[0]->vstring);

	return sm_value_create(1, SM_VALUE_STRING, "<ClusterMonitoring xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/monitoring\"> \
			<Queues> \
			 <Queue name='short' walltime='xml duration' default='true' priority='30' maxJobs='100' runningJobs='20' waitingJobs='40'/> \
			</Queues> \
			<Reservations> \
			<Reservation name='foo' type='ar' startTime='XMLdate' endTime='opcjonalny' owner='versor' > \
				<Resource name='grass1' reservedSlots='3'/> \
			</Reservation> \
			</Reservations> \
			</ClusterMonitoring>");
}

static sm_value_t *
contained_resources_handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	char *contained_resources = {"<ContainedResources xmlns:bes-factory=\"http://schemas.ggf.org/bes/2006/08/bes-factory\" xmlns:jsdl=\"http://schemas.ggf.org/jsdl/2005/11/jsdl\">\
                                        <ContainedResource>\
                                                <bes-factory:ResourceName>alpha.example.com</bes-factory:ResourceName>\
                                                <bes-factory:CPUArchitecture>\
                                                        <jsdl:CPUArchitectureName>x86_64</jsdl:CPUArchitectureName>\
                                                </bes-factory:CPUArchitecture>\
												<bes-factory:CPUCount>10</bes-factory:CPUCount>\
                                                <bes-factory:PhysicalMemory>1073741824</bes-factory:PhysicalMemory>\
                                                <bes-factory:VirtualMemory>1073741824</bes-factory:VirtualMemory>\
                                                <Utilization xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/monitoring\">100</Utilization> \
                                        </ContainedResource>\
                                        <ContainedResource>\
                                                <bes-factory:ResourceName>beta.example.com</bes-factory:ResourceName>\
                                                <bes-factory:CPUArchitecture>\
                                                        <jsdl:CPUArchitectureName>x86_64</jsdl:CPUArchitectureName>\
                                                </bes-factory:CPUArchitecture>\
                                                <bes-factory:CPUCount>2</bes-factory:CPUCount>\
                                                <bes-factory:PhysicalMemory>1073741824</bes-factory:PhysicalMemory>\
                                                <Utilization xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/monitoring\">100</Utilization> \
                                        </ContainedResource>\
                                </ContainedResources>"};

	sm_info("contained_resources_handler handler called for DRMS type: %s", params[0]->vstring);

	return sm_value_create(1, SM_VALUE_STRING, contained_resources);
}

static sm_value_t *
job_info_handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	assert(SM_VALUE_IS_STRING(params[0]));
	assert(SM_VALUE_IS_STRING(params[1]));
	sm_info("job_info_handler  called for local job id: %s (DRMS type = %s)", params[0]->vstring, params[1]->vstring);

	return sm_value_create(1, SM_VALUE_STRING, "<JobMonitoring xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/monitoring\"> \
			<ExecutionHosts><Host name='alpha.example.com'/></ExecutionHosts><CPUTimeUsage>1:00:00</CPUTimeUsage><MemUsage>4000000</MemUsage></JobMonitoring>");
}


