/* $Id: notification_wsn.c 122 2010-06-09 14:10:00Z mamonski $ */

/*
 *  Copyright 2006-2009 QCG Project Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include <qcg-core/sm.h>
#include <qcg-core/conf.h>
#include <qcg-core/client.h>
#include <qcg-core/logger.h>
#include <qcg-core/string.h>
#include <qcg-core/signal.h>
#include <qcg-core/thread.h>
#include <qcg-core/uuid.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/general_module.h>
#include <qcg-core/modules/ntf_client.h>


__SM_ID("$Id: notification_wsn.c 122 2010-06-09 14:10:00Z mamonski $");

#define FULL_TOPIC_DIALECT                   "http://docs.oasis-open.org/wsn/t-1/TopicExpression/Full"
#define NOTIFY_ACTION                        "http://docs.oasis-open.org/wsn/bw-2/NotificationConsumer/Notify"
#define REGISTER_PUBLISHER_REQUEST_ACTION    "http://docs.oasis-open.org/wsn/brw-2/RegisterPublisher/RegisterPublisherRequest"

#define SMC_NOTIFICATION_TOPIC_NAMESPACE     "http://schemas.qoscosgrid.org/comp/2011/04/notification/topic"
#define SMC_NOTIFICATION_MESSAGE_NAMESPACE   "http://schemas.qoscosgrid.org/comp/2011/04/notification/message"

#define SMC_MESSAGE_TEMPLATE                   "<smcm:ActivityState xmlns:smcm='" SMC_NOTIFICATION_MESSAGE_NAMESPACE "'>" \
                                               "  <smcm:ActivityId>%s</smcm:ActivityId>" \
                                               "  <smcm:Status>%s</smcm:Status>" \
                                               "</smcm:ActivityState>"
#define SMC_MESSAGE_WITH_EXIT_STATUS_TEMPLATE  "<smcm:ActivityState xmlns:smcm='" SMC_NOTIFICATION_MESSAGE_NAMESPACE "'>" \
                                               "  <smcm:ActivityId>%s</smcm:ActivityId>" \
                                               "  <smcm:Status>%s</smcm:Status>" \
                                               "  <smcm:ExitStatus>%d</smcm:ExitStatus>" \
                                               "</smcm:ActivityState>"

static int setup(sm_module_t);
static void cleanup(sm_module_t);

static sm_value_t *handle_queued(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_executing(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_finished(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_failed(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_suspended(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_held(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_cancelled(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_stage_in(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_stage_out(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_factory_attributes(sm_signal_t, sm_value_t *, const sm_value_t *);


static sm_status_ret_code_t send_notify(char *topic, const char *activity_id, const char *status, const int *exit_code);

static sm_ntf_client_api_t ntf_api = NULL;
static char* publisher_id = NULL;
static sm_ecm_client_module_t ntf_module;
static char *np_service_url = NULL;

SM_GENERAL_MODULE = {
	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "notification_wsn",
		.descr = "WS-BaseNotifications module",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	}
};

int
setup(sm_module_t this)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	ret  = sm_signal_connect("smc_job_queued", handle_queued);
	ret |= sm_signal_connect("smc_job_executing", handle_executing);
	ret |= sm_signal_connect("smc_job_finished", handle_finished);
	ret |= sm_signal_connect("smc_job_failed", handle_failed);
	ret |= sm_signal_connect("smc_job_held", handle_held);
	ret |= sm_signal_connect("smc_job_suspended", handle_suspended);
	ret |= sm_signal_connect("smc_job_cancelled", handle_cancelled);
	ret |= sm_signal_connect("smc_job_stage_in", handle_stage_in);
	ret |= sm_signal_connect("smc_job_stage_out", handle_stage_out);
	ret |= sm_signal_connect("smc_factory_attributes", handle_factory_attributes);

	if (ret & SM_FAULT) {
		sm_error("Failed to connect to signals");
		return SM_FAULT;
	}

	/* get the WS-Notification client module */
	ntf_module = (sm_ecm_client_module_t) this->module_manager->find_by_name_first(this->module_manager, "ntf_client");
	if (!ntf_module) {
		sm_error("Failed to find ntf_client module");
		return SM_FAULT;
	}

	ntf_api = ntf_module->client_api;



	// TODO registration

	return SM_OK;
}

void
cleanup(sm_module_t this sm_unused)
{
	SM_FREE(np_service_url);
}

sm_value_t *
handle_queued(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	send_notify("ActivityState/Queued", id, "Queued", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_executing(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	send_notify("ActivityState/Executing", id, "Executing", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_held(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	send_notify("ActivityState/Held", id, "Held", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_suspended(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	send_notify("ActivityState/Suspended", id, "Suspended", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_finished(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	int exit_status = params[10]->vint32;

	SM_DEBUG_ENTER_STR(id);
	send_notify("ActivityState/TerminalState/Finished", id, "Finished", &exit_status);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_failed(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	SM_DEBUG_ENTER_STR(id);
	send_notify("ActivityState/TerminalState/Failed", id, "Failed", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_cancelled(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	SM_DEBUG_ENTER_STR(id);
	send_notify("ActivityState/TerminalState/Cancelled", id, "Cancelled", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_stage_in(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	SM_DEBUG_ENTER_STR(id);
	send_notify("ActivityState/Stage-in", id, "Stage-in", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_stage_out(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *id = params[0]->vstring;
	SM_DEBUG_ENTER_STR(id);
	send_notify("ActivityState/Stage-out", id, "Stage-out", NULL);
	return SM_VALUE_NONE_SINGLETON;
}

sm_status_ret_code_t
send_notify(char *topic, const char *activity_id, const char *status, const int *exit_code)
{
	char message[1024];
	sm_client_transport_t transport = NULL;

	if (!(transport = sm_client_create_transport(ntf_module))) {
		sm_error("Failed to create client transport");
		return SM_FAULT;
	}

	if (exit_code) {
		snprintf(message, sizeof(message), SMC_MESSAGE_WITH_EXIT_STATUS_TEMPLATE, activity_id, status, *exit_code);

		sm_info("Sending notify: %s -> %s (exit_status=%d)", activity_id, status, *exit_code);
	} else {
		snprintf(message, sizeof(message), SMC_MESSAGE_TEMPLATE, activity_id, status);

		sm_info("Sending notify: %s -> %s", activity_id, status);
	}

	if (ntf_api->notify(transport, publisher_id, SMC_NOTIFICATION_TOPIC_NAMESPACE, topic, FULL_TOPIC_DIALECT, message) != SM_OK){
		sm_error("Failed to send notification to %s ", transport->service_url);

		sm_client_free_transport(&transport);
		return SM_FAULT;
	}

	sm_client_free_transport(&transport);
	return SM_OK;
}


static char*
get_np_service_url(void)
{
	/* TODO mutex */
	if (!np_service_url) {
		sm_client_transport_t tmp;

		if (!(tmp = sm_client_create_transport(ntf_module))) {
			sm_error("Failed to create client transport");
			return NULL;
		}

		np_service_url = sm_strdup(tmp->service_url);

		sm_client_free_transport(&tmp);
	}

	return np_service_url;
}

sm_value_t *
handle_factory_attributes(sm_signal_t sig sm_unused, sm_value_t * drms_type sm_unused, const sm_value_t *chain sm_unused)
{
	if (!np_service_url && !get_np_service_url())
		return SM_VALUE_NONE_SINGLETON;

	return sm_value_create(1, SM_VALUE_STRING | SM_VALUE_DONT_COPY_FLAG,
			sm_asprintf("<smcf:NotificationProviderURL xmlns:smcf='http://schemas.qoscosgrid.org/comp/2011/04/factory'>%s</smcf:NotificationProviderURL>", np_service_url));
}

