#!/usr/bin/env python

# $Id$

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import datetime, sys, subprocess, re, traceback, time
from common import HostInfo, QueueInfo, JobInfo, ReservationInfo
from gmexceptions import GMInternalError, GMExternalError, GMUnknowJob
import libxml2 #@UnresolvedImport

__jobsRegex = re.compile(r'(.*)/(.*)')
__statusRegex = re.compile(r'([^=]+)=([^,]*),*')
__queueStateRegex = re.compile(r'([^:]+):([^ ]*)[ ]*')
__execHostRegex = re.compile(r'([^/]+)/\d+\+*')
__valueRegex = re.compile(r'([\d\.]+)(\D*)')
__walltimeRegex = re.compile(r'([0-9]+):([0-9]+):([0-9]+)')
__resACLRegex = re.compile(r'^\s*ACL:.*USER==(\w+)', re.MULTILINE)
__resAccountingRegex = re.compile(r'^\s*Accounting.*User:(\w+)', re.MULTILINE)
__newlineRegex = re.compile(r'\n')
__emptyLineRegex = re.compile(r'\n[\s\n]*\n')
__equalRegex = re.compile(r'[ ]+=[ ]+')
__jobIdRegex = re.compile(r'[^\d]*([\d\-]+).*')
__queueRegex = re.compile(r'Queue: (.+)')

def __checknodeResRegexExp(value):
    return "User \'" + value + "\'.*Procs: (\d+)/(\d+)" 

__defaultOwner = "root"
__checknodeROI = "Reservations:"

__postfixTable = {
    "": 1,
    "kb": 1024,
    "mb": 1024*1024,
    "gb": 1024*1024*1024
}

def __extractPostFix(value):
    m = __valueRegex.match(value)
    ret = [m.group(1)]
    if m.group(2)!=None:
        ret.append(m.group(2))
    else:
        ret.append("")
    return ret

def __calculateValue(value):
    p = __extractPostFix(value)
    return float(round(float(p[0]) \
        *__postfixTable.get(p[1]))) 

def __commandOutputToDict(output, items_list=None, key_regex=None):
    blocks_list = __emptyLineRegex.split(output)
    items_dictionary = {}
    for block in blocks_list:
        if block.lstrip() == "":
            continue
        block_lines = __newlineRegex.split(block)

	if key_regex != None:
	    block_name = key_regex.match(block_lines[0]).group(1)
	else:
            block_name = block_lines[0]

        block_lines = block_lines[1:]
        block_dict = { }
        for line in block_lines:
            tuple = __equalRegex.split(line, 1)
            block_dict[ tuple[0].strip() ] = tuple[1].strip()

        if items_list != None:
        	items_list.append(block_name)

        items_dictionary[block_name] = block_dict
     
    return items_dictionary

def getHostData(hostnames=None):
    
    hostsInfo = []
    pbsnodes_output = _pbsnodes(['-a'])
    
    pbsnodes_nodes_list = __emptyLineRegex.split(pbsnodes_output)
    nodes_list = []
    nodes_dictionary = __commandOutputToDict(pbsnodes_output, nodes_list)

    for node_name in nodes_list:
        current_host = HostInfo()
        current_host.HostName = node_name;
        node_properties = nodes_dictionary[node_name];
        if node_properties['state'] != "free" and node_properties['state'] != "job-exclusive":
            continue 
        current_host.CpuCount = int(node_properties['np'])

        #jobs
        if 'jobs' in node_properties:
            m = __jobsRegex.match(node_properties['jobs'])
            while m!=None:
                current_host.Jobs.append(m.group(1))
                jobs = m.group(2)
                m = __jobsRegex.match(jobs)

        #decode status
        attributes = __decodeHostStatusField(node_properties['status'])
	current_host.OSName = attributes["opsys"]
	uname = attributes["uname"]
	if uname.find("x86_64") != -1:
	    current_host.CpuArch = "x86_64"
	elif uname.find("x86") != -1:
	    current_host.CpuArch = "x86"
	elif uname.find("IA-64") != -1:
	    current_host.CpuArch = "IA-64"

	if current_host.CpuArch == "Unknown" and properties != None:
	    if properties.find("x86_64") != -1:
	        current_host.CpuArch = "x86_64"
	    elif properties.find("x86") != -1:
	        current_host.CpuArch = "x86"
	    elif properties.find("IA-64") != -1:
	        current_host.CpuArch = "IA-64"

	current_host.CpuCount = int(attributes["ncpus"])
	current_host.TotalMemory = __calculateValue(attributes["physmem"])
	current_host.TotalSwap = __calculateValue(attributes["totmem"]) - current_host.TotalMemory
	current_host.CpuUtilization = float(attributes["loadave"])

        hostsInfo.append(current_host)

    return hostsInfo

def __decodeHostStatusField(status):
    ret = {} 
    for m in __statusRegex.finditer(status):
        ret[m.group(1)] = m.group(2)        
    return ret

def __decodeQueueStateField(state):
    ret = {}
    for m in __queueStateRegex.finditer(state):
        ret[m.group(1)] = m.group(2)
    return ret

def __decodeTimeDelta(time):
    match = __walltimeRegex.match(time)
    sec = int(match.group(1)) * 3600 + int(match.group(2)) * 60 + int(match.group(3))
    return datetime.timedelta(seconds=sec) 
        
def __decodeHostData(info):
    current_host = HostInfo()
    current_host.HostName = info[0]
    map = info[1]
    if "status" in map.keys():
        attributes = __decodeHostStatusField(info[1]["status"])
    else:
        return current_host
    
    if "properties" in map.keys():
        properties = info[1]["properties"]
    else:
        properties = None
            
    current_host.OSName = attributes["opsys"]   
    
    uname = attributes["uname"]
    if uname.find("x86_64") != -1:
        current_host.CpuArch = "x86_64"
    elif uname.find("x86") != -1:
        current_host.CpuArch = "x86"
    elif uname.find("IA-64") != -1:
        current_host.CpuArch = "IA-64"      

    if current_host.CpuArch == "Unknown" and properties != None:        
        if properties.find("x86_64") != -1:
            current_host.CpuArch = "x86_64"
        elif properties.find("x86") != -1:
            current_host.CpuArch = "x86"
        elif properties.find("IA-64") != -1:
            current_host.CpuArch = "IA-64"
    
    current_host.CpuCount = int(attributes["ncpus"])    
    current_host.TotalMemory = __calculateValue(attributes["physmem"])
    current_host.TotalSwap = __calculateValue(attributes["totmem"]) - current_host.TotalMemory
    current_host.CpuUtilization = float(attributes["loadave"])  

    jobs = attributes["jobs"]
    m = __jobsRegex.match(jobs)
    while m!=None:
        current_host.jobs.append(m.group(1))
        jobs = m.group(2)
        m = __jobsRegex.match(jobs)
    return current_host

def getJobData(jobids=None):
    jobsInfo = []
    if jobids != None:
        for jobid in jobids:        
            jobid = __jobIdRegex.match(jobid).group(1)
            qstat_output = _qstat(['-f', '-1', jobid])
            job_info = __commandOutputToDict(qstat_output, None, __jobIdRegex)[jobid]
            jobsInfo.append(__decodeJobData(jobid, job_info))   
    else:
        qstat_output = _qstat(['-f', '-1'])
        job_list = __commandOutputToDict(qstat_output, None, __jobIdRegex)
        for jobid,job_info in job_list.iteritems():
            jobsInfo.append(__decodeJobData(jobid, job_info))

    return jobsInfo

def __decodeJobData(jobid, attributes):
    current_job = JobInfo()
    current_job.JobID = jobid
    current_job.JobName = attributes["Job_Name"]
    if attributes["job_state"] == "R" or attributes["job_state"] == "E":
        current_job.JobStatus = "Running"
    elif attributes["job_state"] == "Q":
        current_job.JobStatus = "Pending"
    elif attributes["job_state"] == "C":
        current_job.JobStatus = "Finished"
    else:
        current_job.JobStatus = "Unknown"
    current_job.OwnerName = attributes["Job_Owner"]
    current_job.QueueName = attributes["queue"]
    current_job.SubmissionTime = datetime.datetime(*time.strptime(attributes["qtime"])[:6])
    if attributes["job_state"] == "Q":
        current_job.StartTime = None
    else:
        current_job.StartTime = datetime.datetime(*time.strptime(attributes["mtime"])[:6])
    hosts = {}
    if "exec_host" in attributes:
        for m in __execHostRegex.finditer(attributes["exec_host"]): 
            host_name = m.group(1)
            if host_name in hosts:
                hosts[host_name] += 1
            else:
                hosts[host_name]  = 1 

    for host,slots in hosts.iteritems():
        current_job.ExecutionHosts.append((host,slots))

    return current_job
    
    
def getQueueData(queuenames=None):
   
    #TODO try to obtain default queue name
    default_queue = None
    qstat_output = _qstat([ '-f', '-Q'])
    all_queues = __commandOutputToDict(qstat_output, None, __queueRegex)

    queuesInfo = [] 
    for queue_name, queue_dict in all_queues.iteritems():
            decodedQueue = __decodeQueueData(queue_name, queue_dict) 
            if decodedQueue.QueueName == default_queue:
                decodedQueue.IsDefault = True
            queuesInfo.append(decodedQueue)     

    return queuesInfo

def __decodeQueueData(queue_name, attributes):
    current_queue = QueueInfo()
    current_queue.QueueName = queue_name
    queue_state = __decodeQueueStateField(attributes["state_count"])    
    current_queue.PendingJobs = int(queue_state["Queued"])  
    current_queue.RunningJobs = int(queue_state["Running"])
    if "resources_default.walltime" in attributes.keys():
        current_queue.Walltime = __decodeTimeDelta(attributes["resources_default.walltime"])
    else:
        current_queue.Walltime = datetime.timedelta(0)  
    current_queue.job_slots_aval = 0
    current_queue.job_slots_total = 0
    if attributes["enabled"] == "True":
        current_queue.QueueEnabled = True
    else:
        current_queue.QueueEnabled = False  
    return current_queue

def _qstat(args):
    try:
        command = ['qstat']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("qstat command failed, reason:\n" + traceback.format_exc())        
    
    qstat_output = sr.communicate()[0]
    if sr.returncode != 0:
        if sr.returncode == 153:
            raise GMUnknowJob("Unknow Job")
        else:
            raise GMExternalError("qstat command returned code: " + str(sr.returncode) + ". Details: " + qstat_output)
    return qstat_output


def _pbsnodes(args):
    try:
        command = ['pbsnodes']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("pbsnodes command failed, reason:\n" + traceback.format_exc())        
    
    pbsnodes_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("pbsnodes command returned code: " + str(sr.returncode))
    return pbsnodes_output

def _showres(args):
    try:
        command = ['showres']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("showres command failed, reason:\n" + traceback.format_exc())     
    
    showres_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("showres command returned code: " + str(sr.returncode))
    return showres_output

def _checknode(args):
    try:
        command = ['checknode']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("checknode command failed, reason:\n" + traceback.format_exc())       
    
    checknode_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("checknode command returned code: " + str(sr.returncode))
    return checknode_output

def _diagnose(args):
    try:
        command = ['diagnose']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("diagnose command failed, reason:\n" + traceback.format_exc())        
    
    diagnose_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("diagnose command returned code: " + str(sr.returncode))
    return diagnose_output

def getReservationData(reservation_names=None):
    reservations_info = []
    showres_output = _showres(['-n', '--format=xml'])
    doc = libxml2.parseDoc(showres_output)
    ctx = doc.xpathNewContext()
    status_code = ctx.xpathEval("/response/status/@code")
    if len(status_code) != 1:
        raise GMInternalError("showres output is weird. Expected 1 status code attribute.")
    if status_code[0].content != "0":
        raise GMInternalError("showres status code was: " + status_code[0].content)
    reservation_doc = ctx.xpathEval("/response/data/node")
    
    reservations = {}
    
    for node in reservation_doc:        
        print node
        ctx.setContextNode(node)
        host_name = ctx.xpathEval("@name")[0].content
        nres = ctx.xpathEval("nres")[0]
        ctx.setContextNode(nres)
        res_name = ctx.xpathEval("@Name")[0].content        
        if(reservation_names != None and res_name not in reservation_names):
            continue
        
        res_type = ctx.xpathEval("@Type")[0].content
        if res_type == "User":
            resolved_res_type = "ar"
        elif res_type == "System":
            resolved_res_type = "sr"
        elif res_type == "Job":
            resolved_res_type = "job"
        else:
            raise GMInternalError("Reservation type: " + res_type + " is unknown")

        print "res_name = " + res_name
 
        if(res_name not in reservations.keys()):            
            reservations[res_name] = ReservationInfo()
            reservations[res_name].ReservationName = res_name                   
        else:
            reservations[res_name].Resources[host_name] = 1
            continue
        
        reservations[res_name].Resources[host_name] = 1
        reservations[res_name].Type = resolved_res_type
        
        start_time = ctx.xpathEval("@Start")[0].content
        end_time = ctx.xpathEval("@End")[0].content
        
        reservations[res_name].StartTime = datetime.datetime.fromtimestamp(int(start_time))
        reservations[res_name].EndTime = datetime.datetime.fromtimestamp(int(end_time))                     

    reservations_info.extend(reservations.values())
    checknode_map = {}
        
    return reservations_info

    for reservation in reservations_info:       
        diagnose_output = _diagnose(['-r', reservation.ReservationName])        
        aclMatch = __resACLRegex.search(diagnose_output)        
        if aclMatch != None:
            reservation.Owner = aclMatch.group(1)
        else:
            accMatch = __resAccountingRegex.search(diagnose_output)         
            if accMatch != None:
                reservation.Owner = accMatch.group(1)
            else:
                reservation.Owner = __defaultOwner
                
        for key in reservation.Resources.keys():
            if key not in checknode_map.keys():             
                checknode_output = _checknode([key])
                index = checknode_output.index(__checknodeROI)
                checknode_output = checknode_output[index+len(__checknodeROI):len(checknode_output)-1]              
                checknode_map[key] = checknode_output
            else:
                checknode_output = checknode_map[key]               
            
            regex = __checknodeResRegexExp(reservation.ReservationName)         
            match = re.findall(regex, checknode_output, re.MULTILINE|re.DOTALL)
            if len(match) != 1:
                raise GMInternalError("Could not find matching reservation in checknode output")
            if len(match[0]) != 2:
                raise GMInternalError("Could not find matching reservation in checknode output")
            reservation.Resources[key] = int(match[0][0])       
    
    return reservations_info
