# $Id$

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from qcg.core import *
from qcg.comp import *


try:
    import grid_monitoring
    from grid_monitoring.gmexceptions import GMInternalError, GMExternalError, GMUnknowJob
except ImportError:
    error("Unable to load grid_monitoring module")
    raise

from xml.dom.minidom import Document
import datetime

__MonitoringNSPrefix = "monitoring"
__MonitoringNS = "http://schemas.qoscosgrid.org/comp/2011/04/monitoring"
__BesFactoryNSPrefix = "bes-factory"
__BesFactoryNS = "http://schemas.ggf.org/bes/2006/08/bes-factory"
__JsdlNSPrefix = "jsdl"
__JsdlNS = "http://schemas.ggf.org/jsdl/2005/11/jsdl"

def __toXmlDuration(td):
    xmlPeriod = "PT"
    hours = td.days * 24
    hoursA = td.seconds / 3600
    hours = hours + hoursA
    sec = td.seconds - hoursA * 3600
    minutes = sec / 60
    sec = sec - minutes * 60
    xmlPeriod = xmlPeriod + str(hours) + "H" + str(minutes) + "M" + str(sec) + "S"
    return xmlPeriod

def __toXmlDateTime(dt):
    return dt.replace(microsecond=0).isoformat()

def jobMonitoringXmlFromJobInfo(jobInfo):    
    """Gets XML from JobInfo entity"""
    doc = Document()        
    jobMonitoring = doc.createElementNS(__MonitoringNS, "JobMonitoring")        
    jobMonitoring.setAttribute("xmlns", __MonitoringNS)
    doc.appendChild(jobMonitoring)
    executionHosts = doc.createElement("ExecutionHosts")
    jobMonitoring.appendChild(executionHosts)
    for hostTuple in jobInfo.ExecutionHosts:
        host = doc.createElement("Host")
        host.setAttribute("name", hostTuple[0])
        host.setAttribute("slots", str(hostTuple[1]))
        executionHosts.appendChild(host)

    cpuTimeUsage = doc.createElement("CPUTimeUsage")
        
    time = jobInfo.UserTime
    hours = time/3600
    time = time - hours*3600
    minutes = time/60
    time = time - minutes*60
    seconds = time
    userTime = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
    cpuTimeText = doc.createTextNode(str(userTime))        
        
    cpuTimeUsage.appendChild(cpuTimeText)
    jobMonitoring.appendChild(cpuTimeUsage)
        
    memUsage = doc.createElement("MemUsage")
    memUsageText = doc.createTextNode(str(jobInfo.TotalMem))
    memUsage.appendChild(memUsageText)
    jobMonitoring.appendChild(memUsage)
            
    return doc

def containedResourcesFromHostInfos(hostsInfo):
    """Gets XML from list of HostInfo entities"""
   
    doc = Document()

    cluster = doc.createElementNS(__BesFactoryNS, "ContainedResources")    
    cluster.setAttribute("xmlns:" + __BesFactoryNSPrefix, __BesFactoryNS)
    cluster.setAttribute("xmlns:" + __JsdlNSPrefix, __JsdlNS)
    doc.appendChild(cluster)        
    
    for host in hostsInfo:
        containedResource = doc.createElement("ContainedResource")
        cluster.appendChild(containedResource)
        
        resourceName = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":ResourceName")
        containedResource.appendChild(resourceName)
        text = doc.createTextNode(str(host.HostName))
        resourceName.appendChild(text)
        
        if not host.CpuArch == "Unknown":                         
            cpuArchitecture = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":CPUArchitecture")
            containedResource.appendChild(cpuArchitecture)
            cpuArchitectureName = doc.createElementNS(__JsdlNS, __JsdlNSPrefix + ":CPUArchitectureName")
            cpuArchitecture.appendChild(cpuArchitectureName)
            text = doc.createTextNode(str(host.CpuArch))
            cpuArchitectureName.appendChild(text)
        
        cpuCount = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":CPUCount")
        containedResource.appendChild(cpuCount)
        text = doc.createTextNode(str(int(host.CpuCount)))
        cpuCount.appendChild(text)

        if(host.TotalMemory == None):
            raise ValueError("TotalMemory cannot be None")
        physicalMemory = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":PhysicalMemory")
        containedResource.appendChild(physicalMemory)
        text = doc.createTextNode(str(int(host.TotalMemory)))
        physicalMemory.appendChild(text)
        
        if(host.TotalSwap == None):
            raise ValueError("TotalSwap cannot be None")
        virtualMemory = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":VirtualMemory")
        containedResource.appendChild(virtualMemory)
        text = doc.createTextNode(str(int(host.TotalSwap)))
        virtualMemory.appendChild(text)
        
        if(host.CpuUtilization == None):
            raise ValueError("CpuUtilization cannot be None")
        utilization = doc.createElementNS(__MonitoringNS, "Utilization")
        containedResource.appendChild(utilization)
        utilization.setAttribute("xmlns", __MonitoringNS)
        text = doc.createTextNode(str(int(host.CpuUtilization * 100)))
        utilization.appendChild(text)
        
    return doc

def clusterMonitoringFromQueuesAndRes(queuesInfo, reservationsInfo):
    """Gets XML from list of QueueInfo and ReservationInfo entities"""
       
    doc = Document()
    
    cluster = doc.createElementNS(__MonitoringNS, "ClusterMonitoring")    
    cluster.setAttribute("xmlns", __MonitoringNS)
    doc.appendChild(cluster)
    
    queues = doc.createElementNS(__MonitoringNS, "Queues")
    cluster.appendChild(queues)
    
    for queueInfo in queuesInfo:
        queue = doc.createElementNS(__MonitoringNS, "Queue")
        queue.setAttribute("name", queueInfo.QueueName)
        if queueInfo.Walltime == None:
            raise ValueError("Walltime must not be None")
        queue.setAttribute("walltime", __toXmlDuration(queueInfo.Walltime))
        queue.setAttribute("default", str(queueInfo.IsDefault))
        queue.setAttribute("priority", str(queueInfo.QueuePriority))
        queue.setAttribute("maxJobs", str(queueInfo.MaxJobs))
        queue.setAttribute("runningJobs", str(queueInfo.RunningJobs))
        queue.setAttribute("waitingJobs", str(queueInfo.PendingJobs))
        queues.appendChild(queue) 
    
    reservations = doc.createElementNS(__MonitoringNS, "Reservations")
    cluster.appendChild(reservations)
    
    for reservationInfo in reservationsInfo:
        
        reservation = doc.createElementNS(__MonitoringNS, "Reservation")
        reservation.setAttribute("name", reservationInfo.ReservationName)
        reservation.setAttribute("type", reservationInfo.Type)
        if reservationInfo.StartTime == None:
            raise ValueError("StartTime must not be None")
        reservation.setAttribute("startTime", __toXmlDateTime(reservationInfo.StartTime))
        if reservationInfo.EndTime != None:
            reservation.setAttribute("EndTime", __toXmlDateTime(reservationInfo.EndTime))
        reservation.setAttribute("owner", reservationInfo.Owner)
        reservations.appendChild(reservation)
        
        resources = reservationInfo.Resources
        for resourceKey in resources.keys():
            resource = doc.createElementNS(__MonitoringNS, "Resource")
            resource.setAttribute("name", resourceKey)
            resource.setAttribute("reservedSlots", str(resources[resourceKey]))
            reservation.appendChild(resource)
        
    return doc

def removeXmlHeader(string):
    return string.replace("<?xml version=\"1.0\" ?>", "")


_cap_map = grid_monitoring.monitoringCapabilities()
for i in _cap_map.keys():
    if _cap_map[i]:
        info("grid_monitoring reports " + i + " as available")
    else:
        info("grid_monitoring reports " + i + " as not available")

class MonitoringDRMS(Module):
    
    def __init__(self, *unused):
        #signal_connect("smc_factory_attributes", self.factory_attr_handler)
        signal_connect("smc_contained_resources", self.contained_resources_handler)
        #signal_connect("smc_job_info", self.job_info_handler)

    def factory_attr_handler(self, signal, drms_type, unused):
        queuesInfo = grid_monitoring.getQueueData(drms_type, None)
        reservationsInfo = grid_monitoring.getReservationData(drms_type, None)
        doc = clusterMonitoringFromQueuesAndRes(queuesInfo, reservationsInfo)                
        return removeXmlHeader(doc.toxml()) 
            
    def contained_resources_handler(self, signal, drms_type, unused):
        hostsInfo = grid_monitoring.getHostData(drms_type, None)
        doc = containedResourcesFromHostInfos(hostsInfo)                  
        return removeXmlHeader(doc.toxml())
                                
    def job_info_handler(self, signal, drms_type, job_id, unused):
        try:
            jobInfo = grid_monitoring.getJobData(drms_type, [job_id])[0]
            doc = jobMonitoringXmlFromJobInfo(jobInfo)
            return removeXmlHeader(doc.toxml())           
        except GMUnknowJob:
            return ""


