/* $Id: qcg_comp_client.c 302 2010-08-31 11:28:06Z mamonski $ */

/*
 *  Copyright 2006-2009 QCG Project Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <assert.h>
#include <stdlib.h>

#include <qcg-core/sm.h>
#include <qcg-core/closure.h>
#include <qcg-core/client.h>
#include <qcg-core/event_timer.h>
#include <qcg-core/logger.h>
#include <qcg-core/metric.h>
#include <qcg-core/python.h>
#include <qcg-core/signal.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/uuid.h>

#include <qcg-core/modules/atc_transport_module.h>
#include <qcg-core/modules/ecm_module.h>
#include <qcg-core/modules/ecm_client_module.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include "stub/qcgcomp.nsmap"
#include "stub/qcgcompStub.h"

#include <include/namespaces.h>

#ifdef WITH_LIBRSYNC
#include <rsync/client.h>
#endif

__SM_ID("$Id: qcg_comp_client.c 302 2010-08-31 11:28:06Z mamonski $");

static void print_usage(FILE *f);
static sm_status_ret_code_t opt_handler(char opt, char *optvalue);
static sm_status_ret_code_t args_handler(int argc, char **args);

static int start_client(sm_client_transport_t transport);
static void print_fault(sm_client_transport_t transport);
static const void *client_params(void);
static int client_setup(sm_module_t this);
static void client_cleanup(sm_module_t this);

SM_ECM_CLIENT_MODULE = {

	.header = {
		.entry_size = SM_ECM_CLIENT_MODULE_ENTRY_SIZE,
		.api_type = SM_API_ECM_CLIENT,

		.name = "qcg_comp_client",
		.descr = "QCG Comp gSOAP client module",

		.version_major = 0,
		.version_minor = 1,

		.setup = client_setup,
		.cleanup = client_cleanup
	},

	.print_usage = print_usage,
	.opt_handler = opt_handler,
	.args_handler = args_handler,
	.start_client = start_client,
	.params = client_params,
	.opt_string = "csgtGeElioS:L:M:xJ:r:R:D:a:A:X:I:U:T:PC:K:m:",

	.client_api = NULL,

	.have_wsa_header = SM_TRUE,
	.have_wsse_header = SM_TRUE
};

/* WS-Addressing plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsaapi.c>

/* WS-Security plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsseapi.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/smdevp.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/threads.c>

typedef enum {
	OP_CREATE_ACTIVITY,
	OP_GET_ACTIVITY_STATUS,
	OP_GET_ACTIVITY_DOC,
	OP_TERMINATE_ACTIVITY,
	OP_GET_FACTORY_DOC,
	OP_CREATE_RESERVATION,
	OP_GET_RESERVATION_STATUS,
	OP_GET_RESERVATION_DOC,
	OP_TERMINATE_RESERVATION,
	OP_LIST_RESERVATIONS,
	OP_ENABLE_SUB,
	OP_DISABLE_SUB,
	OP_STAGE_IN,
	OP_STAGE_OUT,
	OP_LIST_DIRECTORY,
	OP_STAT,
	OP_DELETE,
	OP_CHMOD,
	OP_MKDIR,
#ifdef WITH_LIBRSYNC
	OP_RSYNC,
	OP_RSYNC_LOCAL_TO_REMOTE,
	OP_RSYNC_REMOTE_TO_LOCAL,
#endif
	OP_SOCKS_GET_CONNECTION_TOKEN
} operation_code_t;

static char *jsdl_str = NULL;
static char *ardl_str = NULL;
static char *activity_id = NULL;
static char *activity_epr_file = NULL;
static char *reservation_id = NULL;
static char *reservation_epr_file = NULL;
static const char *service_url = NULL;
static char operation_opt = 0;
static operation_code_t operation_code;
static sm_alloc_tracked_t tracked = NULL;
static const char *op_namespace = NULL;
static const char *op_name = NULL;
static SM_BOOL extra_info = SM_FALSE;
#ifdef WITH_LIBRSYNC
static const char *rsync_localdir = NULL;
static char *rsync_filter = NULL;
#endif
static const char *path = NULL;
static const char *mode = NULL;

void
print_usage(FILE *f)
{
	fprintf(f, "qcg-comp [common-options] (operation) [operation-args]\n");
	fprintf(f, "\n");
	fprintf(f, "BES-Factory port usage:\n");
	fprintf(f, "  -c -J jsdl.xml [-r reservationUUID|-R reservation-epr.xml] [-A activity-epr.xml]\n");
	fprintf(f, "                                                        create activity, optionally save activity EPR to file\n");
	fprintf(f, "  -s [-x] (-a activityUUID|-A activity-epr.xml)         check status of the activity\n");
	fprintf(f, "  -g [-x] (-a activityUUID|-A activity-epr.xml)         get activity document\n");
	fprintf(f, "  -t (-a activityUUID|-A activity-epr.xml)              terminate activity\n");
	fprintf(f, "  -G                                                    get factory document\n");
	fprintf(f, "\n");
	fprintf(f, "BES-Management port usage:\n");
	fprintf(f, "  -E                                                    enable submission\n");
	fprintf(f, "  -e                                                    disable submission\n");
	fprintf(f, "\n");
	fprintf(f, "QCG-ARES port usage:\n");
	fprintf(f, "  -c -D ardl.xml [-R reservation-epr.xml]\n");
	fprintf(f, "                                                        create reservation, optionally save reservation EPR to file\n");
	fprintf(f, "  -s [-x] (-r reservationUUID|-R reservation-epr.xml)   check status of the reservation\n");
	fprintf(f, "  -g [-x] (-r reservationUUID|-R reservation-epr.xml)   get reservation document\n");
	fprintf(f, "  -t (-r reservationUUID|-R reservation-epr.xml)        terminate reservation\n");
	fprintf(f, "  -l                                                    list active reservations\n");
	fprintf(f, "\n");
	fprintf(f, "  -x                                                    print extra information about activity/reservation\n");
	fprintf(f, "\n");
	fprintf(f, "QCG-Staging port usage:\n");
	fprintf(f, "  -i (-j jsdl|-J jsdl.xml)                              stage files in\n");
	fprintf(f, "  -o (-j jsdl|-J jsdl.xml)                              stage files out\n");
	fprintf(f, "  -I directory                                          list files\n");
	fprintf(f, "  -U file                                               delete file or directory\n");
	fprintf(f, "  -T file                                               stat file\n");
	fprintf(f, "  -C file -m mode (octal representation)                change mode of file\n");
	fprintf(f, "  -K directory [-m mode (octal representation)]         create directory\n");
#ifdef WITH_LIBRSYNC
	fprintf(f, "  -S localDirectory (-j jsdl|-J jsdl.xml) [-X filter]   rsync files\n");
	fprintf(f, "  -L localDirectory (-j jsdl|-J jsdl.xml) [-X filter]   rsync files from local site (updates remote site)\n");
	fprintf(f, "  -M localDirectory (-j jsdl|-J jsdl.xml) [-X filter]   rsync files from remote site (updates local site)\n");
#endif
	fprintf(f, "\n");
	fprintf(f, "QCG-Socks port usage:\n");
	fprintf(f, "  -P                                                    get connection token\n");
}

#define CHECK_OP(opt) \
	if (operation_opt != 0) {\
		sm_error("Ambiguous operations -%c and -%c", operation_opt, opt); \
		return SM_NEG; \
	} else { \
		operation_opt = opt; \
		return SM_OK; \
	}

static char *
file_to_string(const char *file_name)
{
	struct stat sb;
	char *str = NULL;
	FILE *f = NULL;

	if (stat(file_name, &sb) == -1) {
		sm_error("Unable to load file: %s", file_name);
		sm_perror("stat failed ");
		return NULL;
	}

	if (((str = sm_alloc_tracked_malloc(tracked, sb.st_size + 1)) == NULL)) {
		sm_error("unable to allocate %d bytes for file %s", (int)(sb.st_size + 1), file_name);
		return NULL;
	}

	if ((f = fopen(file_name, "r")) == NULL) {
		sm_error("Unable to open file: %s", file_name);
		sm_perror("fopen failed");
	}

	if (fread(str, sb.st_size ,1 ,f) != 1) {
		sm_error("Reading of file %s failed", file_name);
		sm_perror("fread failed");
		fclose(f);
		return NULL;
	}

	str[sb.st_size] = '\0';
	fclose(f);

	return str;
}

static int epr_to_file(struct wsa5__EndpointReferenceType *epr, const char *tag, const char *file_name)
{
	char *epr_str = NULL;
	FILE *f;

	if ((epr_str = sm_gsoap_buffer_put(epr, wsa5__EndpointReferenceType, tag, qcgcomp_namespaces)) == NULL)
		return SM_FAULT;

	if ((f = fopen(file_name, "w")) == NULL) {
		sm_error("unable to open file for writing: %s", file_name);
		sm_perror("fopen failed");
		goto fault;
	}

	if (fwrite(epr_str, strlen(epr_str), 1, f) != 1) {
		sm_perror("fwrite failed");
		fclose(f);
		goto fault;
	}

	fclose(f);
	SM_FREE(epr_str);

	return SM_OK;
fault:
	SM_FREE(epr_str);
	return SM_FAULT;
}

static int print_activity_id(struct wsa5__EndpointReferenceType *ActivityIdentifier)
{
	SM_DEBUG_ENTER();

	if (ActivityIdentifier && ActivityIdentifier->ReferenceParameters) {
		int i;

		for (i = 0; i < ActivityIdentifier->ReferenceParameters->__size; i++) {
			struct soap_dom_element *dom = ActivityIdentifier->ReferenceParameters->__any + i;

			sm_debug("dom->type = %d", dom->type);

			if (dom->type == SOAP_TYPE__qcg_comp_factory__ActivityId) {
				const char *value = (const char *) dom->node;
				size_t ignore = strlen("urn:uuid:");

				if (strncmp(value, "urn:uuid:", ignore))
					return SM_FAULT;

				printf("Activity Id: %s\n", (value + ignore));
				return SM_OK;
			}
		}
	}

	sm_error("No Activity Id found in EPR");
	return SM_FAULT;
}

static int print_reservation_id(struct wsa5__EndpointReferenceType *ReservationIdentifier)
{
	SM_DEBUG_ENTER();

	if (ReservationIdentifier && ReservationIdentifier->ReferenceParameters) {
		int i;

		for (i = 0; i < ReservationIdentifier->ReferenceParameters->__size; i++) {
			struct soap_dom_element *dom = ReservationIdentifier->ReferenceParameters->__any + i;

			sm_debug("dom->type = %d", dom->type);

			if (dom->type == SOAP_TYPE__qcg_comp_ares__ReservationId) {
				const char *value = (const char *) dom->node;
				size_t ignore = strlen("urn:uuid:");

				if (strncmp(value, "urn:uuid:", ignore))
					return SM_FAULT;

				printf("Reservation Id: %s\n", (value + ignore));
				return SM_OK;
			}
		}
	}

	sm_error("No Reservation Id found in EPR");
	return SM_FAULT;
}

static sm_alloc_tracked_t
create_activity_epr(struct wsa5__EndpointReferenceType *epr)
{
	sm_alloc_tracked_t handle = NULL;

	if (activity_epr_file) {
		char *epr_str = file_to_string(activity_epr_file);

		if (!epr_str)
			return NULL;

		handle = sm_gsoap_buffer_get(epr_str, epr, wsa5__EndpointReferenceType, NULL, qcgcomp_namespaces);
		if (handle)
			service_url = epr->Address;

		return handle;
	} else if (activity_id) {
		char buf[1024];
		struct soap_dom_element *dom = NULL;

		snprintf(buf, sizeof(buf), "<ActivityId xmlns=\"" SMC_NAMESPACE_FACTORY "\">urn:uuid:%s</ActivityId>", activity_id);

		dom = sm_alloc_tracked_malloc(tracked, sizeof(struct soap_dom_element));
		handle = sm_gsoap_buffer_get(buf, dom, xsd__anyType, NULL, qcgcomp_namespaces);

		memset(epr, 0, sizeof(*epr));
		epr->ReferenceParameters = sm_alloc_tracked_malloc(tracked, sizeof(struct wsa5__ReferenceParametersType));
		soap_default_wsa5__ReferenceParametersType(NULL, epr->ReferenceParameters);
		epr->ReferenceParameters->__size = 1;
		epr->ReferenceParameters->__any = dom;
		epr->Address = (char*)service_url;

		return handle;
	} else {
		assert(0); /*already checked */
	}

	return NULL;
}

static sm_alloc_tracked_t
create_reservation_epr(struct wsa5__EndpointReferenceType *epr)
{
	sm_alloc_tracked_t handle;

	if (reservation_epr_file) {
		char *epr_str = file_to_string(reservation_epr_file);

		if (!epr_str)
			return NULL;

		handle = sm_gsoap_buffer_get(epr_str, epr, wsa5__EndpointReferenceType, NULL, qcgcomp_namespaces);

		if (handle)
			service_url = epr->Address;

		return handle;
	} else if (reservation_id) {
		char buf[1024];
		struct soap_dom_element *dom = NULL;

		snprintf(buf, sizeof(buf), "<ReservationId xmlns=\"" SMC_NAMESPACE_ARES "\">urn:uuid:%s</ReservationId>", reservation_id);

		dom = sm_alloc_tracked_malloc(tracked, sizeof(struct soap_dom_element));
		handle = sm_gsoap_buffer_get(buf, dom, xsd__anyType, NULL, qcgcomp_namespaces);

		memset(epr, 0, sizeof(*epr));
		epr->ReferenceParameters = sm_alloc_tracked_malloc(tracked, sizeof(struct wsa5__ReferenceParametersType));
		soap_default_wsa5__ReferenceParametersType(NULL, epr->ReferenceParameters);
		epr->ReferenceParameters->__size = 1;
		epr->ReferenceParameters->__any = dom;
		epr->Address = (char*)service_url;

		return handle;
	} else {
		assert(0); /*already checked */
	}

	return NULL;
}

sm_status_ret_code_t
opt_handler(char opt, char *optvalue)
{
	SM_DEBUG_ENTER_INT(opt);

	switch (opt) {
		case 'c' :
		{
			CHECK_OP(opt);
			break;
		}
		case 's' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'g' :
		{
			CHECK_OP(opt);
			break;
		}
		case 't' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'G' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'e' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'E' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'l' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'i' :
		{
			CHECK_OP(opt);
			break;
		}
#ifdef WITH_LIBRSYNC
		case 'S' :
		case 'L' :
		case 'M' :
		{
			rsync_localdir=sm_alloc_tracked_strdup(tracked, optvalue);
			CHECK_OP(opt);
			break;
		}
		case 'X':
		{
			int i, len;

			rsync_filter = sm_alloc_tracked_malloc(tracked, strlen(optvalue)+2);
			strcpy(rsync_filter, optvalue);
			strcat(rsync_filter, ":");
			len = strlen(rsync_filter);

			sm_debug("Filter string: '%s'", rsync_filter);

			for(i=0 ; i<len ; ++i) {
				if(rsync_filter[i]==':')
					rsync_filter[i]='\0';
			}

			break;
		}
#endif
		case 'o' :
		{
			CHECK_OP(opt);
			break;
		}
		case 'J' :
		{
			if ((jsdl_str = file_to_string(optvalue)) == NULL) {
				return SM_FAULT;
			}

			break;
		}
		case 'r' :
		{
			reservation_id = sm_alloc_tracked_strdup(tracked, optvalue);

			break;
		}
		case 'R' :
		{
			reservation_epr_file = sm_alloc_tracked_strdup(tracked, optvalue);

			break;
		}
		case 'D' :
		{
			if ((ardl_str = file_to_string(optvalue)) == NULL) {
				return SM_FAULT;
			}

			break;
		}
		case 'a' :
		{
			activity_id = sm_alloc_tracked_strdup(tracked, optvalue);

			break;
		}
		case 'A' :
		{
			activity_epr_file = sm_alloc_tracked_strdup(tracked, optvalue);

			break;
		}
		case 'x' :
		{
			extra_info = SM_TRUE;

			break;
		}
		case 'I' :
		case 'U' :
		case 'T' :
		case 'C' :
		case 'K' :
		{
			path = sm_alloc_tracked_strdup(tracked, optvalue);
			sm_debug("File operation on path %s", path);
			CHECK_OP(opt);
			break;
		}
		case 'm' :
		{
			mode = sm_alloc_tracked_strdup(tracked, optvalue);
			sm_debug("Using mode %s", mode);
			break;
		}
		case 'P' :
		{
			CHECK_OP(opt);
			break;
		}
		default:
		{
			sm_error("Illegal option: -%c", opt);
			return SM_NEG;
		}
	}

	return SM_OK;
}

sm_status_ret_code_t
args_handler(int argc, char **args sm_unused)
{
	SM_DEBUG_ENTER_INT(argc);

	if (argc > 0) {
		sm_error("Illegal extra arguments");
		return SM_NEG;
	}

	switch (operation_opt) {
		case 'c' :
		{

			if (jsdl_str) {
				operation_code = OP_CREATE_ACTIVITY;
				op_namespace = SMC_NAMESPACE_BES_FACTORY;
				op_name = "CreateActivity";

				if (ardl_str
					|| activity_id
					|| extra_info)
					return SM_NEG;

			} else if (ardl_str) {
				operation_code = OP_CREATE_RESERVATION;
				op_namespace = SMC_NAMESPACE_ARES;
				op_name = "CreateReservation";

				if (activity_id
					|| activity_epr_file
					|| reservation_id
					|| jsdl_str
					|| extra_info)
					return SM_NEG;

			} else {
				return SM_NEG;
			}

			break;
		}
		case 's' :
		{
			if (activity_id && activity_epr_file)
				return SM_NEG;

			if (reservation_id && reservation_epr_file)
				return SM_NEG;

			if (activity_id || activity_epr_file) {
				if ( reservation_id
					|| reservation_epr_file
					|| jsdl_str
					|| ardl_str)
					return SM_NEG;

				operation_code = OP_GET_ACTIVITY_STATUS;
				op_namespace = SMC_NAMESPACE_BES_FACTORY;
				op_name = "GetActivityStatuses";
			} else if (reservation_id || reservation_epr_file) {
				if (activity_id
					|| activity_epr_file
					|| jsdl_str
					|| ardl_str)
					return SM_NEG;

				operation_code = OP_GET_RESERVATION_STATUS;
				op_namespace = SMC_NAMESPACE_ARES;
				op_name = "GetReservationStatus";
			} else {
				return SM_NEG;
			}

			break;
		}
		case 'g' :
		{
			if (activity_id && activity_epr_file)
				return SM_NEG;

			if (reservation_id && reservation_epr_file)
				return SM_NEG;

			if (activity_id || activity_epr_file) {
				if ( reservation_id
					|| reservation_epr_file
					|| jsdl_str
					|| ardl_str)
					return SM_NEG;

				operation_code = OP_GET_ACTIVITY_DOC;
				op_namespace = SMC_NAMESPACE_BES_FACTORY;
				op_name = "GetActivityDocuments";
			} else if (reservation_id || reservation_epr_file) {
				if (activity_id
					|| activity_epr_file
					|| jsdl_str
					|| ardl_str)
					return SM_NEG;

				operation_code = OP_GET_RESERVATION_DOC;
				op_namespace = SMC_NAMESPACE_ARES;
				op_name = "GetReservationDocument";
			} else {
				return SM_NEG;
			}

			break;
		}
		case 't' :
		{
			if (activity_id && activity_epr_file)
				return SM_NEG;

			if (reservation_id && reservation_epr_file)
				return SM_NEG;

			if (activity_id || activity_epr_file) {
				if ( reservation_id
					|| reservation_epr_file
					|| jsdl_str
					|| ardl_str
					|| extra_info)
					return SM_NEG;

				operation_code = OP_TERMINATE_ACTIVITY;
				op_namespace = SMC_NAMESPACE_BES_FACTORY;
				op_name = "TerminateActivites";
			} else if (reservation_id || reservation_epr_file) {
				if (activity_id
					|| activity_epr_file
					|| jsdl_str
					|| ardl_str
					|| extra_info)
					return SM_NEG;

				operation_code = OP_TERMINATE_RESERVATION;
				op_namespace = SMC_NAMESPACE_ARES;
				op_name = "TerminateReservation";
			} else {
				return SM_NEG;
			}

			break;
		}
		case 'G' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_GET_FACTORY_DOC;
			op_namespace = SMC_NAMESPACE_BES_FACTORY;
			op_name = "GetFactoryAttributesDocument";

			break;
		}
		case 'e' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_DISABLE_SUB;
			op_namespace = SMC_NAMESPACE_BES_MANAGEMENT;
			op_name = "StopAcceptingNewActivities";

			break;
		}
		case 'E' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_ENABLE_SUB;
			op_namespace = SMC_NAMESPACE_BES_MANAGEMENT;
			op_name = "StartAcceptingNewActivities";

			break;
		}
		case 'l' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_LIST_RESERVATIONS;
			op_namespace = SMC_NAMESPACE_ARES;
			op_name = "GetActiveReservations";

			break;
		}
		case 'i' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| !jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_STAGE_IN;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "StageIn";
			break;
		}
		case 'o' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| !jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_STAGE_OUT;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "StageOut";
			break;
		}
		case 'I' :
		{
			if (!path)
				return SM_NEG;

			operation_code = OP_LIST_DIRECTORY;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "ListDirectory";
			sm_debug("op_name = %s", op_name);
			break;
		}
		case 'U' :
		{
			if (!path)
				return SM_NEG;

			operation_code = OP_DELETE;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "Delete";
			break;
		}
		case 'T' :
		{
			if (!path)
				return SM_NEG;

			operation_code = OP_STAT;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "StatFile";
			break;
		}
		case 'C' :
		{
			if (!path || !mode)
				return SM_NEG;

			operation_code = OP_CHMOD;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "ChModFile";
			break;
		}
		case 'K' :
		{
			if (!path)
				return SM_NEG;

			operation_code = OP_MKDIR;
			op_namespace = SMC_NAMESPACE_STAGING;
			op_name = "MkDirFile";
			break;
		}
#ifdef WITH_LIBRSYNC
		case 'S' :
		{
			if (activity_id
				|| activity_epr_file
				|| reservation_id
				|| reservation_epr_file
				|| !jsdl_str
				|| ardl_str
				|| extra_info)
				return SM_NEG;

			operation_code = OP_RSYNC;
			op_namespace = SMC_NAMESPACE_RSYNC;
			op_name = "Rsync";
			break;
		}
		case 'L' :
		{
			if (activity_id
					|| activity_epr_file
					|| reservation_id
					|| reservation_epr_file
					|| !jsdl_str
					|| ardl_str
					|| extra_info)
				return SM_NEG;

			operation_code = OP_RSYNC_LOCAL_TO_REMOTE;
			op_namespace = SMC_NAMESPACE_RSYNC;
			op_name = "Rsync";
			break;
		}
		case 'M' :
		{
			if (activity_id
					|| activity_epr_file
					|| reservation_id
					|| reservation_epr_file
					|| !jsdl_str
					|| ardl_str
					|| extra_info)
				return SM_NEG;

			operation_code = OP_RSYNC_REMOTE_TO_LOCAL;
			op_namespace = SMC_NAMESPACE_RSYNC;
			op_name = "Rsync";
			break;
		}
#endif
		case 'P':
		{
			operation_code = OP_SOCKS_GET_CONNECTION_TOKEN;
			op_namespace = SMC_NAMESPACE_SOCKS;
			op_name = "GetConnectionToken";
			break;
		}
		default :
			sm_error("No operation specified");
			return SM_NEG;
	}

	return SM_OK;
}

static int
create_activity(sm_client_transport_t transport)
{
	int ret, i;
	struct bes_factory__CreateActivityType request;
	struct bes_factory__CreateActivityResponseType response;
	struct bes_factory__ActivityDocumentType activity_doc;
	struct jsdl__JobDefinition_USCOREType jsdl;
	sm_alloc_tracked_t handle = NULL;
	sm_alloc_tracked_t handle2 = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));
	memset(&activity_doc, 0, sizeof(activity_doc));
	activity_doc.jsdl__JobDefinition = &jsdl;
	request.bes_factory__ActivityDocument = &activity_doc;

	handle = sm_gsoap_buffer_get(jsdl_str, &jsdl, jsdl__JobDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle) {
		sm_error("Failed to parse JSDL from string");
		goto fault;
	}

	/* remove non-protocol DS elements */
	for ( i = jsdl.jsdl__JobDescription->__sizeDataStaging - 1; i>=0; i--) {
		struct jsdl__DataStaging_USCOREType *ds = jsdl.jsdl__JobDescription->jsdl__DataStaging + i;

		if ((ds->jsdl__Source && ds->jsdl__Source->jsdl__URI && !strchr(ds->jsdl__Source->jsdl__URI, ':'))
			|| (ds->jsdl__Target && ds->jsdl__Target->jsdl__URI && !strchr(ds->jsdl__Target->jsdl__URI, ':'))) {
			if (i != jsdl.jsdl__JobDescription->__sizeDataStaging - 1) {
				ds[i] = ds[jsdl.jsdl__JobDescription->__sizeDataStaging - 1];
			}

			jsdl.jsdl__JobDescription->__sizeDataStaging--;
		}
	}

	if (reservation_epr_file || reservation_id) {

		if (!jsdl.jsdl__JobDescription->jsdl__Resources)
			jsdl.jsdl__JobDescription->jsdl__Resources = sm_alloc_tracked_malloc(tracked, sizeof(*jsdl.jsdl__JobDescription->jsdl__Resources));

		jsdl.jsdl__JobDescription->jsdl__Resources->jsdl_qcg_comp_ares__Reservation =
			sm_alloc_tracked_malloc(tracked, sizeof(*jsdl.jsdl__JobDescription->jsdl__Resources->jsdl_qcg_comp_ares__Reservation));

		if ((handle2 = create_reservation_epr(&jsdl.jsdl__JobDescription->jsdl__Resources->jsdl_qcg_comp_ares__Reservation->ReservationIdentifier)) == NULL)
			goto fault;
	}

	ret = soap_call___bes_factory__CreateActivity(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		if (activity_epr_file) {

			if (epr_to_file(&response.ActivityIdentifier, "bes-factory:ActivityIdentifier", activity_epr_file) != SM_OK)
				goto fault;
		}

		if (print_activity_id(&response.ActivityIdentifier) != SM_OK)
			goto fault;

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	sm_alloc_tracked_free(&handle2);
	return exitcode;
}

static int
create_reservation(sm_client_transport_t transport)
{
	int ret;
	struct qcg_comp_ares__CreateReservationType request;
	struct qcg_comp_ares__CreateReservationResponseType response;
	struct qcg_comp_ares__ARESReservationDocumentType reservation_doc;
	struct ardl__ARDLReservationDefinition_USCOREType ardl;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));
	memset(&reservation_doc, 0, sizeof(reservation_doc));
	reservation_doc.ardl__ReservationDefinition = &ardl;
	request.qcg_comp_ares__ReservationDocument = &reservation_doc;

	handle = sm_gsoap_buffer_get(ardl_str, &ardl, ardl__ARDLReservationDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle) {
		sm_error("Failed to parse ARDL from string");
		goto fault;
	}

	ret = soap_call___qcg_comp_ares__CreateReservation(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		if (reservation_epr_file) {

			if (epr_to_file(&response.ReservationIdentifier, "qcg-comp-ares:ReservationIdentifier", reservation_epr_file) != SM_OK)
				goto fault;

		}

		if (print_reservation_id(&response.ReservationIdentifier) != SM_OK)
			goto fault;

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
get_activity_status(sm_client_transport_t transport)
{
	int ret;
	struct bes_factory__GetActivityStatusesType request;
	struct bes_factory__GetActivityStatusesResponseType response;
	struct wsa5__EndpointReferenceType epr;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = create_activity_epr(&epr);

	if (!handle) {
		goto fault;
	}

	request.__sizeActivityIdentifier = 1;
	request.ActivityIdentifier = &epr;

	ret = soap_call___bes_factory__GetActivityStatuses(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		const char *status = NULL;

		assert(response.__sizeResponse == 1);

		if (response.Response[0].bes_factory__ActivityStatus->__union_ActivityStatusType) {
			switch (response.Response[0].bes_factory__ActivityStatus->__union_ActivityStatusType) {
				case SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Stage_in:
					status = "Stage-in";
					break;
				case SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Stage_out:
					status = "Stage-out";
					break;
				case SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Executing:
					status = "Executing";
					break;
				case SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Queued:
					status = "Queued";
					break;
				case SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Held:
					status = "Held";
					break;
				case SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Suspended:
					status = "Suspended";
					break;
				default:
					assert(0);
			}
		} else {
			status = soap_bes_factory__ActivityStateEnumeration2s(((ecm_gsoap_client_transport_t)transport)->soap, response.Response[0].bes_factory__ActivityStatus->state);
		}

		if (response.Response[0].bes_factory__ActivityStatus->qcg_comp_factory__ActivityEndStatus &&
			response.Response[0].bes_factory__ActivityStatus->qcg_comp_factory__ActivityEndStatus->ExitStatus) {
			printf("status = %s (exit status = %d)\n", status, (int)*(response.Response[0].bes_factory__ActivityStatus->qcg_comp_factory__ActivityEndStatus->ExitStatus));
		} else {
			printf("status = %s\n", status);
		}

		if (response.Response[0].bes_factory__ActivityStatus->__size && extra_info) {
			printf("\nadditional info:\n%s\n", sm_gsoap_buffer_put(response.Response[0].bes_factory__ActivityStatus->__any, xsd__anyType, "", qcgcomp_namespaces));
		}

		if (response.Response[0].__size && extra_info) {
			printf("\n%s\n", sm_gsoap_buffer_put(response.Response[0].__any, xsd__anyType, "", qcgcomp_namespaces));
		}

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
get_activity_document(sm_client_transport_t transport)
{
	int ret;
	struct bes_factory__GetActivityDocumentsType request;
	struct bes_factory__GetActivityDocumentsResponseType response;
	struct wsa5__EndpointReferenceType epr;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = create_activity_epr(&epr);

	if (!handle) {
		goto fault;
	}

	request.__sizeActivityIdentifier = 1;
	request.ActivityIdentifier = &epr;

	ret = soap_call___bes_factory__GetActivityDocuments(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		struct jsdl__JobDefinition_USCOREType *jsdl;
		char *tmp;

		assert(response.__sizeResponse == 1);

		jsdl = response.Response[0].JobDefinition;

		if (!(tmp = sm_gsoap_buffer_put(jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", qcgcomp_namespaces)))
			goto fault;

		printf("%s\n", tmp);
		SM_FREE(tmp);

		if (extra_info) {
			printf("\nLocal job ID: %s\n", response.Response[0].qcg_comp_factory__LocalActivityId);
		}

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
terminate_activity(sm_client_transport_t transport)
{
	int ret;
	struct bes_factory__TerminateActivitiesType request;
	struct bes_factory__TerminateActivitiesResponseType response;
	struct wsa5__EndpointReferenceType epr;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = create_activity_epr(&epr);

	if (!handle) {
		goto fault;
	}

	request.__sizeActivityIdentifier = 1;
	request.ActivityIdentifier = &epr;

	ret = soap_call___bes_factory__TerminateActivities(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {

		assert(response.__sizeResponse == 1);

		if (response.Response[0].Terminated)
			printf("Activity terminated.\n");
		else
			printf("Activity is being terminated.\n");

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
get_factory_document(sm_client_transport_t transport)
{
	int ret;
	struct bes_factory__GetFactoryAttributesDocumentType request;
	struct bes_factory__GetFactoryAttributesDocumentResponseType response;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	ret = soap_call___bes_factory__GetFactoryAttributesDocument(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		struct bes_factory__FactoryResourceAttributesDocumentType *fd = response.bes_factory__FactoryResourceAttributesDocument;
		char *tmp;

		if (!(tmp = sm_gsoap_buffer_put(fd, bes_factory__FactoryResourceAttributesDocumentType, "bes-factory:FactoryResourceAttributesDocument", qcgcomp_namespaces)))
			goto fault;

		printf("%s\n", tmp);
		SM_FREE(tmp);

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	return exitcode;
}

static int
get_reservation_status(sm_client_transport_t transport)
{
	int ret;
	struct qcg_comp_ares__GetReservationStatusType request;
	struct qcg_comp_ares__GetReservationStatusResponseType response;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = create_reservation_epr(&request.ReservationIdentifier);

	if (!handle) {
		goto fault;
	}

	ret = soap_call___qcg_comp_ares__GetReservationStatus(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		int i;

		printf("Reserved hosts:\n");
		for (i = 0 ; i < response.qcg_comp_ares__ReservationStatus->__sizeReservedHost; i++) {
			struct qcg_comp_ares__ARESReservedHost_USCOREType *reservedHost = (response.qcg_comp_ares__ReservationStatus->ReservedHost + i);
			printf("%s[used=%d,reserved=%d,total=%d]\n", reservedHost->HostName, (int)*(reservedHost->UsedSlotsCount), (int)*(reservedHost->ReservedSlotsCount), (int)(*reservedHost->TotalSlotsCount));
		}

		if (response.qcg_comp_ares__ReservationStatus->AssociatedActivities &&
				response.qcg_comp_ares__ReservationStatus->AssociatedActivities->__sizeActivityIdentifier) {
			printf("Associated activities:\n");
			for (i = 0; i < response.qcg_comp_ares__ReservationStatus->AssociatedActivities->__sizeActivityIdentifier; i++) {
				print_activity_id(response.qcg_comp_ares__ReservationStatus->AssociatedActivities->ActivityIdentifier + i);
			}
		}

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
get_reservation_document(sm_client_transport_t transport)
{
	int ret;
	struct qcg_comp_ares__GetReservationDocumentType request;
	struct qcg_comp_ares__GetReservationDocumentResponseType response;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = create_reservation_epr(&request.ReservationIdentifier);

	if (!handle) {
		goto fault;
	}

	ret = soap_call___qcg_comp_ares__GetReservationDocument(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		struct ardl__ARDLReservationDefinition_USCOREType *ardl = response.qcg_comp_ares__ReservationDocument->ardl__ReservationDefinition;
		char *tmp;

		if (!(tmp = sm_gsoap_buffer_put(ardl, ardl__ARDLReservationDefinition_USCOREType, "ardl:ReservationDefinition", qcgcomp_namespaces)))
			goto fault;

		printf("%s\n", tmp);
		SM_FREE(tmp);

		if (extra_info) {
			printf("\nLocal reservation ID: %s\n", response.qcg_comp_ares__LocalReservationId);
		}

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
terminate_reservation(sm_client_transport_t transport)
{
	int ret;
	struct qcg_comp_ares__TerminateReservationType request;
	struct qcg_comp_ares__TerminateReservationResponseType response;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = create_reservation_epr(&request.ReservationIdentifier);

	if (!handle) {
		goto fault;
	}

	ret = soap_call___qcg_comp_ares__TerminateReservation(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		printf("Reservation terminated.\n");
		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
list_reservations(sm_client_transport_t transport)
{
	int ret;
	struct qcg_comp_ares__GetActiveReservationsType request;
	struct qcg_comp_ares__GetActiveReservationsResponseType response;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	ret = soap_call___qcg_comp_ares__GetActiveReservations(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		int i;

		for (i=0; i < response.__sizeReservationIdentifier; i++)
			print_reservation_id(response.ReservationIdentifier + i);

		printf("Total number of reservations: %d\n", response.__sizeReservationIdentifier);

		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	return exitcode;
}

static int
disable_sub(sm_client_transport_t transport)
{
	int ret;
	struct bes_management__StopAcceptingNewActivitiesType request;
	struct bes_management__StopAcceptingNewActivitiesResponseType response;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	ret = soap_call___bes_management__StopAcceptingNewActivities(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		printf("Submission disabled.\n");
		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	return exitcode;
}

static int
enable_sub(sm_client_transport_t transport)
{
	int ret;
	struct bes_management__StartAcceptingNewActivitiesType request;
	struct bes_management__StartAcceptingNewActivitiesResponseType response;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	ret = soap_call___bes_management__StartAcceptingNewActivities(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		printf("Submission enabled.\n");
		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	return exitcode;
}

static int
add_mime_attachment(struct soap *soap, const char *filename, const char *id)
{
	int fd;
	void *ptr = NULL;
	void *ptr2 = NULL;
	struct stat stat_buf;
	int retval = SM_FAULT;

	SM_DEBUG_ENTER_STR(filename);

	if (-1 == (fd = open(filename, O_RDONLY))) {
		sm_error("Unable to open file: %s", filename);
		sm_perror("open");
		return SM_FAULT;
	}

	if (-1 == fstat(fd, &stat_buf)) {
		sm_perror("fstat");
		close(fd);
		return SM_FAULT;
	}

	if (0 == stat_buf.st_size) {
		ptr2 = soap_malloc(soap, 1);
		if (!ptr2) {
			close(fd);
			return SM_FAULT;
		}

		*((char *)ptr2) = 0;
	} else {
		ptr = mmap(NULL, stat_buf.st_size, PROT_READ, MAP_SHARED, fd, 0);
		if (MAP_FAILED == ptr) {
			sm_perror("mmap");
			close(fd);
			return SM_FAULT;
		}

		ptr2 = soap_malloc(soap, stat_buf.st_size + 1);
		if (!ptr2) {
			munmap(ptr, stat_buf.st_size);
			close(fd);
			return SM_FAULT;
		}

		((char *)ptr2)[stat_buf.st_size] = 0;
		memcpy(ptr2, ptr, stat_buf.st_size);
	}

	if (!(soap->omode & SOAP_ENC_MIME)) {
		soap_set_mime(soap, NULL, NULL);
	}

	if (SOAP_OK ==
		soap_set_mime_attachment(soap, ptr2, stat_buf.st_size, SOAP_MIME_BINARY, "application/octet-stream", sm_alloc_tracked_asprintf(tracked, "<%s>", id), NULL, NULL)) {
		retval = SM_OK;
	} else {
		sm_error("soap_set_mime_attachment failed");
	}

	munmap(ptr, stat_buf.st_size);
	close(fd);
	return retval;
}

static char *
generate_id(void)
{
	uuid_t uuid;
	char buf[SM_UUID_SIZE];

	uuid_generate(uuid);
	uuid_unparse(uuid, buf);

	return sm_alloc_tracked_strdup(tracked, buf);
}

static int
stage_in(sm_client_transport_t transport)
{
	int ret, i;
	struct qcg_comp_staging__StageFilesInType request;
	struct qcg_comp_staging__StageFilesInResponseType response;
	struct jsdl__JobDefinition_USCOREType jsdl;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = sm_gsoap_buffer_get(jsdl_str, &jsdl, jsdl__JobDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle) {
		sm_error("Failed to parse JSDL from string");
		goto fault;
	}

	if (!jsdl.jsdl__JobDescription || !jsdl.jsdl__JobDescription->__sizeDataStaging) {
		printf("No data staging elements found.\n");
		goto ret;
	}

	for ( i = 0; i < jsdl.jsdl__JobDescription->__sizeDataStaging; i++) {
		struct jsdl__DataStaging_USCOREType *ds = jsdl.jsdl__JobDescription->jsdl__DataStaging + i;
		if (ds->jsdl__Source && ds->jsdl__Source->jsdl__URI && !strchr(ds->jsdl__Source->jsdl__URI, ':')) {
			const char *file_name = ds->jsdl__Source->jsdl__URI;

			ds->jsdl__Source->jsdl__URI = sm_alloc_tracked_asprintf(tracked, "cid:%s",generate_id());

			if (add_mime_attachment(((ecm_gsoap_client_transport_t)transport)->soap, file_name, ds->jsdl__Source->jsdl__URI) == SM_FAULT)
				goto fault;

			printf("Staging file: %s\n", file_name);
		}
	}

	request.jsdl__JobDefinition = &jsdl;

	ret = soap_call___qcg_comp_staging__StageFilesIn(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		printf("All files staged in.\n");
		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

static int
save_all_files(struct soap_multipart *attachment, struct jsdl__JobDefinition_USCOREType *jsdl)
{
	FILE *handle;
	size_t len, nwritten;
	char *buf;
	char *id2;
	const char *file_name;
	int i;

	SM_DEBUG_ENTER_PTR(attachment);

	for ( ; attachment; attachment = attachment->next) {

		sm_debug("attachment->id = %s", SM_SAFE_STR(attachment->id));

		if ( !attachment->id || !strcmp(attachment->id, "<SOAP-ENV:Envelope>"))
			continue;

		id2 = (char*)attachment->id;

		if (!strstr(id2, "<cid:")) {
			sm_error("Illegal attachment id format: %s", id2);
			return SM_FAULT;
		}

		id2+=1;
		id2[strlen(id2)-1]='\0';

		file_name = NULL;
		for ( i = 0; i < jsdl->jsdl__JobDescription->__sizeDataStaging; i++) {
			struct jsdl__DataStaging_USCOREType *ds = jsdl->jsdl__JobDescription->jsdl__DataStaging + i;
			if (ds->jsdl__Target && ds->jsdl__Target->jsdl__URI && !strcmp(ds->jsdl__Target->jsdl__URI, id2)) {
				file_name = (const char *)ds->__any;
				break;
			}
		}

		if (!file_name) {
			sm_error("Unknown attachment id: %s", id2);
			return SM_FAULT;
		}

		if (!(handle = fopen(file_name, "w+"))) {
			sm_error("Unable to open file %s for writing", file_name);
			sm_perror("fopen");
			return SM_FAULT;
		}

		len = attachment->size;
		buf = attachment->ptr;

		while (len) {
			nwritten = fwrite(buf, 1, len, handle);
			if (!nwritten) {
				fclose(handle);
				return SM_FAULT;
			}
			len -= nwritten;
			buf += nwritten;
		}

		fsync(fileno(handle));

		if (fclose(handle) != 0)
			return SM_FAULT;

		printf("File %s staged out.\n", file_name);
	}

	return SM_OK;
}

static int
stage_out(sm_client_transport_t transport)
{
	int ret, i;
	struct qcg_comp_staging__StageFilesOutType request;
	struct qcg_comp_staging__StageFilesOutResponseType response;
	struct jsdl__JobDefinition_USCOREType jsdl;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;

	memset(&request,0, sizeof(request));

	handle = sm_gsoap_buffer_get(jsdl_str, &jsdl, jsdl__JobDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle) {
		sm_error("Failed to parse JSDL from string");
		goto fault;
	}

	if (!jsdl.jsdl__JobDescription || !jsdl.jsdl__JobDescription->__sizeDataStaging) {
		printf("No data staging elements found.\n");
		goto ret;
	}

	for ( i = 0; i < jsdl.jsdl__JobDescription->__sizeDataStaging; i++) {
		struct jsdl__DataStaging_USCOREType *ds = jsdl.jsdl__JobDescription->jsdl__DataStaging + i;
		if (ds->jsdl__Target && ds->jsdl__Target->jsdl__URI && !strchr(ds->jsdl__Target->jsdl__URI, ':')) {
			const char *file_name = ds->jsdl__Target->jsdl__URI;
			ds->jsdl__Target->jsdl__URI = sm_alloc_tracked_asprintf(tracked, "cid:%s", generate_id());
			if (ds->__size) {
				sm_error("Unsupported extension in DataStaging element");
				goto fault;
			}
			ds->__any = (void*)file_name;
		}
	}

	request.jsdl__JobDefinition = &jsdl;

	ret = soap_call___qcg_comp_staging__StageFilesOut(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		if (save_all_files(((ecm_gsoap_client_transport_t)transport)->soap->mime.list,&jsdl) == SM_FAULT)
			goto fault;
		printf("All files staged out.\n");
		goto ret;
	} else {
		goto fault;
	}

fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}

#ifdef WITH_LIBRSYNC
static int
rsync(sm_client_transport_t transport)
{
	struct jsdl__JobDefinition_USCOREType jsdl;
	sm_alloc_tracked_t handle = NULL;
	int exitcode = 0;
	struct jsdl_hpcpa__HPCProfileApplication_USCOREType *hpcprof = NULL;

	handle = sm_gsoap_buffer_get(jsdl_str, &jsdl, jsdl__JobDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle) {
		sm_error("Failed to parse JSDL from string");
		goto fault;
	}

	if(!jsdl.jsdl__JobDescription  || !jsdl.jsdl__JobDescription->jsdl__Application)
	{
		sm_error("Missing mandatory element: jsdl:Application");
		goto fault;
	}
	hpcprof=jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;
	if(!hpcprof)
	{
		sm_error("Missing mandatory element: jsdl-hpcpa:HPCProfileApplication");
		goto fault;
	}
	const char *username=NULL;
	if(hpcprof->jsdl_hpcpa__UserName)
		username = hpcprof->jsdl_hpcpa__UserName->__item;

	const char *remote = NULL;
	if (!hpcprof->jsdl_hpcpa__WorkingDirectory)
	{
		sm_error("Missing mandatory element: jsdl-hpcpa:WorkingDirectory");
		goto fault;
	}
	remote = hpcprof->jsdl_hpcpa__WorkingDirectory->__item;
	printf("Syncing local %s and %s:%s\n", rsync_localdir, service_url, remote);
	switch(operation_code)
	{
	case OP_RSYNC:
		rsync_sync(transport, service_url, username, rsync_localdir, remote, rsync_filter);
		break;
	case OP_RSYNC_LOCAL_TO_REMOTE:
		printf("The local side is the origin.\n");
		rsync_local_to_remote(transport, service_url, username, rsync_localdir, remote, rsync_filter);
		break;
	case OP_RSYNC_REMOTE_TO_LOCAL:
		printf("The remote side is the origin.\n");
		rsync_remote_to_local(transport, service_url, username, rsync_localdir, remote, rsync_filter);
		break;
	default:
		assert(0);
	}
	goto ret;
fault:
	exitcode = 1;
ret:
	sm_alloc_tracked_free(&handle);
	return exitcode;
}
#endif

static int
list_directory(sm_client_transport_t transport)
{
	struct qcg_comp_staging__ListDirectoryType request;
	struct qcg_comp_staging__ListDirectoryResponseType response;
	int i;
	struct jsdl__DataStaging_USCOREType *ds;
	int ret;

	memset(&request,0,sizeof(request));

	request.Directory = transport->transport_strdup(transport, path);
	request.UserName = NULL;
	

	ret = soap_call___qcg_comp_staging__ListDirectory(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		goto fault;
	}

	ds = response.jsdl__JobDefinition->jsdl__JobDescription->jsdl__DataStaging;

	printf("File list in directory %s:\n", path);
	for(i=0 ; i<response.jsdl__JobDefinition->jsdl__JobDescription->__sizeDataStaging ; ++i) {
		printf("%s\n", ds[i].jsdl__FileName);
	}

	return 0;
fault:
	return 1;
}

static int
stat_file(sm_client_transport_t transport)
{
	struct qcg_comp_staging__StatFileType request;
	struct qcg_comp_staging__StatFileResponseType response;
	int ret;

	memset(&request,0,sizeof(request));

	request.__sizePath = 1;
	SM_TRY_NOT(request.Path = transport->transport_malloc(transport, sizeof(*request.Path)));
	SM_TRY_NOT(request.Path[0] = transport->transport_strdup(transport, path));
	request.UserName = NULL;

	printf("Stats for %s:\n", request.Path[0]);

	ret = soap_call___qcg_comp_staging__StatFile(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		goto fault;
	}

	printf("Access time: %s", ctime(&response.Files[0].ATime));
	printf("Creation time: %s", ctime(&response.Files[0].CTime));
	printf("Modification time: %s", ctime(&response.Files[0].MTime));
	printf("Owner: %s:%s\n", response.Files[0].UID, response.Files[0].GID);
	printf("Mode: %o\n", response.Files[0].Mode);
	printf("Number of hard links: %i\n", response.Files[0].NLink);
	printf("Size: %i\n", response.Files[0].Size);

	return 0;
fault:
	return 1;
}

static int
delete(sm_client_transport_t transport)
{
	struct qcg_comp_staging__DeleteType request;
	struct qcg_comp_staging__DeleteResponseType response;
	int ret;

	memset(&request,0,sizeof(request));

	request.Path = transport->transport_strdup(transport, path);
	request.UserName = NULL;

	printf("Deleting %s\n", request.Path);

	ret = soap_call___qcg_comp_staging__Delete(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		goto fault;
	}

	printf("Done\n");

	return 0;
fault:
	return 1;
}

static int
op_chmod(sm_client_transport_t transport)
{
	struct qcg_comp_staging__ChModType request;
	struct qcg_comp_staging__ChModResponseType response;
	int ret;

	memset(&request,0,sizeof(request));

	request.Path = transport->transport_strdup(transport, path);
	request.Mode = transport->transport_strdup(transport, mode);
	request.UserName = NULL;

	printf("Changing mode of %s to %s\n", request.Path, request.Mode);

	ret = soap_call___qcg_comp_staging__ChMod(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		goto fault;
	}

	printf("Done\n");

	return 0;
fault:
	return 1;
}

static int
op_mkdir(sm_client_transport_t transport)
{
	struct qcg_comp_staging__MkDirType request;
	struct qcg_comp_staging__MkDirResponseType response;
	int ret;

	memset(&request,0,sizeof(request));

	request.Path = transport->transport_strdup(transport, path);

	if(mode != NULL)
		request.Mode = transport->transport_strdup(transport, mode);
	else
		request.Mode = NULL;

	request.UserName = NULL;

	printf("Creating directory %s", request.Path);

	if(request.Mode != NULL)
		printf("(with mode %s)\n", request.Mode);
	else
		printf("\n");

	ret = soap_call___qcg_comp_staging__MkDir(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		goto fault;
	}

	printf("Done\n");

	return 0;
fault:
	return 1;
}

static int
get_connection_token(sm_client_transport_t transport)
{
	struct qcg_comp_socks__GetConnectionTokenType request;
	struct qcg_comp_socks__GetConnectionTokenResponseType response;
	int ret;

	memset(&request, 0, sizeof(request));

	ret = soap_call___qcg_comp_socks__GetConnectionToken(((ecm_gsoap_client_transport_t)transport)->soap, service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		goto fault;
	}

	printf("Connection token: %s\n", response.SecurityToken);

	return 0;
fault:
	return 1;
}

int
start_client(sm_client_transport_t transport)
{

	if (transport->pre_call_setup(transport, op_namespace, op_name, NULL, NULL, NULL) != SM_OK) {
		return 1;
	}

	service_url = transport->service_url;

	switch (operation_code) {
		case OP_CREATE_ACTIVITY :
			return create_activity(transport);
		case OP_GET_ACTIVITY_STATUS :
			return get_activity_status(transport);
		case OP_GET_ACTIVITY_DOC :
			return get_activity_document(transport);
		case OP_TERMINATE_ACTIVITY :
			return terminate_activity(transport);
		case OP_CREATE_RESERVATION :
			return create_reservation(transport);
		case OP_GET_RESERVATION_STATUS :
			return get_reservation_status(transport);
		case OP_GET_RESERVATION_DOC :
			return get_reservation_document(transport);
		case OP_TERMINATE_RESERVATION :
			return terminate_reservation(transport);
		case OP_GET_FACTORY_DOC :
			return get_factory_document(transport);
		case OP_LIST_RESERVATIONS :
			return list_reservations(transport);
		case OP_DISABLE_SUB :
			return disable_sub(transport);
		case OP_ENABLE_SUB :
			return enable_sub(transport);
		case OP_STAGE_IN :
			return stage_in(transport);
		case OP_STAGE_OUT :
			return stage_out(transport);
		case OP_LIST_DIRECTORY:
			return list_directory(transport);
		case OP_STAT:
			return stat_file(transport);
		case OP_DELETE:
			return delete(transport);
		case OP_CHMOD:
			return op_chmod(transport);
		case OP_MKDIR:
			return op_mkdir(transport);
#ifdef WITH_LIBRSYNC
		case OP_RSYNC:
		case OP_RSYNC_LOCAL_TO_REMOTE:
		case OP_RSYNC_REMOTE_TO_LOCAL:
			return rsync(transport);
#endif
		case OP_SOCKS_GET_CONNECTION_TOKEN:
			return get_connection_token(transport);
		default:
			assert(0);
	}

}

void
print_fault(sm_client_transport_t transport)
{
	char *fault_detail;
	struct soap *soap = ((ecm_gsoap_client_transport_t)transport)->soap;

	soap_print_fault(soap, stderr);

	if (soap->fault && soap->fault->detail) {
		fault_detail = sm_gsoap_buffer_put(soap->fault->detail, SOAP_ENV__Detail, "SOAP-ENV:Detail", qcgcomp_namespaces);

		if (fault_detail) {
			char *b = strchr(fault_detail, '>');
			char *e = strrchr(fault_detail, '<');

			if (b && e) {
				*e = '\0';
				fprintf(stderr,"%s", b + 1);
			} else {
				fprintf(stderr,"%s", fault_detail);
			}
			fprintf(stderr, "\n");

			SM_FREE(fault_detail);
		}
	}
}

const void *
client_params(void)
{
	static struct ecm_gsoap_client_transport_params_s params = {
		.namespaces = qcgcomp_namespaces,
		.print_fault = print_fault
	};

	return &params;
}

int
client_setup(sm_module_t this sm_unused)
{
	if (!(tracked = sm_alloc_tracked_create()))
		return SM_FAULT;

	return SM_OK;
}

void
client_cleanup(sm_module_t this sm_unused)
{
	sm_alloc_tracked_free(&tracked);
}
