/* $Id: icomm.c 294 2010-08-30 08:51:28Z smaug $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>

#include <qcg-core/sm.h>
#include <qcg-core/conf.h>
#include <qcg-core/logger.h>
#include <qcg-core/privsep.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <icomm.h>
#include <stub/qcgcompH.h>
#include <stub/qcgcompStub.h>

__SM_ID("$Id: icomm.c 294 2010-08-30 08:51:28Z smaug $");

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "icomm"
        }
};

static sm_value_t *session_process_data;

sm_status_ret_code_t
smc_icomm_setup(sm_module_t this)
{
	char *service_config = NULL;
	const char *service_config_namespace;
	const char *service_config_xpath;
	const char *log_file;
	int log_level;

	SM_DEBUG_ENTER();

	service_config_namespace = sm_conf_ctxt_get_service_namespace(this->conf_ctxt);
	service_config_xpath = sm_conf_ctxt_get_service_prefix(this->conf_ctxt);

	if (!service_config_namespace || !service_config_xpath)
		return SM_FAULT;

	/* We keep the configuration params instead of filename because
 	 * session process might not have appropriate permission to read the file. */
	if (sm_conf_node_raw(this->conf_ctxt, SM_TRUE, "/", &service_config) != SM_OK)
		return SM_FAULT;

	log_file = sm_log_get_filename();
	log_level = sm_log_get_log_level();

	session_process_data = sm_value_create(3,
		SM_VALUE_STRING, service_config, SM_VALUE_STRING, service_config_namespace, SM_VALUE_STRING, service_config_xpath);

	SM_FREE(service_config);

	if (!session_process_data)
		return SM_FAULT;

	return SM_OK;
}

void
smc_icomm_cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();

	sm_value_free(&session_process_data);
}

static sm_status_ret_code_t
smc_icomm_send_value(int fd, sm_value_t *params, sm_value_t **result, smc_icomm_request_t type)
{
	sm_value_t *value = NULL;

	SM_MANDATORY_P(params);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	SM_VALUE_DEBUG(params);

	assert(fd != -1);

	value = sm_value_create(1, SM_VALUE_INT32, type);
	if (!value)
		goto fault;

	if (sm_value_to_fd(fd, value) == SM_FAULT)
		goto fault;

	if (sm_value_to_fd(fd, params) == SM_FAULT)
		goto fault;

	if (!(*result = sm_value_from_fd(fd)))
		goto fault;

	sm_value_free(&value);
	return SM_OK;

fault:
	sm_value_free(&value);
	return SM_FAULT;
}

static sm_status_ret_code_t
smc_icomm_create_process_and_send(uid_t uid, SM_BOOL one_shot, sm_value_t *params, sm_value_t **result, smc_icomm_request_t type)
{
	sm_status_ret_code_t ret;
	sm_privsep_process_handle_t handle = NULL;
	int fd = -1;

	SM_DEBUG_ENTER();

	if (sm_privsep_get_process(uid, one_shot, session_process_data, &fd, &handle) != SM_OK) {
		sm_error("Failed to create session process for uid: %u (request: %d)", (unsigned int) uid, type);
		return SM_FAULT;
	}

	if ((ret = smc_icomm_send_value(fd, params, result, type)) != SM_OK) {
		sm_error("Failed to send request to session process for uid: %u (request: %d)", (unsigned int) uid, type);
		/* Continue and try to release the process. */
	}

	if (sm_privsep_release_process(&fd, &handle) != SM_OK)
		sm_error("Failed to release session process for uid: %u (request: %d)", (unsigned int) uid, type);

	return ret;
}

sm_status_ret_code_t
smc_icomm_request_CreateActivity(uid_t uid, sm_value_t *params, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_FALSE, params, result, smc_icomm_CreateActivity);
}

sm_status_ret_code_t
smc_icomm_request_GetActivityStatuses(uid_t uid, sm_value_t *params, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_FALSE, params, result, smc_icomm_GetActivityStatuses);
}

sm_status_ret_code_t
smc_icomm_request_TerminateActivities(uid_t uid, sm_value_t *params, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_FALSE, params, result, smc_icomm_TerminateActivities);
}

sm_status_ret_code_t
smc_icomm_request_CreateReservation(uid_t uid, sm_value_t *params, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_TRUE, params, result, smc_icomm_CreateReservation);
}

sm_status_ret_code_t
smc_icomm_request_GetReservationStatus(uid_t uid, sm_value_t *params, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_TRUE, params, result, smc_icomm_GetReservationStatus);
}

sm_status_ret_code_t
smc_icomm_request_GetActiveReservations(uid_t uid, sm_value_t *empty, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_TRUE, empty, result, smc_icomm_GetActiveReservations);
}

sm_status_ret_code_t
smc_icomm_request_TerminateReservation(uid_t uid, sm_value_t *params, sm_value_t **result)
{
	SM_DEBUG_ENTER();
	return smc_icomm_create_process_and_send(uid, SM_TRUE, params, result, smc_icomm_TerminateReservation);
}

sm_status_ret_code_t
smc_icomm_request_StageInCreateContext(uid_t uid, sm_value_t *params, sm_value_t **result, int *context, void **handle)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if (sm_privsep_get_process(uid, SM_TRUE, session_process_data, context, (sm_privsep_process_handle_t *)handle) != SM_OK) {
                sm_error("Failed to create session process for uid: %u (request: smc_icomm_StageInCreateContext)", (unsigned int) uid);
                return SM_FAULT;
        }

        if ((ret = smc_icomm_send_value(*context, params, result, smc_icomm_StageInCreateContext)) != SM_OK) {
                sm_error("Failed to send request to session process for uid: %u (request: smc_icomm_StageInCreateContext)", (unsigned int) uid);
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageInOpen(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_StageInOpen)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageInOpen");
                /* Continue and try to release the process. */
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageInWrite(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_StageInWrite)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageInWrite");
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageInClose(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_StageInClose)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageInClose");
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageInDestroyContext(int *context, void **handle, sm_value_t *empty, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(*context, empty, result, smc_icomm_StageInDestroyContext)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageInDestroyContext");
                /* Continue and try to release the process. */
        }

        if (sm_privsep_release_process(context, (sm_privsep_process_handle_t *)handle) != SM_OK)
                sm_error("Failed to release session process for request: smc_icomm_StageInDestroyContext");

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageInRelease(int *context, void **handle)
{
        sm_status_ret_code_t ret;

        if ((ret = sm_privsep_release_process(context, (sm_privsep_process_handle_t *)handle)) != SM_OK)
                sm_error("Failed to release session process for request: smc_icomm_StageInRelease");

	return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageOutCreateContext(uid_t uid, sm_value_t *params, sm_value_t **result, int *context, void **handle)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if (sm_privsep_get_process(uid, SM_TRUE, session_process_data, context, (sm_privsep_process_handle_t *)handle) != SM_OK) {
                sm_error("Failed to create session process for uid: %u (request: smc_icomm_StageOutCreateContext)", (unsigned int) uid);
                return SM_FAULT;
        }

        if ((ret = smc_icomm_send_value(*context, params, result, smc_icomm_StageOutCreateContext)) != SM_OK) {
                sm_error("Failed to send request to session process for uid: %u (request: smc_icomm_StageOutCreateContext)", (unsigned int) uid);
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageOutOpen(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_StageOutOpen)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageOutOpen");
                /* Continue and try to release the process. */
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageOutRead(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_StageOutRead)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageOutRead");
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageOutClose(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_StageOutClose)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageOutClose");
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageOutDestroyContext(int *context, void **handle, sm_value_t *empty, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(*context, empty, result, smc_icomm_StageOutDestroyContext)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_StageOutDestroyContext");
                /* Continue and try to release the process. */
        }

        if (sm_privsep_release_process(context, (sm_privsep_process_handle_t *)handle) != SM_OK)
                sm_error("Failed to release session process for request: smc_icomm_StageOutDestroyContext");

        return ret;
}

sm_status_ret_code_t
smc_icomm_request_StageOutRelease(int *context, void **handle)
{
        sm_status_ret_code_t ret;

        if ((ret = sm_privsep_release_process(context, (sm_privsep_process_handle_t *)handle)) != SM_OK)
                sm_error("Failed to release session process for request: smc_icomm_StageOutRelease");

	return ret;
}

sm_status_ret_code_t
smc_icomm_begin_receive(int fd, smc_icomm_request_t *result)
{
	sm_value_t *value = NULL;

	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	assert(fd != -1);

	value = sm_value_from_fd(fd);
	if (!value)
		return SM_FAULT;

	assert(sm_value_count(value) == 1);
	assert(SM_VALUE_IS_INT(value[0]));

	*result = value[0]->vint32;

	sm_value_free(&value);
	return SM_OK;
}

static sm_status_ret_code_t
smc_icomm_receive_value(int fd, sm_value_t **value)
{
	SM_MANDATORY_P(value);

	SM_DEBUG_ENTER();

	assert(fd != -1);

	*value = sm_value_from_fd(fd);

	if (*value)
		return SM_OK;
	else
		return SM_FAULT;
}

sm_status_ret_code_t
smc_icomm_receive_CreateActivity(int fd, sm_value_t **jsdl)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, jsdl);
}

sm_status_ret_code_t
smc_icomm_receive_GetActivityStatuses(int fd, sm_value_t **params)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_TerminateActivities(int fd, sm_value_t **params)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_CreateReservation(int fd, sm_value_t **params)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_GetReservationStatus(int fd, sm_value_t **params)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_GetActiveReservations(int fd, sm_value_t **empty)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, empty);
}

sm_status_ret_code_t
smc_icomm_receive_TerminateReservation(int fd, sm_value_t **params)
{
	SM_DEBUG_ENTER();
	return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageInCreateContext(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageInOpen(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageInWrite(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageInClose(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageInDestroyContext(int fd, sm_value_t **empty)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, empty);
}

sm_status_ret_code_t
smc_icomm_receive_StageOutCreateContext(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageOutOpen(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageOutRead(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageOutClose(int fd, sm_value_t **params)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, params);
}

sm_status_ret_code_t
smc_icomm_receive_StageOutDestroyContext(int fd, sm_value_t **empty)
{
        SM_DEBUG_ENTER();
        return smc_icomm_receive_value(fd, empty);
}

sm_status_ret_code_t
smc_icomm_respond_CreateActivity(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_GetActivityStatuses(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_TerminateActivities(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_CreateReservation(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_GetReservationStatus(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_GetActiveReservations(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_TerminateReservation(int fd, sm_value_t *result)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (!result)
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;

	ret = sm_value_to_fd(fd, result);
	sm_value_free(&result);

	return ret;
}

sm_status_ret_code_t
smc_icomm_respond_StageIn(int fd, sm_value_t *result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if (!result)
                result = SM_VALUE_RET_CODE_OOM_SINGLETON;

        ret = sm_value_to_fd(fd, result);
        sm_value_free(&result);

        return ret;
}

sm_status_ret_code_t
smc_icomm_respond_StageOut(int fd, sm_value_t *result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if (!result)
                result = SM_VALUE_RET_CODE_OOM_SINGLETON;

        ret = sm_value_to_fd(fd, result);
        sm_value_free(&result);

        return ret;
}

#define ICOMM_REQUEST(name) sm_status_ret_code_t \
		smc_icomm_request_##name(uid_t uid, sm_value_t *params, sm_value_t **result)\
		{\
	SM_DEBUG_ENTER();\
return smc_icomm_create_process_and_send(uid, SM_FALSE, params, result, smc_icomm_##name);\
		}\

#define ICOMM_RECEIVE(name)\
	sm_status_ret_code_t\
		smc_icomm_receive_##name(int fd, sm_value_t **params)\
		{\
			SM_DEBUG_ENTER();\
			return smc_icomm_receive_value(fd, params);\
		}
#define ICOMM_RESPOND(name)\
sm_status_ret_code_t smc_icomm_respond_##name(int fd, sm_value_t *result) {\
	sm_status_ret_code_t ret;\
\
	SM_DEBUG_ENTER();\
\
	if (!result)\
		result = SM_VALUE_RET_CODE_OOM_SINGLETON;\
\
	ret = sm_value_to_fd(fd, result);\
	sm_value_free(&result);\
\
	return ret;\
}

#define ICOMM_DEFINE(name) \
	ICOMM_REQUEST(name) \
	ICOMM_RECEIVE(name) \
	ICOMM_RESPOND(name)

ICOMM_DEFINE(GetFiles)
ICOMM_DEFINE(GetSignature)

ICOMM_RECEIVE(GetDeltaBegin)
ICOMM_RESPOND(GetDeltaBegin)
sm_status_ret_code_t
smc_icomm_request_GetDeltaBegin(uid_t uid, sm_value_t *params, sm_value_t **result, int *context, void **handle)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if (sm_privsep_get_process(uid, SM_TRUE, session_process_data, context, (sm_privsep_process_handle_t *)handle) != SM_OK) {
                sm_error("Failed to create session process for uid: %u (request: smc_icomm_GetDeltaBegin)", (unsigned int) uid);
                return SM_FAULT;
        }

        if ((ret = smc_icomm_send_value(*context, params, result, smc_icomm_GetDeltaBegin)) != SM_OK) {
                sm_error("Failed to send request to session process for uid: %u (request: smc_icomm_GetDeltaBegin)", (unsigned int) uid);
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_GetDeltaRelease(int *context, void **handle)
{
	sm_status_ret_code_t ret;

	if ((ret = sm_privsep_release_process(context,
			(sm_privsep_process_handle_t *) handle)) != SM_OK)
		sm_error("Failed to release session process");

	return ret;
}

ICOMM_RECEIVE(GetDeltaProcess)
ICOMM_RESPOND(GetDeltaProcess)
sm_status_ret_code_t
smc_icomm_request_GetDeltaProcess(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_GetDeltaProcess)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_GetDeltaProcess");
                /* Continue and try to release the process. */
        }

        return ret;
}

ICOMM_RECEIVE(PutDeltaBegin)
ICOMM_RESPOND(PutDeltaBegin)
sm_status_ret_code_t
smc_icomm_request_PutDeltaBegin(uid_t uid, sm_value_t *params, sm_value_t **result, int *context, void **handle)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if (sm_privsep_get_process(uid, SM_TRUE, session_process_data, context, (sm_privsep_process_handle_t *)handle) != SM_OK) {
                sm_error("Failed to create session process for uid: %u (request: smc_icomm_PutDeltaBegin)", (unsigned int) uid);
                return SM_FAULT;
        }

        if ((ret = smc_icomm_send_value(*context, params, result, smc_icomm_PutDeltaBegin)) != SM_OK) {
                sm_error("Failed to send request to session process for uid: %u (request: smc_icomm_PutDeltaBegin)", (unsigned int) uid);
        }

        return ret;
}

sm_status_ret_code_t
smc_icomm_PutDeltaRelease(int *context, void **handle)
{
	sm_status_ret_code_t ret;

	if ((ret = sm_privsep_release_process(context,
			(sm_privsep_process_handle_t *) handle)) != SM_OK)
		sm_error("Failed to release session process");

	return ret;
}

ICOMM_RECEIVE(PutDeltaProcess)
ICOMM_RESPOND(PutDeltaProcess)
sm_status_ret_code_t
smc_icomm_request_PutDeltaProcess(int context, sm_value_t *params, sm_value_t **result)
{
        sm_status_ret_code_t ret;

        SM_DEBUG_ENTER();

        if ((ret = smc_icomm_send_value(context, params, result, smc_icomm_PutDeltaProcess)) != SM_OK) {
                sm_error("Failed to send request to session process for request: smc_icomm_PutDeltaProcess");
                /* Continue and try to release the process. */
        }

        return ret;
}

ICOMM_DEFINE(ListDirectory)
ICOMM_DEFINE(StatFile)
ICOMM_DEFINE(Delete)
ICOMM_DEFINE(ChMod)
ICOMM_DEFINE(MkDir)
