/* $Id: jsdl.c 32 2009-12-14 15:43:29Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <common.h>
#include <jsdl.h>
#include <include/jsdl_filter.h>
#include <stub/qcgcompStub.h>
#include <stub/qcgcompH.h>

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "jsdl"
	}
};

extern struct Namespace qcgcomp_namespaces[];

sm_status_ret_code_t
smc_jsdl_setup(void)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	ret = sm_signal_register("smc_jsdl_filter", SM_SIGNAL_DEFAULT,
		(sm_value_type_t []) { SM_VALUE_STRING }, 1,
		(sm_value_type_t []) { SM_VALUE_RET_CODE, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG }, 2);

	return ret;
}

void
smc_jsdl_cleanup(void)
{
	SM_DEBUG_ENTER();

	sm_signal_unregister("smc_jsdl_filter");
}

char *
smc_jsdl_str(_jsdl__JobDefinition *jsdl, smc_strdup_function func, void *ctxt)
{
	char *tmp = NULL;
	char *result = NULL;

	SM_MANDATORY_P(jsdl);

	SM_DEBUG_ENTER();

	tmp = sm_gsoap_buffer_put(jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", qcgcomp_namespaces);

	if (tmp) {
		if (func) {
			SM_MANDATORY_P(ctxt);

			result = func(ctxt, tmp);
			SM_FREE(tmp);
		} else
			result = tmp;
	} else
		sm_error("Failed to get string representation of JSDL");

	return result;
}

sm_alloc_tracked_t
smc_jsdl_from_str(const char *jsdl_str, _jsdl__JobDefinition *jsdl)
{
	sm_alloc_tracked_t handle;

	SM_MANDATORY_P(jsdl_str);
	SM_MANDATORY_P(jsdl);

	SM_DEBUG_ENTER();

	handle = sm_gsoap_buffer_get(jsdl_str, jsdl, jsdl__JobDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle)
		sm_error("Failed to parse JSDL from string");

	return handle;
}

const char *
smc_jsdl_get_username(_jsdl__JobDefinition *jsdl)
{
	SM_MANDATORY_P(jsdl);

	SM_DEBUG_ENTER();

	if (jsdl->jsdl__JobDescription->jsdl__Application) {
		_jsdl_hpcpa__HPCProfileApplication *hpcpapp = jsdl->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;

		if (hpcpapp && hpcpapp->jsdl_hpcpa__UserName)
			return SM_DEBUG_LEAVE_STR(hpcpapp->jsdl_hpcpa__UserName->__item);
	}

	return SM_DEBUG_LEAVE_STR(NULL);
}

const char *
smc_jsdl_get_working_dir(_jsdl__JobDefinition *jsdl, SM_BOOL create)
{
	const char *working_dir = NULL;
	SM_MANDATORY_P(jsdl);

	SM_DEBUG_ENTER();

	if (jsdl->jsdl__JobDescription->jsdl__Application) {
		_jsdl_hpcpa__HPCProfileApplication *hpcpapp = jsdl->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;

		if (hpcpapp && hpcpapp->jsdl_hpcpa__WorkingDirectory) {
			working_dir = hpcpapp->jsdl_hpcpa__WorkingDirectory->__item;
		}
	}

	if (working_dir && create == SM_TRUE) {
		struct stat wd_dir_stat;

		if (stat(working_dir, &wd_dir_stat) != 0 && errno == ENOENT) {
			sm_info("Working directory: %s does not exist. Creating...", working_dir);

			if (mkdir(working_dir, 0755) != 0 && errno != EEXIST)
				sm_perror("Failed to create working directory  (the job will fail)");
		}
	}

	return SM_DEBUG_LEAVE_STR(working_dir);
}

bool
smc_jsdl_has_application(_jsdl__JobDefinition *jsdl)
{
	SM_MANDATORY_P(jsdl);

	SM_DEBUG_ENTER();

	if (jsdl->jsdl__JobDescription->jsdl__Application)
		return true;

	return false;
}

bool
smc_jsdl_stage_in_requested(_jsdl__JobDefinition *jsdl)
{
	int i;

	SM_DEBUG_ENTER();

	for (i = 0; jsdl && i < jsdl->jsdl__JobDescription->__sizeDataStaging; i++) {
		_jsdl__DataStaging *ds = jsdl->jsdl__JobDescription->jsdl__DataStaging + i;

		if (ds->jsdl__Source && ds->jsdl__Source->jsdl__URI)
			return true;
	}

	return false;
}

bool
smc_jsdl_stage_out_requested(_jsdl__JobDefinition *jsdl)
{
	int i;

	SM_DEBUG_ENTER();

	for (i = 0; jsdl && i < jsdl->jsdl__JobDescription->__sizeDataStaging; i++) {
		_jsdl__DataStaging *ds = jsdl->jsdl__JobDescription->jsdl__DataStaging + i;

		if (ds->jsdl__Target && ds->jsdl__Target->jsdl__URI)
			return true;
	}

	return false;
}

sm_ret_code_t
smc_jsdl_handle_filters(const char *jsdl_str, const char **out, smc_strdup_function func, void *ctxt)
{
	sm_list_t return_values = NULL;
	sm_list_t l = NULL;
	sm_value_t *value = NULL;

	SM_DEBUG_ENTER();

	SM_MANDATORY_P(func);
	SM_MANDATORY_P(ctxt);

	SM_TRY_FAULT(sm_signal_emit("smc_jsdl_filter", jsdl_str, &return_values));

	/* We're interested in the output of the last one. */
	for (l = return_values; l; l = l->next) {
		if (!l->next)
			break;
	}

	if (l && (value = l->data)) {
		if (SM_VALUE_RET_CODE_C(value[0]))
			return sm_ret_code_create(SM_VALUE_RET_CODE_C(value[0]), SM_VALUE_RET_CODE_STRING(value[0]));

		sm_debug("After going through a chain of %zu filters, the final JSDL is: %s", sm_list_count(return_values), value[1]->vstring);
		SM_TRY_NOT(*out = func(ctxt, value[1]->vstring));
	} else
		SM_TRY_NOT(*out = func(ctxt, jsdl_str));

	sm_list_free_user(return_values, (sm_list_free_user_function) sm_value_free);
	return sm_ret_code_create(JSDL_FILTER_OK, NULL);

fault:
	sm_list_free_user(return_values, (sm_list_free_user_function) sm_value_free);
	return sm_ret_code_create(JSDL_FILTER_INTERNAL_FAULT, "Internal fault in JSDL filter");
}
