/* $Id: qcg_comp_staging.c 265 2010-08-16 11:22:07Z smaug $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "qcg_comp_socks_server.h"

#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <qcg-core/logger.h>
#include <qcg-core/sm.h>
#include <qcg-core/value.h>
#include <qcg-core/random.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>
#include <qcg-core/thread.h>

#include <faults.h>
#include <common.h>
#include <icomm.h>
#include <jsdl.h>
#include <user.h>
#include <qcg_comp_socks.h>

#include <stub/qcgcompH.h>
#include <stub/qcgcompStub.h>

__SM_ID("$Id: qcg_comp_staging.c 265 2010-08-16 11:22:07Z smaug $");

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "qcg_comp_socks"
	}
};

extern struct Namespace qcg_comp_namespaces[];

struct smc_socks_config_s socks_config;

#define INVALID_THREAD 0
pthread_t smc_socks_thread = INVALID_THREAD;

#define TOKEN_LENGTH 10
#define TOKEN_TTL 3600	//in seconds

static void
smc_socks_token_release(smc_socks_token_t *token)
{
	SM_DEBUG_ENTER_STR((*token)->token);

	sm_free((*token)->token);
	sm_free(*token);
	*token = NULL;

	return SM_DEBUG_LEAVE_VOID();
}

//beware! the lock has to be locked!
static void
smc_socks_tokens_gc(struct smc_socks_tokens_s *tokens)
{
	time_t now;
	size_t i;

	time(&now);

	for(i = 0; i < tokens->size; ++i) {
		if(tokens->data[i] == NULL) {
			continue;
		}

		if(tokens->data[i]->registered + TOKEN_TTL < now) {
			sm_debug("Token %s expired", tokens->data[i]->token);
			smc_socks_token_release(&tokens->data[i]);
		}
	}
}

static SM_BOOL
smc_socks_tokens_add(struct smc_socks_tokens_s *tokens, const char *token_str)
{
	size_t i = 0;
	SM_BOOL result = SM_TRUE;
	smc_socks_token_t token = NULL;

	SM_DEBUG_ENTER();

	SM_TRY_NOT(token = sm_malloc(sizeof(*token)));
	SM_TRY_NOT(token->token = sm_strdup(token_str));
	time(&token->registered);

	sm_mutex_lock(&tokens->lock);

	smc_socks_tokens_gc(tokens);

	for(i = 0; i < tokens->size; ++i) {

		if(tokens->data[i] == NULL) {

			tokens->data[i] = token;
			token = NULL;
			break;
		}
	}

	if(token) {
		smc_socks_token_t *new_data = NULL;
		size_t new_size = tokens->size + 10;

		new_data = sm_realloc(tokens->data, new_size * sizeof(*tokens->data));

		if(new_data) {

			for(i = tokens->size; i < new_size; ++i)
				new_data[i] = NULL;

			new_data[tokens->size] = token;

			tokens->data = new_data;
			tokens->size = new_size;
		}
		else
			result = SM_FALSE;
	}

	sm_mutex_unlock(&tokens->lock);

	return SM_DEBUG_LEAVE_INT(result);

fault:
	return SM_DEBUG_LEAVE_INT(SM_FALSE);
}


static SM_BOOL
smc_socks_tokens_check(struct smc_socks_tokens_s *tokens, const char *token)
{
	SM_BOOL result = SM_FALSE;
	size_t i = 0;

	SM_DEBUG_ENTER();

	sm_mutex_lock(&tokens->lock);

	smc_socks_tokens_gc(tokens);

	for(i = 0; i < tokens->size; ++i) {
		if(tokens->data[i] == NULL) {
			continue;
		}

		if(strcmp(tokens->data[i]->token, token) == 0) {
			result = SM_TRUE;
			smc_socks_token_release(&(tokens->data[i]));
			break;
		}
	}

	sm_mutex_unlock(&tokens->lock);

	return SM_DEBUG_LEAVE_INT(result);
}

static smc_socks_tokens_t
smc_socks_tokens_init() {
	smc_socks_tokens_t tokens = NULL;

	SM_DEBUG_ENTER();

	SM_TRY_NOT(tokens = sm_malloc(sizeof(*tokens)));

	tokens->data = NULL;
	tokens->size = 0;
	SM_TRY_NOT_OK(sm_mutex_init(&tokens->lock));

	tokens->add = smc_socks_tokens_add;
	tokens->check = smc_socks_tokens_check;

	return SM_DEBUG_LEAVE_PTR(tokens);

fault:
	return SM_DEBUG_LEAVE_PTR(NULL);
}

static void
smc_socks_tokens_release(smc_socks_tokens_t *tokens) {
	size_t i;

	SM_DEBUG_ENTER();

	//don't lock mutext, because there must be no other threads that wants to use this list

	for(i = 0; i < (*tokens)->size; ++i) {
		if((*tokens)->data[i] != NULL) {
			smc_socks_token_release(&((*tokens)->data[i]));
		}
	}

	sm_mutex_destroy(&((*tokens)->lock));	//there is nothing we can do if it fails
	sm_free(*tokens);
	*tokens = NULL;
}

sm_status_ret_code_t
smc_socks_setup(sm_module_t this)
{
	char *socks_policy = NULL;
	int i;
	char *str = NULL;

	SM_DEBUG_ENTER();

	if (sm_conf_node_raw(this->conf_ctxt, SM_FALSE, "/conf:SOCKSPolicy", &socks_policy) == SM_FAULT)
		goto fault;

	sm_info("socks_policy = '%s'", socks_policy);

	if (!socks_policy || !strlen(socks_policy)) {
		SM_FREE(socks_policy);
		sm_debug("No SOCKS policy defined. Socks module not started");
		return SM_NEG;
	}

	if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, "/conf:SOCKSPolicy/@bindPort", &socks_config.socks_port) != SM_OK) {
		sm_error("No SOCKS port defined.");
		goto fault;
	}

	if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, "/conf:SOCKSPolicy/@authentication", &str) != SM_OK) {
		sm_error("No SOCKS authentication method defined.");
		goto fault;
	}

	if (strcmp(str,"perConnection") == 0) {
		socks_config.authentication = smc_socks_atn_per_connection;
	} else if (strcmp(str, "perSession") == 0) {
		socks_config.authentication = smc_socks_atn_per_session;
	} else if (strcmp(str, "none") == 0) {
		socks_config.authentication = smc_socks_atn_none;
	} else {
		SM_ASSERT(0, "Unknown SOCKS authentication method");
	}

	sm_info("socks_port = %s", socks_config.socks_port);
	sm_info("socks_authentication = %s", str);
	SM_FREE(str);

	if ((socks_config.allowed_src_hosts_count = sm_conf_node_count(this->conf_ctxt, SM_FALSE, "/conf:SOCKSPolicy/conf:AllowSrcHosts")) < 0) {
		goto fault;
	}

	if ((socks_config.allowed_dest_ports_count = sm_conf_node_count(this->conf_ctxt, SM_FALSE, "/conf:SOCKSPolicy/conf:AllowDestPorts")) < 0) {
		goto fault;
	}

	if ((socks_config.allowed_dest_hosts_count = sm_conf_node_count(this->conf_ctxt, SM_FALSE, "/conf:SOCKSPolicy/conf:AllowDestHosts")) < 0) {
		goto fault;
	}

	sm_debug("allowed_dest_ports_count = %d", socks_config.allowed_dest_ports_count);
	sm_debug("allowed_dest_hosts_count = %d", socks_config.allowed_dest_hosts_count);
	sm_debug("allowed_src_hosts_count  = %d", socks_config.allowed_src_hosts_count);


	socks_config.allowed_ports = sm_calloc(socks_config.allowed_dest_ports_count, sizeof(*socks_config.allowed_ports));

	for (i = 0; i < socks_config.allowed_dest_ports_count; i++) {
		char xpath[128] = "";
		int min_port = -1;
		int max_port = -1;

		snprintf(xpath, sizeof(xpath), "/conf:SOCKSPolicy/conf:AllowDestPorts[%d]/@minPort", i+1);

		if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, xpath, &min_port) != SM_OK) {
			goto fault;
		}

		snprintf(xpath, sizeof(xpath), "/conf:SOCKSPolicy/conf:AllowDestPorts[%d]/@maxPort", i+1);

		if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, xpath, &max_port) != SM_OK) {
			goto fault;
		}

		socks_config.allowed_ports[i].min_port = min_port;
		socks_config.allowed_ports[i].max_port = max_port;

		sm_info("Allowed dest ports[%d]: <%d,%d>", i+1, min_port, max_port);
	}

	socks_config.allowed_dst_hosts = sm_calloc(socks_config.allowed_dest_hosts_count, sizeof(*socks_config.allowed_dst_hosts));

	for (i = 0; i < socks_config.allowed_dest_hosts_count; i++) {
		char xpath[128] = "";
		struct in_addr host, mask;
		char *str_host, *str_mask;

		snprintf(xpath, sizeof(xpath), "/conf:SOCKSPolicy/conf:AllowDestHosts[%d]", i+1);

		if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, xpath, &str) != SM_OK) {
			goto fault;
		}

		sm_debug("AllowDestHosts[%d] = %s",i, str);

		SM_TRY_NOT(inet_aton(str, &host));
		SM_FREE(str);

		snprintf(xpath, sizeof(xpath), "/conf:SOCKSPolicy/conf:AllowDestHosts[%d]/@mask", i+1);

		if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, xpath, &str) != SM_OK) {
			str = sm_strdup("255.255.255.255");
		}

		sm_debug("AllowDestHosts[%d]/@mask = %s", i, str);

		SM_TRY_NOT(inet_aton(str, &mask));
		SM_FREE(str);

		sm_debug("host = %u, mask = %u", host.s_addr, mask.s_addr);

		socks_config.allowed_dst_hosts[i].address = host.s_addr;
		socks_config.allowed_dst_hosts[i].mask = mask.s_addr;

		str_host = sm_strdup(inet_ntoa(host));
		str_mask = sm_strdup(inet_ntoa(mask));

		sm_info("Allowed dest hosts[%d]: %s/%s", i+1, str_host, str_mask);

		SM_FREE(str_host);
		SM_FREE(str_mask);
	}

	socks_config.allowed_src_hosts = sm_calloc(socks_config.allowed_src_hosts_count, sizeof(*socks_config.allowed_src_hosts));

	for (i = 0; i < socks_config.allowed_src_hosts_count; i++) {
		char xpath[128] = "";
		struct in_addr host, mask;
		char *str_host, *str_mask;

		snprintf(xpath, sizeof(xpath), "/conf:SOCKSPolicy/conf:AllowSrcHosts[%d]", i+1);

		if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, xpath, &str) != SM_OK) {
			goto fault;
		}

		sm_debug("AllowSrcHosts[%d] = %s",i, str);

		SM_TRY_NOT(inet_aton(str, &host));
		SM_FREE(str);

		snprintf(xpath, sizeof(xpath), "/conf:SOCKSPolicy/conf:AllowSrcHosts[%d]/@mask", i+1);

		if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, xpath, &str) != SM_OK) {
			str = sm_strdup("255.255.255.255");
		}

		sm_debug("AllowSrcHosts[%d]/@mask = %s", i, str);

		SM_TRY_NOT(inet_aton(str, &mask));
		SM_FREE(str);

		sm_debug("host = %u, mask = %u", host.s_addr, mask.s_addr);

		socks_config.allowed_src_hosts[i].address = host.s_addr;
		socks_config.allowed_src_hosts[i].mask = mask.s_addr;

		str_host = sm_strdup(inet_ntoa(host));
		str_mask = sm_strdup(inet_ntoa(mask));

		sm_info("Allowed src hosts[%d]: %s/%s", i+1, str_host, str_mask);

		SM_FREE(str_host);
		SM_FREE(str_mask);
	}

	SM_TRY_NOT(socks_config.tokens = smc_socks_tokens_init());

	SM_TRY_NOT_OK(sm_thread_create(&smc_socks_thread, NULL, (void*(*)(void*))socks_main, &socks_config));

	SM_FREE(socks_policy);
	return SM_OK;
fault:
	SM_FREE(str);
	SM_FREE(socks_policy);
	return SM_FAULT;
}

void
smc_socks_cleanup(void)
{
	SM_DEBUG_ENTER();

	if(smc_socks_thread != INVALID_THREAD) {
		(void)sm_thread_cancel(smc_socks_thread);	//even if this fails what can we do?
	} else  {
		sm_debug("SOCKS server disabled");
		return;
	}

	sm_free(socks_config.allowed_dst_hosts);
	sm_free(socks_config.allowed_src_hosts);
	sm_free(socks_config.allowed_ports);

	smc_socks_tokens_release(&socks_config.tokens);
}

int
__qcg_comp_socks__GetConnectionToken(struct soap* soap, struct qcg_comp_socks__GetConnectionTokenType *request sm_unused, struct qcg_comp_socks__GetConnectionTokenResponseType *response)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_transport_t transport = soap->user;
	char *token = NULL;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetConnectionToken", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "GetConnectionToken", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	memset(response, 0, sizeof(*response));

	sm_debug("Generating token");

	SM_TRY_FAULT(sm_random_string(TOKEN_LENGTH, &token));

	sm_debug("Token %s generated", token);

	if(socks_config.tokens->add(socks_config.tokens, token)) {
		sm_debug("Token registered");
		SM_TRY_NOT(response->SecurityToken = transport->transport_strdup(transport, token));
	}
	else {
		response->SecurityToken = NULL;
	}

	sm_debug("token = %s", response->SecurityToken);

	SM_FREE(token);
	SM_DEBUG_LEAVE_VOID();
	return transport->exit(transport, NULL);
fault:
	SM_FREE(token);
	SM_DEBUG_LEAVE_VOID();
	return transport->fault(transport, fault);
}

