/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "qcg_comp_socks_server.h"
#include "qcg_comp_socks_serve.h"

#include <qcg-core/logger.h>
#include <qcg-core/modules/modules.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "socks_client"
	}
};

static void client_fail_without_resp(client_context_t context) {
	sm_debug("client %i miserably failed", context->socket);
	context->status = Fail;
}

static void client_fail_with_resp(client_context_t context) {
	char response[8]={0,91};

	send(context->socket, response, sizeof(response), 0);

	client_fail_without_resp(context);
}

static void client_ok(client_context_t context) {
	char response[8]={0,90};

	sm_debug("client %i ok", context->socket);

	send(context->socket, response, sizeof(response), 0);
	context->status = Ok;
}

static enum client_status transfer(int fd, char *buffer, int *len) {
	ssize_t n;

	sm_debug("Transferring %u bytes to %i", *len, fd);

	while(*len > 0) {
		n = send(fd, buffer, *len, 0);

		if(n == 0)
			return Fail;

		 if(n == -1) {
			 if(errno == EAGAIN || errno == EINTR)
				 return Blocked;
			 else
				 return Fail;
		 }

		*len -= n;
	}
	return Ok;
}

static void client_process(client_context_t context) {
	enum client_status status;

	status = transfer(context->remote, context->buffer, &(context->len));

	if(status != Ok) {
		context->status = status;
		return;
	}

	status = transfer(context->socket, context->remote_buffer, &(context->remote_len));

	if(status != Ok) {
		context->status = status;
		return;
	}

	context->status = Blocked;
}

static void client_connected(client_context_t context) {
	context->process = client_process;
	context->fail = client_fail_without_resp;
	client_ok(context);
}

static void client_wait_for_write(client_context_t context) {
	assert(context->remote_fd);

	if(context->remote_fd->revents & POLLOUT) {
		context->remote_events = 0;
		context->process = client_connected;
		context->status = Ok;
	}
	else {
		context->status = Blocked;
	}
}

static int client_make_connection(client_context_t context, int family, int type, int proto, struct sockaddr* addr, socklen_t addr_len) { /*TODO be consequent : SM_OK == 0 */
	int s, result;
	int yes = 1;

	s = socket(family, type, proto);

	if(s < 0) {
		return 0;
	}

	result = ioctl(s, FIONBIO, &yes);

	if(result != 0) {
		sm_perror("ioctl");
		return 0;
	}

	result = connect(s, addr, addr_len);

	if(result != 0) {
		if(errno != EINPROGRESS && errno != EINTR) {
			sm_perror("connect");
			return 0;
		}
		context->remote_events = POLLOUT;
		context->process = client_wait_for_write;
		context->status = Blocked;
	}
	else {
		context->process = client_connected;
		context->status = Ok;
	}

	context->remote = s;
	return 1;
}

static void client_connect(client_context_t context) {
	SM_DEBUG_ENTER();

	if(!client_make_connection(context, AF_INET, SOCK_STREAM, IPPROTO_TCP, (struct sockaddr*)&(context->dst), sizeof(context->dst)))
		context->fail(context);

	SM_DEBUG_LEAVE_VOID();
}


static void client_check(client_context_t context) {
	int i = 0;
	short port = 0;
	SM_BOOL ok_port = SM_FALSE;
	SM_BOOL ok_addr = SM_FALSE;

	SM_DEBUG_ENTER();

	port = ntohs(context->dst.sin_port);

	for(i = 0; i < context->config->allowed_dest_ports_count; ++i) {

		if(context->config->allowed_ports[i].min_port <= port && port <= context->config->allowed_ports[i].max_port) {
			sm_debug("Port accepted: %i <= %i <= %i", context->config->allowed_ports[i].min_port, port, context->config->allowed_ports[i].max_port);
			ok_port = SM_TRUE;
			break;
		}
	}

	if (!ok_port) {
		sm_notice("Destination port = %d not allowed", port);
		context->fail(context);
	        SM_DEBUG_LEAVE_VOID();
	}

	for(i = 0; i < context->config->allowed_dest_hosts_count; ++i) {

		if((context->config->allowed_dst_hosts[i].address & context->config->allowed_dst_hosts[i].mask) == (context->dst.sin_addr.s_addr & context->config->allowed_dst_hosts[i].mask)) {
			sm_debug("Address accepted");
			ok_addr = SM_TRUE;
			break;
		}
	}

	if (!ok_addr) {
		sm_notice("Destination host not allowed");
		context->fail(context);
	        SM_DEBUG_LEAVE_VOID();
	}


	context->process = client_connect;
	context->status = Ok;

	SM_DEBUG_LEAVE_VOID();
}

static void client_get_username(client_context_t context) {
	int i;
	SM_BOOL result = SM_FALSE;

	SM_DEBUG_ENTER();

	context->status = Blocked;

	for(i=0 ; i<context->len ; ++i) {

		if(context->buffer[i]=='\0') {

			sm_debug("username=%s", context->buffer);
			
			if ( 1 ) {
				context->status = Ok;
                                context->process = client_check;
				context->len-=i+1;
				memmove(context->buffer, context->buffer+i+1, context->len);
			} 
			else {
				result = context->config->tokens->check(context->config->tokens, context->buffer);
				context->len-=i+1;
				memmove(context->buffer, context->buffer+i+1, context->len);

				if(result) {
					context->status = Ok;
					context->process = client_check;
				}
				else
					context->fail(context);
			} 

			break;
		}
	}

	SM_DEBUG_LEAVE_VOID();
}

static void client_init(client_context_t context) {
	SM_DEBUG_ENTER();

	if(context->len < 7) {
		context->status = Blocked;
		return SM_DEBUG_LEAVE_VOID();
	}

	if(context->buffer[0] != 1) { //accept only CONNECT requests
		sm_notice("Client %i, invalid request", context->socket);
		return context->fail(context);
	}

	memcpy(&context->dst.sin_port, context->buffer+1, 2);
	memcpy(&context->dst.sin_addr.s_addr, context->buffer+3, 4);
	context->dst.sin_family = AF_INET;

	context->len-=7;
	memmove(context->buffer, context->buffer+7, context->len);

	context->process = client_get_username;
	context->status = Ok;

	SM_DEBUG_LEAVE_VOID();
}

static void client_check_version(client_context_t context) {
	SM_DEBUG_ENTER();

	if(context->len < 1) {
		context->status = Blocked;
		return SM_DEBUG_LEAVE_VOID();
	}

	if(context->buffer[0] != 4) {	//protocol version number
		sm_notice("Client %i, protocol version mismatch: %d", context->socket, context->buffer[0]);
		return context->fail(context);
	}

	context->len -= 1;
	memmove(context->buffer, context->buffer+1, context->len);

	context->process = client_init;
	context->status = Ok;

	SM_DEBUG_LEAVE_VOID();
}

client_context_t create_client_context(int fd, smc_socks_config_t config, struct sockaddr* addr, socklen_t addr_len) {
	char host[128], service[32];
	client_context_t result = NULL;

	SM_DEBUG_ENTER();

	if(getnameinfo(addr, addr_len, host, sizeof(host), service, sizeof(service), NI_NUMERICHOST|NI_NUMERICSERV) == 0) {
		SM_BOOL ok_addr = SM_FALSE;
		int i;

		sm_info("Client from %s:%s at socket %i", host, service, fd);

		for (i = 0; i < config->allowed_src_hosts_count; ++i) {

			if((config->allowed_src_hosts[i].address & config->allowed_src_hosts[i].mask) == (((struct sockaddr_in*)addr)->sin_addr.s_addr & config->allowed_src_hosts[i].mask)) {
				sm_debug("Address accepted");
				ok_addr = SM_TRUE;
				break;
			}
		}

		if (!ok_addr) {
			sm_notice("Client address rejected");
			goto fail;
                }
	}
	else {
		sm_error("Unable to get peer address at socket %i", fd);
		goto fail;
	}

	result = sm_malloc(sizeof(*result));

	if(!result)
		goto fail;

	memset(result, 0, sizeof(*result));
	result->socket = fd;
	result->remote = -1;
	result->config = config;
	result->process = client_check_version;
	result->fail = client_fail_with_resp;

	result->size = 100*1024;
	result->buffer = sm_malloc(result->size);

	if(!result->buffer)
		goto fail;

	result->remote_size = 100*1024;
	result->remote_buffer = sm_malloc(result->remote_size);

	if(!result->remote_buffer)
		goto fail;

	result->status = Ok;

	return SM_DEBUG_LEAVE_PTR(result);
fail:

	if(result) {
		if(result->fail)
			result->fail(result);

		if(result->buffer)
			sm_free(result->buffer);

		if(result->remote_buffer)
			sm_free(result->remote_buffer);

		sm_free(result);
	}

	close(fd);

	return SM_DEBUG_LEAVE_PTR(NULL);
}

void release_client_context(client_context_t client) {

	SM_DEBUG_ENTER();

	sm_debug("Releasing client fd=%i (remote=%i)", client->socket, client->remote);

	if(client->buffer)
		free(client->buffer);

	if(client->remote_buffer)
		free(client->remote_buffer);

	if(client->domain)
		free(client->domain);

	close(client->socket);
	close(client->remote);

	free(client);

	SM_DEBUG_LEAVE_VOID();
}
