/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "qcg_comp_socks_server.h"
#include "qcg_comp_socks_serve.h"

#include <qcg-core/logger.h>
#include <qcg-core/modules/modules.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "socks_server"
	}
};

static int*
start_server(const char *host, const char *port) {
	struct addrinfo hints = {
			.ai_family = PF_INET,
			.ai_socktype = SOCK_STREAM,
			.ai_protocol = IPPROTO_TCP,
			.ai_flags = AI_PASSIVE
	};
	struct addrinfo *result = NULL, *i;
	int rv;
	int n = 0;
	int *sockets = NULL;
	int pos = 0;

	SM_DEBUG_ENTER();

	rv = getaddrinfo(host, port, &hints, &result);

	if(rv != 0) {
		sm_info("getaddrinfo: %s (%i)", gai_strerror(rv), rv);
		goto fault;
	}

	for(i = result ; i ; i = i->ai_next) {
		++n;
	}

	if(n == 0)
		goto fault;

	++n;
	sockets = sm_calloc(n, sizeof(int));
	pos = 0;

	for(i = result ; i ; i = i->ai_next) {
		int s;
		int yes = 1;

		s = socket(i->ai_family, i->ai_socktype, i->ai_protocol);

		if(s == -1)
			continue;

		(void)setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));	//ignoring result

		if(ioctl(s, FIONBIO, &yes) != 0) {
			sm_perror("ioctl");
			close(s);
			continue;
		}

		if(bind(s, i->ai_addr, i->ai_addrlen) != 0) {
			sm_perror("bind");
			close(s);
			continue;
		}

		if(listen(s, SOMAXCONN) != 0) {
			sm_perror("listen");
			close(s);
			continue;
		}

		sockets[pos++] = s;
	}

	sm_debug("Created %i sockets", pos);

	for(; pos < n; ++pos)
		sockets[pos] = -1;

	sm_debug("pos = %i", pos); /* print at least error message if pos == 0?*/

quit:

	if(result)
		freeaddrinfo(result);

	return sockets;

fault:

	if(sockets) {
		sm_free(sockets);
		sockets = NULL;
	}

	goto quit;
}

typedef struct clients_queue_s {
	client_context_t head;
	int n;
} *clients_queue_t;

static void
process_new_client(int fd, clients_queue_t clients, smc_socks_config_t config) {

	struct sockaddr_storage addr;
	socklen_t len = sizeof(addr);
	int client;
	client_context_t context;

	SM_DEBUG_ENTER();

	client = accept(fd, (struct sockaddr*)&addr, &len);

	if(client >= 0) {
		context = create_client_context(client, config, (struct sockaddr*)&addr, len);

		if(context) {
			context->next = clients->head;
			clients->head = context;
			clients->n++;
		}
	}
	else {
		sm_perror("accept");
	}

	SM_DEBUG_LEAVE_VOID();
}

static void
purge_dead_clients(clients_queue_t clients) {
	client_context_t client = clients->head;
	client_context_t prev = NULL;
	client_context_t to_free;

	SM_DEBUG_ENTER();

	while(client) {

		if(client->status == Fail) {
			if(prev == NULL)
				clients->head = client->next;
			else
				prev->next = client->next;

			to_free = client;
			client = client->next;
			release_client_context(to_free);

			clients->n--;
		}
		else {
			prev = client;
			client = client->next;
		}
	}

	SM_DEBUG_LEAVE_VOID();
}

void*
socks_main(smc_socks_config_t config) {
	int *srv_sockets;
	size_t n_srv = 0, i;
	ssize_t n;
	struct pollfd *fds = NULL;
	int rv;
	struct clients_queue_s clients;
	client_context_t client;

	SM_DEBUG_ENTER();

	memset(&clients, 0, sizeof(clients));
	signal(SIGPIPE, SIG_IGN);

	srv_sockets = start_server(NULL, config->socks_port);

	while(srv_sockets[n_srv] != -1)
		++n_srv;

	for(;;) {
		size_t n_fds = n_srv + 2*clients.n;

		sm_thread_testcancel();

		fds = realloc(fds, n_fds * sizeof(struct pollfd));

		sm_debug("poll table size %u (%u server sockets, %u client sockets)", n_fds, n_srv, clients.n);

		for(i=0 ; i<n_srv ; ++i) {
			fds[i].fd = srv_sockets[i];
			fds[i].events = POLLIN;
		}

		for(client = clients.head ; client ; client=client->next) {
			fds[i].fd = client->socket;
			fds[i].events = POLLIN;
			client->fd = fds + i;
			++i;

			if(client->remote >= 0) {
				fds[i].fd = client->remote;
				fds[i].events = POLLIN | client->remote_events;
				client->remote_fd = fds + i;
				++i;
			}
			else
				client->remote_fd = NULL;
		}

		rv = poll(fds, i, -1);

		sm_debug("poll=%i", rv);

		for(i=0 ; i<n_srv ; ++i) {

			sm_debug("fds[%i].revents = %i", i, fds[i].revents);

			if(fds[i].revents & POLLIN)
				process_new_client(fds[i].fd, &clients, config);
		}


		for(client = clients.head ; client ; client=client->next) {

			if(!client->fd)	//ignore brand new clients
				continue;

			sm_debug("client->fd->revents = %i", client->fd->revents);

			if(client->fd->revents & (POLLERR|POLLHUP|POLLNVAL)) {
				sm_debug("client is broken");
				client->status = Fail;
				continue;
			}

			if(client->remote_fd) {
				if(client->remote_fd->revents & (POLLERR|POLLHUP|POLLNVAL)) {
					sm_debug("remote is broken, remote_fd = %p, revents = %i", client->remote_fd, client->remote_fd->revents);
					client->status = Fail;
					continue;
				}
			}

			if(client->fd->revents & POLLIN) {
				n = recv(client->socket, client->buffer+client->len, client->size-client->len, 0);

				if(n == -1) {
					if(errno != EINTR && errno != EAGAIN) {
						perror("recv");
						client->fail(client);
					}
				}
				else if(n == 0) {
					client->fail(client);
				}
				else	//new data available
					client->len += n;
			}

			if(client->remote_fd) {
				sm_debug("client->remote = %i, client->remote_fd->revents = %i", client->remote, client->remote_fd->revents);

				if(client->remote_fd->revents & POLLIN) {
					n = recv(client->remote, client->remote_buffer+client->remote_len, client->remote_size-client->remote_len, 0);

					if(n == -1) {
						if(errno != EINTR && errno != EAGAIN) {
							perror("recv");
							client->fail(client);
						}
					}
					else if(n == 0) {
						client->fail(client);
					}
					else	//new data available
						client->remote_len += n;
				}
			}

			if(client->status != Fail) {
				do {
					client->process(client);
				}
				while(client->status == Ok);
			}

			sm_debug("client %i status %i", client->socket, client->status);
		}

		purge_dead_clients(&clients);
	}

	return SM_DEBUG_LEAVE_PTR(NULL);
}
