/* $Id: rsvn.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <stdbool.h>
#include <stdlib.h>
#include <time.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/odbc.h>
#include <qcg-core/signal.h>
#include <qcg-core/uuid.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <ardl.h>
#include <identifiers.h>
#include <rsvn.h>
#include <user.h>
#include <stub/qcgcompH.h>

#define RESERVATION_TIME_SHIFT		15

#define SMC_RSVN_INTERNAL_FAULT_STR	"Internal fault in rsvn module"
#define SMC_RSVN_DATABASE_FAULT_STR	"Database fault in rsvn module"
#define SMC_RSVN_NOT_FOUND_STR		"Reservation not found"

#if defined(TEST) && !defined(RSVN_NO_TEST)
#include <stub/qcgcomp.nsmap>
#else
extern struct Namespace qcgcomp_namespaces[];
#endif

__SM_ID("$Id: rsvn.c 2 2009-10-05 10:09:52Z mamonski $");

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "rsvn"
        }
};

typedef struct smc_concrete_rsvn_s {
	struct smc_rsvn_s super;

	char *id;
	char *drms_id;
	char *submitted_ardl;
	char *ardl;
	_ardl__ReservationDefinition *ardl_parsed;
	char *subject;
	char *mapped_user;
	struct tm submission_time;
	struct tm start_time;
	struct tm end_time;

	bool submission_time_set;
	bool start_time_set;
	bool end_time_set;

	sm_alloc_tracked_t tracked;
	sm_ret_code_t error;
	bool in_transaction;
	bool new_entry;
	bool modified;
	bool readonly;

	/** Was it read from accounting table? */
	bool accounting;

	/** Should it be archived in accounting table? */
	bool archive;

	bool terminated;
} *smc_concrete_rsvn_t;

static const char *find_txt = "SELECT * FROM reservations WHERE id = ? FOR UPDATE";
static const char *find_by_drms_id_txt = "SELECT * FROM reservations WHERE drms_id = ? FOR UPDATE";

static const char *find_readonly_txt = "SELECT * FROM reservations WHERE id = ?";

/* We limit the query to only 1 row because we do reservations->reservations_acc
 * movement with a delay and theoretically there's a possibility of drms_id wind up. */
static const char *find_by_drms_id_readonly_txt = "SELECT * FROM reservations WHERE drms_id = ? ORDER BY submission_time DESC LIMIT 1";

static const char *find_acc_readonly_txt = "SELECT * FROM reservations_acc WHERE id = ?";
static const char *find_acc_by_drms_id_readonly_txt = "SELECT * FROM reservations_acc WHERE drms_id = ? ORDER BY submission_time DESC LIMIT 1";

static const char *insert_txt = "INSERT INTO reservations (id, drms_id,"
	"submitted_ardl, ardl, subject, mapped_user, submission_time, start_time, end_time)"
	"VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";

static const char *update_txt = "UPDATE reservations SET id = ?, drms_id = ?,"
	"submitted_ardl = ?, ardl = ?, subject = ?, mapped_user = ?,"
	"submission_time = ?, start_time = ?, end_time = ? WHERE id = ?";

static const char *delete_txt = "DELETE FROM reservations WHERE id = ?";

static const char *insert_acc_txt = "INSERT INTO reservations_acc (id, drms_id,"
	"submitted_ardl, ardl, subject, mapped_user, submission_time, start_time, end_time)"
	"VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";

static const char *count_txt = "SELECT count(*) FROM reservations";
static const char *find_all_readonly_txt = "SELECT * from reservations";

static sm_odbc_handle_t odbc;

/*
 * Statements are global as we assume that odbc->begin_transaction locks the
 * ODBC handle and thus only one statement can be used at a time.
 */
static sm_odbc_stmt_t find_stmt;
static sm_odbc_stmt_t find_by_drms_id_stmt;
static sm_odbc_stmt_t find_readonly_stmt;
static sm_odbc_stmt_t find_by_drms_id_readonly_stmt;
static sm_odbc_stmt_t find_acc_readonly_stmt;
static sm_odbc_stmt_t find_acc_by_drms_id_readonly_stmt;
static sm_odbc_stmt_t insert_stmt;
static sm_odbc_stmt_t insert_stmt;
static sm_odbc_stmt_t update_stmt;
static sm_odbc_stmt_t delete_stmt;
static sm_odbc_stmt_t insert_acc_stmt;
static sm_odbc_stmt_t count_stmt;
static sm_odbc_stmt_t find_all_readonly_stmt;

static const char *smc_rsvn_get_id(smc_concrete_rsvn_t this);
static const char *smc_rsvn_get_drms_id(smc_concrete_rsvn_t this);
static const char *smc_rsvn_get_submitted_ardl(smc_concrete_rsvn_t this);
static const char *smc_rsvn_get_ardl(smc_concrete_rsvn_t this);
static _ardl__ReservationDefinition *smc_rsvn_get_ardl_parsed(smc_concrete_rsvn_t this);
static const char *smc_rsvn_get_subject(smc_concrete_rsvn_t this);
static const char *smc_rsvn_get_mapped_user(smc_concrete_rsvn_t this);
static struct tm smc_rsvn_get_submission_time(smc_concrete_rsvn_t this);
static struct tm smc_rsvn_get_start_time(smc_concrete_rsvn_t this);
static struct tm smc_rsvn_get_end_time(smc_concrete_rsvn_t this);
static bool smc_rsvn_is_readonly(smc_concrete_rsvn_t this);

static const char *smc_rsvn_set_drms_id(smc_concrete_rsvn_t this, const char *value);
static const char *smc_rsvn_set_submitted_ardl(smc_concrete_rsvn_t this, const char *value);
static const char *smc_rsvn_set_ardl(smc_concrete_rsvn_t this, const char *value);
static const char *smc_rsvn_set_ardl_parsed(smc_concrete_rsvn_t this, _ardl__ReservationDefinition *value);
static const char *smc_rsvn_set_subject(smc_concrete_rsvn_t this, const char *value);
static const char *smc_rsvn_set_mapped_user(smc_concrete_rsvn_t this, const char *value);
static struct tm smc_rsvn_set_start_time(smc_concrete_rsvn_t this, struct tm value);
static struct tm smc_rsvn_set_end_time(smc_concrete_rsvn_t this, struct tm value);

static sm_ret_code_t smc_rsvn_save(smc_concrete_rsvn_t this);

static void smc_rsvn_gen_id(smc_concrete_rsvn_t this);
static smc_rsvn_t smc_rsvn_create_internal(bool readonly, bool new, bool transaction, bool accounting);
static sm_status_ret_code_t smc_rsvn_insert(smc_concrete_rsvn_t this);
static sm_status_ret_code_t smc_rsvn_update(smc_concrete_rsvn_t this);
static sm_status_ret_code_t smc_rsvn_move_to_acc(smc_concrete_rsvn_t this);
static sm_ret_code_t smc_rsvn_process_ardl(smc_concrete_rsvn_t this);
static sm_ret_code_t smc_rsvn_get_internal(const char *id, smc_rsvn_t *result, bool readonly);
static sm_ret_code_t smc_rsvn_get_by_drms_id_internal(const char *drms_id, smc_rsvn_t *result, bool readonly);
static struct tm timetm(void);
static sm_ret_code_t smc_rsvn_get_times(smc_concrete_rsvn_t this, _ardl__ReservationDefinition *ardl);

static void smc_rsvn_archive(smc_concrete_rsvn_t this);
static void smc_rsvn_terminate(smc_concrete_rsvn_t this);

sm_status_ret_code_t
smc_rsvn_setup(sm_odbc_handle_t odbc_handle)
{
	int ret;

	SM_MANDATORY_P(odbc_handle);

	SM_DEBUG_ENTER();

	odbc = odbc_handle;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	ret = sm_signal_register("smc_rsvn_created", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING /* drms_id */,
		SM_VALUE_STRING /* submitted_ardl */, SM_VALUE_STRING /* ardl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */, SM_VALUE_INT32 /* start_time */, SM_VALUE_INT32 /* end_time */ },
		9,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);

	ret = sm_signal_register("smc_rsvn_terminated", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING /* drms_id */,
		SM_VALUE_STRING /* submitted_ardl */, SM_VALUE_STRING /* ardl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */, SM_VALUE_INT32 /* start_time */, SM_VALUE_INT32 /* end_time */ },
		9,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);

	if (ret & SM_FAULT)
		goto fault;

	SM_TRY_NOT(find_stmt = sm_odbc_stmt_create(odbc, find_txt));
	SM_TRY_NOT(find_by_drms_id_stmt = sm_odbc_stmt_create(odbc, find_by_drms_id_txt));
	SM_TRY_NOT(find_readonly_stmt = sm_odbc_stmt_create(odbc, find_readonly_txt));
	SM_TRY_NOT(find_by_drms_id_readonly_stmt = sm_odbc_stmt_create(odbc, find_by_drms_id_readonly_txt));
	SM_TRY_NOT(find_acc_readonly_stmt = sm_odbc_stmt_create(odbc, find_acc_readonly_txt));
	SM_TRY_NOT(find_acc_by_drms_id_readonly_stmt = sm_odbc_stmt_create(odbc, find_acc_by_drms_id_readonly_txt));
	SM_TRY_NOT(insert_stmt = sm_odbc_stmt_create(odbc, insert_txt));
	SM_TRY_NOT(update_stmt = sm_odbc_stmt_create(odbc, update_txt));
	SM_TRY_NOT(insert_acc_stmt = sm_odbc_stmt_create(odbc, insert_acc_txt));
	SM_TRY_NOT(delete_stmt = sm_odbc_stmt_create(odbc, delete_txt));
	SM_TRY_NOT(count_stmt = sm_odbc_stmt_create(odbc, count_txt));
	SM_TRY_NOT(find_all_readonly_stmt = sm_odbc_stmt_create(odbc, find_all_readonly_txt));

	SM_TRY_FAULT(odbc->commit(odbc));

	return SM_OK;

fault:
	smc_rsvn_cleanup();
	odbc->rollback(odbc);
	return SM_FAULT;
}

void
smc_rsvn_cleanup(void)
{
	SM_DEBUG_ENTER();

	sm_odbc_stmt_free(&find_stmt);
	sm_odbc_stmt_free(&find_by_drms_id_stmt);
	sm_odbc_stmt_free(&find_readonly_stmt);
	sm_odbc_stmt_free(&find_by_drms_id_readonly_stmt);
	sm_odbc_stmt_free(&find_acc_readonly_stmt);
	sm_odbc_stmt_free(&find_acc_by_drms_id_readonly_stmt);
	sm_odbc_stmt_free(&insert_stmt);
	sm_odbc_stmt_free(&update_stmt);
	sm_odbc_stmt_free(&delete_stmt);
	sm_odbc_stmt_free(&insert_acc_stmt);
	sm_odbc_stmt_free(&count_stmt);
	sm_odbc_stmt_free(&find_all_readonly_stmt);

	sm_signal_unregister("smc_rsvn_created");
	sm_signal_unregister("smc_rsvn_terminated");
}

smc_rsvn_t
smc_rsvn_create(void)
{
	SM_DEBUG_ENTER();
	return smc_rsvn_create_internal(false, true, false, false);
}

static smc_rsvn_t
smc_rsvn_create_internal(bool readonly, bool new, bool transaction, bool accounting)
{
	smc_concrete_rsvn_t ret;

	SM_DEBUG_ENTER();

	if (!(ret = sm_malloc(sizeof(*ret))))
		return NULL;

	ret->super.get_id = (smc_rsvn_string_getter_function) smc_rsvn_get_id;
	ret->super.get_drms_id = (smc_rsvn_string_getter_function) smc_rsvn_get_drms_id;
	ret->super.get_submitted_ardl = (smc_rsvn_string_getter_function) smc_rsvn_get_submitted_ardl;
	ret->super.get_ardl = (smc_rsvn_string_getter_function) smc_rsvn_get_ardl;
	ret->super.get_ardl_parsed = (smc_rsvn_ardl_parsed_getter_function) smc_rsvn_get_ardl_parsed;
	ret->super.get_subject = (smc_rsvn_string_getter_function) smc_rsvn_get_subject;
	ret->super.get_mapped_user = (smc_rsvn_string_getter_function) smc_rsvn_get_mapped_user;
	ret->super.get_submission_time = (smc_rsvn_time_getter_function) smc_rsvn_get_submission_time;
	ret->super.get_start_time = (smc_rsvn_time_getter_function) smc_rsvn_get_start_time;
	ret->super.get_end_time = (smc_rsvn_time_getter_function) smc_rsvn_get_end_time;

	ret->super.set_drms_id = (smc_rsvn_string_setter_function) smc_rsvn_set_drms_id;
	ret->super.set_submitted_ardl = (smc_rsvn_string_setter_function) smc_rsvn_set_submitted_ardl;
	ret->super.set_ardl = (smc_rsvn_string_setter_function) smc_rsvn_set_ardl;
	ret->super.set_ardl_parsed = (smc_rsvn_ardl_parsed_setter_function) smc_rsvn_set_ardl_parsed;
	ret->super.set_subject = (smc_rsvn_string_setter_function) smc_rsvn_set_subject;
	ret->super.set_mapped_user = (smc_rsvn_string_setter_function) smc_rsvn_set_mapped_user;
	ret->super.set_start_time = (smc_rsvn_time_setter_function) smc_rsvn_set_start_time;
	ret->super.set_end_time = (smc_rsvn_time_setter_function) smc_rsvn_set_end_time;

	ret->super.archive = (smc_rsvn_void_setter_function) smc_rsvn_archive;
	ret->super.terminate = (smc_rsvn_void_setter_function) smc_rsvn_terminate;

	ret->super.is_readonly = (smc_rsvn_bool_getter_function) smc_rsvn_is_readonly;
	ret->super.save = (smc_rsvn_save_function) smc_rsvn_save;

	ret->new_entry = new;
	ret->in_transaction = transaction;
	ret->accounting = accounting;
	ret->readonly = readonly;

	return (smc_rsvn_t) ret;
}

#define CHECK_THIS	SM_ASSERT(this, "You must supply 'this' pointer")

#define SET(field) \
	do { \
		this->modified = true; \
		this->field = value; \
		this->field##_set = true; \
		return this->field; \
	} while (0)

#define SET_STR(field) \
	do {  \
		this->modified = true; \
		SM_FREE(this->field); \
		this->field = sm_strdup(value); \
		if (!this->field) { \
			this->error = sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, "Out of memory"); \
			return NULL; \
		} \
		return this->field; \
	} while (0)

static const char *smc_rsvn_get_id(smc_concrete_rsvn_t this) { CHECK_THIS; smc_rsvn_gen_id(this); return this->id; }
static const char *smc_rsvn_get_drms_id(smc_concrete_rsvn_t this) { CHECK_THIS; return this->drms_id; }
static const char *smc_rsvn_get_submitted_ardl(smc_concrete_rsvn_t this) { CHECK_THIS; return this->submitted_ardl; }
static const char *smc_rsvn_get_ardl(smc_concrete_rsvn_t this) { CHECK_THIS; smc_rsvn_process_ardl(this); return this->ardl; }
static const char *smc_rsvn_get_subject(smc_concrete_rsvn_t this) { CHECK_THIS; return this->subject; }
static const char *smc_rsvn_get_mapped_user(smc_concrete_rsvn_t this) { CHECK_THIS; return this->mapped_user; }
static struct tm smc_rsvn_get_submission_time(smc_concrete_rsvn_t this) { CHECK_THIS; return this->submission_time; }
static struct tm smc_rsvn_get_start_time(smc_concrete_rsvn_t this) { CHECK_THIS; return this->start_time; }
static struct tm smc_rsvn_get_end_time(smc_concrete_rsvn_t this) { CHECK_THIS; return this->end_time; }

static const char *smc_rsvn_set_drms_id(smc_concrete_rsvn_t this, const char *value) { CHECK_THIS; SET_STR(drms_id); }
static const char *smc_rsvn_set_submitted_ardl(smc_concrete_rsvn_t this, const char *value) { CHECK_THIS; SET_STR(submitted_ardl); }
static const char *smc_rsvn_set_ardl(smc_concrete_rsvn_t this, const char *value) { CHECK_THIS; SET_STR(ardl); }
static const char *smc_rsvn_set_subject(smc_concrete_rsvn_t this, const char *value) { CHECK_THIS; SET_STR(subject); }
static const char *smc_rsvn_set_mapped_user(smc_concrete_rsvn_t this, const char *value) { CHECK_THIS; SET_STR(mapped_user); }
static struct tm smc_rsvn_set_start_time(smc_concrete_rsvn_t this, struct tm value) { CHECK_THIS; SET(start_time); }
static struct tm smc_rsvn_set_end_time(smc_concrete_rsvn_t this, struct tm value) { CHECK_THIS; SET(end_time); }

static void
smc_rsvn_gen_id(smc_concrete_rsvn_t this)
{
	if (this->id)
		return;

	this->id = smc_generate_id(NULL, NULL);

	if (!this->id)
		this->error = sm_ret_code_create_oom();
}

static bool
smc_rsvn_is_readonly(smc_concrete_rsvn_t this)
{
	CHECK_THIS;
	return this->readonly;
}

static _ardl__ReservationDefinition *
smc_rsvn_get_ardl_parsed(smc_concrete_rsvn_t this)
{
	_ardl__ReservationDefinition j;

	CHECK_THIS;

	if (this->ardl_parsed && !this->modified)
		return this->ardl_parsed;

	if (!this->ardl)
		return NULL;

	this->ardl_parsed = NULL;
	sm_alloc_tracked_free(&this->tracked);

	memset(&j, 0, sizeof(j));
	SM_TRY_NOT(this->tracked = smc_ardl_from_str(this->ardl, &j));
	return sm_alloc_tracked_memdup(this->tracked, &j, sizeof(j));

fault:
	return NULL;
}

static const char *
smc_rsvn_set_ardl_parsed(smc_concrete_rsvn_t this, _ardl__ReservationDefinition *value)
{
	CHECK_THIS;

	this->modified = true;

	SM_FREE(this->ardl);

	if (!(this->ardl = smc_ardl_str(value, NULL, NULL)))
		this->error = sm_ret_code_create_oom();

	return this->ardl;
}

static void
smc_rsvn_archive(smc_concrete_rsvn_t this)
{
	CHECK_THIS;

	if (!this->accounting) {
		this->modified = true;
		this->archive = true;
	}
}

static void
smc_rsvn_terminate(smc_concrete_rsvn_t this)
{
	CHECK_THIS;

	if (!this->accounting) {
		time_t start;
		time_t end;

		this->end_time = timetm();
		this->end_time_set = true;

		start = mktime(&this->start_time);
		end = mktime(&this->end_time);

		if (!(start <= end)) {
			/* The reservation couldn't have started yet. */
			memset(&this->start_time, 0, sizeof(struct tm));
			memset(&this->end_time, 0, sizeof(struct tm));
			this->start_time_set = false;
			this->end_time_set = false;
		}

		this->modified = true;
		this->archive = true;

		this->terminated = true;
	}
}

static sm_status_ret_code_t
smc_rsvn_read_from_db(smc_concrete_rsvn_t rsvn, sm_odbc_stmt_t stmt)
{
	struct tm db_time;
	sm_status_ret_code_t status;

	SM_MANDATORY_P(rsvn);
	SM_MANDATORY_P(stmt);

	SM_DEBUG_ENTER();

	memset(&db_time, 0, sizeof(db_time));

	SM_TRY_FAULT(stmt->get_string(stmt, 1, &rsvn->id));
	SM_TRY_FAULT(stmt->get_string(stmt, 2, &rsvn->drms_id));
	SM_TRY_FAULT(stmt->get_string(stmt, 3, &rsvn->submitted_ardl));
	SM_TRY_FAULT(stmt->get_string(stmt, 4, &rsvn->ardl));
	SM_TRY_FAULT(stmt->get_string(stmt, 5, &rsvn->subject));
	SM_TRY_FAULT(stmt->get_string(stmt, 6, &rsvn->mapped_user));

	SM_TRY_FAULT(status = stmt->get_timestamp(stmt, 7, &db_time));
	rsvn->submission_time = db_time;
	rsvn->submission_time_set = true;

	SM_TRY_FAULT(status = stmt->get_timestamp(stmt, 8, &db_time));
	rsvn->start_time = db_time;
	rsvn->start_time_set = true;

	SM_TRY_FAULT(status = stmt->get_timestamp(stmt, 9, &db_time));
	rsvn->end_time = db_time;
	rsvn->end_time_set = true;

	return SM_OK;

fault:
	return SM_FAULT;
}

#define SMC_DB_TRY_FAULT(value) \
	do { \
		if ((value) == SM_FAULT) \
			goto db_fault; \
	} while(0);

static sm_ret_code_t
smc_rsvn_get_internal(const char *id, smc_rsvn_t *result, bool readonly)
{
	sm_status_ret_code_t status;
	sm_odbc_stmt_t first_stmt = NULL;
	sm_odbc_stmt_t stmt = NULL;
	bool accounting;

	SM_MANDATORY_P(id);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER_STR(id);

	*result = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return sm_ret_code_create(SMC_RSVN_DATABASE_FAULT, SMC_RSVN_DATABASE_FAULT_STR);

	accounting = false;

	/* Try in reservations first. */
	if (readonly)
		stmt = find_readonly_stmt;
	else
		stmt = find_stmt;

	first_stmt = stmt;

	SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", id));
	SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

	/* Now try in reservations_accounting. */
	if (status == SM_NEG) {
		if (readonly)
			stmt = find_acc_readonly_stmt;
		else
			goto not_found;

		SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", id));
		SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

		if (status == SM_NEG)
			goto not_found;

		accounting = true;
	}

	SM_TRY_NOT(*result = smc_rsvn_create_internal(readonly, false, !readonly, accounting));

	SMC_DB_TRY_FAULT(smc_rsvn_read_from_db((smc_concrete_rsvn_t) *result, stmt));

	first_stmt->close_cursor(first_stmt);
	stmt->close_cursor(stmt);

	if (readonly)
		SM_TRY_FAULT(odbc->commit(odbc));

	return sm_ret_code_create(SMC_RSVN_OK, NULL);

not_found:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	return sm_ret_code_create(SMC_RSVN_NOT_FOUND, SMC_RSVN_NOT_FOUND_STR);

db_fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_rsvn_free(result);
	return sm_ret_code_create(SMC_RSVN_DATABASE_FAULT, SMC_RSVN_DATABASE_FAULT_STR);

fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_rsvn_free(result);
	return sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, SMC_RSVN_INTERNAL_FAULT_STR);
}

sm_ret_code_t
smc_rsvn_get(const char *id, smc_rsvn_t *result)
{
	SM_DEBUG_ENTER_STR(id);
	return smc_rsvn_get_internal(id, result, false);
}

sm_ret_code_t
smc_rsvn_get_readonly(const char *id, smc_rsvn_t *result)
{
	SM_DEBUG_ENTER_STR(id);
	return smc_rsvn_get_internal(id, result, true);
}

static sm_ret_code_t
smc_rsvn_get_by_drms_id_internal(const char *drms_id, smc_rsvn_t *result, bool readonly)
{
	sm_status_ret_code_t status;
	sm_odbc_stmt_t stmt = NULL;
	sm_odbc_stmt_t first_stmt = NULL;
	bool accounting;

	SM_MANDATORY_P(drms_id);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER_STR(drms_id);

	*result = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return sm_ret_code_create(SMC_RSVN_DATABASE_FAULT, SMC_RSVN_DATABASE_FAULT_STR);

	accounting = false;

	/* Try in reservations first. */
	if (readonly)
		stmt = find_by_drms_id_readonly_stmt;
	else
		stmt = find_by_drms_id_stmt;

	first_stmt = stmt;

	SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", drms_id));
	SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

	/* Now try in reservations_accounting. */
	if (status == SM_NEG) {
		if (readonly)
			stmt = find_acc_by_drms_id_readonly_stmt;
		else
			goto not_found;

		SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", drms_id));
		SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

		if (status == SM_NEG)
			goto not_found;

		accounting = true;
	}

	SM_TRY_NOT(*result = smc_rsvn_create_internal(readonly, false, !readonly, accounting));

	SMC_DB_TRY_FAULT(smc_rsvn_read_from_db((smc_concrete_rsvn_t) *result, stmt));

	first_stmt->close_cursor(first_stmt);
	stmt->close_cursor(stmt);

	if (readonly)
		SM_TRY_FAULT(odbc->commit(odbc));

	return sm_ret_code_create(SMC_RSVN_OK, NULL);

not_found:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	return sm_ret_code_create(SMC_RSVN_NOT_FOUND, SMC_RSVN_NOT_FOUND_STR);

db_fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_rsvn_free(result);
	return sm_ret_code_create(SMC_RSVN_DATABASE_FAULT, SMC_RSVN_DATABASE_FAULT_STR);

fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_rsvn_free(result);
	return sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, SMC_RSVN_INTERNAL_FAULT_STR);
}

sm_ret_code_t
smc_rsvn_get_by_drms_id_readonly(const char *drms_id, smc_rsvn_t *result)
{
	SM_DEBUG_ENTER_STR(drms_id);
	return smc_rsvn_get_by_drms_id_internal(drms_id, result, true);
}

sm_ret_code_t
smc_rsvn_get_by_drms_id(const char *drms_id, smc_rsvn_t *result)
{
	SM_DEBUG_ENTER_STR(drms_id);
	return smc_rsvn_get_by_drms_id_internal(drms_id, result, false);
}

sm_ret_code_t
smc_rsvn_get_active_readonly(sm_list_t *result)
{
	sm_status_ret_code_t status;
	sm_odbc_stmt_t stmt = NULL;
	smc_rsvn_t rsvn = NULL;

	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	*result = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return sm_ret_code_create(SMC_RSVN_DATABASE_FAULT, SMC_RSVN_DATABASE_FAULT_STR);

	stmt = find_all_readonly_stmt;

	SMC_DB_TRY_FAULT(stmt->execute(stmt, NULL));
	SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

	if (status == SM_OK) {

		for (;;) {
			SM_TRY_NOT(rsvn = smc_rsvn_create_internal(true, false, false, false));

			SMC_DB_TRY_FAULT(smc_rsvn_read_from_db((smc_concrete_rsvn_t) rsvn, stmt));

			SM_TRY_NOT(sm_list_add(result, rsvn, 0));
			rsvn = NULL;

			SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));
			if (status == SM_NEG)
				break;
		}
	}

	stmt->close_cursor(stmt);

	SM_TRY_FAULT(odbc->commit(odbc));

	return sm_ret_code_create(SMC_RSVN_OK, NULL);

db_fault:
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_rsvn_free(&rsvn);
	sm_list_free_user(*result, (sm_list_free_user_function) smc_rsvn_free);
	return sm_ret_code_create(SMC_RSVN_DATABASE_FAULT, SMC_RSVN_DATABASE_FAULT_STR);

fault:
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_rsvn_free(&rsvn);
	sm_list_free_user(*result, (sm_list_free_user_function) smc_rsvn_free);
	return sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, SMC_RSVN_INTERNAL_FAULT_STR);
}

void
smc_rsvn_free(smc_rsvn_t *rsvn)
{
	smc_concrete_rsvn_t concrete;

	SM_MANDATORY_P(rsvn);

	SM_DEBUG_ENTER();

	concrete = (smc_concrete_rsvn_t) *rsvn;

	if (!concrete)
		return;

	if (concrete->in_transaction) {
		odbc->rollback(odbc);
		concrete->in_transaction = false;
	}

	SM_FREE(concrete->id);
	SM_FREE(concrete->drms_id);
	SM_FREE(concrete->submitted_ardl);
	SM_FREE(concrete->ardl);
	SM_FREE(concrete->subject);
	SM_FREE(concrete->mapped_user);

	sm_alloc_tracked_free(&concrete->tracked);

	SM_FREE(*rsvn);
}


#define FIELD2DB(field)	((field##_set) ? &(field) : NULL)

static struct tm
timetm(void)
{
	time_t now = time(NULL);
	struct tm nowtm;

	localtime_r(&now, &nowtm);
	return nowtm;
}

static sm_status_ret_code_t
smc_rsvn_insert(smc_concrete_rsvn_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(!this->in_transaction, "Reservation object inconsistent state");

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(insert_stmt->execute(insert_stmt, "ssssssttt",
			this->id,
			this->drms_id,
			this->submitted_ardl,
			this->ardl,
			this->subject,
			this->mapped_user,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time)));

	SM_TRY_FAULT(odbc->commit(odbc));
	return SM_OK;

fault:
	odbc->rollback(odbc);
	return SM_FAULT;
}

static sm_status_ret_code_t
smc_rsvn_update(smc_concrete_rsvn_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(this->in_transaction, "Reservation object inconsistent state");

	SM_TRY_FAULT(update_stmt->execute(update_stmt, "ssssssttts",
			this->id,
			this->drms_id,
			this->submitted_ardl,
			this->ardl,
			this->subject,
			this->mapped_user,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time),
			this->id));

	SM_TRY_FAULT(odbc->commit(odbc));
	this->in_transaction = false;
	return SM_OK;

fault:
	odbc->rollback(odbc);
	this->in_transaction = false;
	return SM_FAULT;
}

static sm_status_ret_code_t
smc_rsvn_move_to_acc(smc_concrete_rsvn_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(this->in_transaction, "Reservation object inconsistent state");

	SM_TRY_FAULT(delete_stmt->execute(delete_stmt, "s", this->id));

	SM_TRY_FAULT(insert_acc_stmt->execute(insert_acc_stmt, "ssssssttt",
			this->id,
			this->drms_id,
			this->submitted_ardl,
			this->ardl,
			this->subject,
			this->mapped_user,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time)));

	SM_TRY_FAULT(odbc->commit(odbc));
	this->in_transaction = false;
	return SM_OK;

fault:
	odbc->rollback(odbc);
	this->in_transaction = false;
	return SM_FAULT;
}

#define READ_COMPONENT(str, component, designator) \
	do { \
		char *__p; \
		if ((__p = strchr(str, designator))) { \
			char *__tmp = __p; \
			unsigned int __value = 0; \
			\
			__tmp--; \
			while (*__tmp && isdigit(*__tmp)) \
				__tmp--; \
			__tmp++; \
			if (sscanf(__tmp, "%u", &__value) == 1) \
				component = __value; \
		} \
	} while (0)

static int
smc_rsvn_parse_duration(char *str)
{
	int years = 0, months = 0, days = 0, hours = 0, minutes = 0, seconds = 0;
	int duration = -1;
	char *p = NULL;

	SM_DEBUG_ENTER_STR(str);

	if (!str || str[0] == '-' || str[0] != 'P')
		return -1;

	if ((p = strchr(str, '.'))) {
		p[0] = 'S';
		p[1] = '\0';
	}

	if ((p = strchr(str, 'T'))) {
		*p = '\0';
		p++;
	}

	/* TODO: It's not that easy to add months and years. We need to know the
	 * start point to perform the addition. See XML Schema specification for
	 * the algorithm. */
	if (strchr(str, 'Y') || strchr(str, 'M'))
		return -1;

	READ_COMPONENT(str, years, 'Y');
	READ_COMPONENT(str, months, 'M');
	READ_COMPONENT(str, days, 'D');

	if (p) {
		READ_COMPONENT(p, hours, 'H');
		READ_COMPONENT(p, minutes, 'M');
		READ_COMPONENT(p, seconds, 'S');
	}

	sm_debug("days = %d, hours = %d, minutes = %d, seconds = %d", days, hours, minutes, seconds);

	duration = seconds;
	duration += minutes * 60;
	duration += hours * 3600;
	duration += days * 86400;

	return duration;
}

static sm_ret_code_t
smc_rsvn_get_times(smc_concrete_rsvn_t this, _ardl__ReservationDefinition *ardl)
{
	struct ardl__ARDLTimeWindow_USCOREType *tw;
	time_t start = 0;
	time_t end = 0;

	SM_DEBUG_ENTER();

	tw = ardl->ardl__ReservationDescription->TimeWindow;
	assert(tw);

	if (!this->start_time_set) {
		if (tw->StartTime) {
			/* Ignore seconds -- round down. */
			*tw->StartTime /= 60;
			*tw->StartTime *= 60;

			sm_debug("StartTime = %d", (int) *tw->StartTime);
			localtime_r(tw->StartTime, &this->start_time);
			start = *tw->StartTime;
		} else {
			time_t now = time(NULL);
			time_t t;

			t = now;
			t /= 60;
			t *= 60;

			/* Round up to the next whole minute. */
			t += 60;

			/* But make sure there's at least RESERVATION_TIME_SHIFT
			 * time left until the start. */
			if (t - now < RESERVATION_TIME_SHIFT)
				t += 60;

			assert(t - now >= RESERVATION_TIME_SHIFT);

			localtime_r(&t, &this->start_time);
			start = t;
		}

		this->start_time_set = true;
	}

	if (!this->end_time_set) {
		if (tw->__union_ARDLTimeWindow_USCOREType == SOAP_UNION__ardl__union_ARDLTimeWindow_USCOREType_EndTime) {
			SM_BOOL rest = ((tw->union_ARDLTimeWindow_USCOREType.EndTime  % 60) != 0);
			/* Ignore seconds -- round up. Rationale: Created reservation must not be shorter than user expects */
			tw->union_ARDLTimeWindow_USCOREType.EndTime /= 60;
			tw->union_ARDLTimeWindow_USCOREType.EndTime *= 60;

			if (rest)
				tw->union_ARDLTimeWindow_USCOREType.EndTime +=60;

			sm_debug("EndTime = %d", (int) tw->union_ARDLTimeWindow_USCOREType.EndTime);
			localtime_r(&tw->union_ARDLTimeWindow_USCOREType.EndTime, &this->end_time);
		} else {
			int duration;

			duration = smc_rsvn_parse_duration(tw->union_ARDLTimeWindow_USCOREType.Duration);

			if (duration == -1)
				return sm_ret_code_create(SMC_RSVN_INVALID_ARGUMENT, "Invalid Duration element");

			sm_debug("Duration parsed as %d seconds", (int) duration);
			end = start + duration;
			localtime_r(&end, &this->end_time);
		}

		this->end_time_set = true;
	}

	/* All reservations must last for at least 1 minute. */
	if (mktime(&this->end_time) - mktime(&this->start_time) < 60) {
		time_t t;

		sm_debug("Reservation shorter than 1 minute, setting duration to 1 minute.");

		t = mktime(&this->start_time) + 60;
		localtime_r(&t, &this->end_time);
	}

	return sm_ret_code_create_ok();
}

static sm_ret_code_t
smc_rsvn_process_ardl(smc_concrete_rsvn_t this)
{
	sm_alloc_tracked_t tracked = NULL;
	_ardl__ReservationDefinition ardl;
	sm_ret_code_t ret;
	struct ardl__ARDLTimeWindow_USCOREType *tw;

	SM_DEBUG_ENTER_STR(this->id);

	memset(&ret, 0, sizeof(ret));

	if (!this->ardl)
		return sm_ret_code_create(SMC_RSVN_OK, NULL);

	if (!this->modified)
		return sm_ret_code_create(SMC_RSVN_OK, NULL);

	if (!(tracked = smc_ardl_from_str(this->ardl, &ardl)))
		return sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, "Failed to parse ARDL document");

	if (this->new_entry) {
		ret = smc_rsvn_get_times(this, &ardl);
		if (ret.code)
			return ret;
	}

	/* Use this->mapped_user to set UserName if not already set. */
	if (this->mapped_user) {
		if (!ardl.ardl__ReservationDescription->ardl__Resources)
			SM_TRY_NOT(ardl.ardl__ReservationDescription->ardl__Resources = sm_alloc_tracked_malloc(tracked, sizeof(struct ardl__ARDLResources_USCOREType)));

		if (!ardl.ardl__ReservationDescription->ardl__Resources->UserName)
			SM_TRY_NOT(ardl.ardl__ReservationDescription->ardl__Resources->UserName = sm_alloc_tracked_strdup(tracked, this->mapped_user));
	}

	/* We always may update start/end time so modify ARDL accordingly. */
	if (this->start_time_set && this->end_time_set) {
		tw = ardl.ardl__ReservationDescription->TimeWindow;
		tw->__union_ARDLTimeWindow_USCOREType = SOAP_UNION__ardl__union_ARDLTimeWindow_USCOREType_EndTime;
		SM_TRY_NOT(tw->StartTime = sm_alloc_tracked_malloc(tracked, sizeof(time_t)));

		*tw->StartTime = mktime(&this->start_time);
		tw->union_ARDLTimeWindow_USCOREType.EndTime = mktime(&this->end_time);
	}

	SM_FREE(this->ardl);
	SM_TRY_NOT(this->ardl = smc_ardl_str(&ardl, NULL, NULL));

	sm_alloc_tracked_free(&tracked);
	return sm_ret_code_create(SMC_RSVN_OK, NULL);

fault:
	sm_alloc_tracked_free(&tracked);
	if (ret.code)
		return ret;
	else
		return sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, SMC_RSVN_INTERNAL_FAULT_STR);
}

sm_ret_code_t
smc_rsvn_save(smc_concrete_rsvn_t this)
{
	sm_status_ret_code_t (*save_func)(smc_concrete_rsvn_t) = NULL;
	sm_ret_code_t ret;

	CHECK_THIS;

	SM_DEBUG_ENTER_STR(this->id);

	if (this->readonly)
		SM_ASSERT(false, "You cannot save an object that is read only");

	if (this->error.code)
		return this->error;

	if (!this->modified) {
		if (this->in_transaction) {
			odbc->commit(odbc);
			this->in_transaction = false;
		}
		return sm_ret_code_create(SMC_RSVN_OK, NULL);
	}

	if (this->new_entry) {
		this->submission_time = timetm();
		this->submission_time_set = true;

		if (!this->id) {
			smc_rsvn_gen_id(this);

			if (this->error.code)
				return this->error;
		}
	}

	ret = smc_rsvn_process_ardl(this);
	if (ret.code)
		return ret;

	if (this->new_entry) {
		save_func = smc_rsvn_insert;
	} else {
		if (this->archive)
			save_func = smc_rsvn_move_to_acc;
		else
			save_func = smc_rsvn_update;
	}

	if (save_func(this) != SM_OK)
		return sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, "Failed to save rsvn object to database");

	if (this->new_entry) {
		sm_signal_emit("smc_rsvn_created", this->id, this->drms_id,
			this->submitted_ardl, this->ardl, this->subject,
			this->mapped_user, mktime(&this->submission_time),
			mktime(&this->start_time), mktime(&this->end_time), NULL);
	} else if (this->terminated) {
		sm_signal_emit("smc_rsvn_terminated", this->id, this->drms_id,
			this->submitted_ardl, this->ardl, this->subject,
			this->mapped_user, mktime(&this->submission_time),
			mktime(&this->start_time), mktime(&this->end_time), NULL);
	}

	this->modified = false;
	this->error = sm_ret_code_create(SMC_RSVN_INTERNAL_FAULT, "Cannot save the same rsvn object twice. You need to fetch a new one.");

	return sm_ret_code_create(SMC_RSVN_OK, NULL);
}

int
smc_rsvn_count(void)
{
	double result = 0;

	SM_DEBUG_ENTER();

	odbc->begin_transaction(odbc);

	SM_TRY_NOT_OK(count_stmt->execute(count_stmt, NULL));
	SM_TRY_NOT_OK(count_stmt->fetch_row(count_stmt));
	SM_TRY_NOT_OK(count_stmt->get_double(count_stmt, 1, &result));

	count_stmt->close_cursor(count_stmt);
	odbc->commit(odbc);

	SM_ASSERT(result <= INT_MAX, "smc_rsvn_count needs refactoring ;-)");

	return (int) result;

fault:
	count_stmt->close_cursor(count_stmt);
	odbc->rollback(odbc);
	return -1;
}

#if defined(TEST) && !defined(RSVN_NO_TEST)
static const char *ardl_st_end =
"\
<ReservationDefinition xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/ardl\">\
<ReservationDescription>\
<ReservationIdentification>\
<ReservationName>SimpleReservation</ReservationName>\
</ReservationIdentification>\
<TimeWindow>\
<StartTime>2012-08-26T21:00:00+02:00</StartTime>\
<EndTime>2012-08-26T23:00:00+02:00</EndTime>\
</TimeWindow>\
<Resources>\
<ReservedSlotsCount>1</ReservedSlotsCount>\
</Resources>\
</ReservationDescription>\
</ReservationDefinition>\
";

static const char *ardl_just_end =
"\
<ReservationDefinition xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/ardl\">\
<ReservationDescription>\
<ReservationIdentification>\
<ReservationName>SimpleReservation</ReservationName>\
</ReservationIdentification>\
<TimeWindow>\
<EndTime>2012-08-26T23:00:00+02:00</EndTime>\
</TimeWindow>\
<Resources>\
<ReservedSlotsCount>1</ReservedSlotsCount>\
</Resources>\
</ReservationDescription>\
</ReservationDefinition>\
";

static const char *ardl_st_duration =
"\
<ReservationDefinition xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/ardl\">\
<ReservationDescription>\
<ReservationIdentification>\
<ReservationName>SimpleReservation</ReservationName>\
</ReservationIdentification>\
<TimeWindow>\
<StartTime>2012-08-26T21:00:00+02:00</StartTime>\
<Duration>PT02H00M00S</Duration>\
</TimeWindow>\
<Resources>\
<ReservedSlotsCount>1</ReservedSlotsCount>\
</Resources>\
</ReservationDescription>\
</ReservationDefinition>\
";

static const char *ardl_just_duration =
"\
<ReservationDefinition xmlns=\"http://schemas.qoscosgrid.org/comp/2011/04/ardl\">\
<ReservationDescription>\
<ReservationIdentification>\
<ReservationName>SimpleReservation</ReservationName>\
</ReservationIdentification>\
<TimeWindow>\
<Duration>PT02H00M00S</Duration>\
</TimeWindow>\
<Resources>\
<ReservedSlotsCount>1</ReservedSlotsCount>\
</Resources>\
</ReservationDescription>\
</ReservationDefinition>\
";

static void
test_prepare(void)
{
	sm_odbc_stmt_t stmt;

	SM_DEBUG_ENTER();

	assert(odbc->begin_transaction(odbc) == SM_OK);

	stmt = sm_odbc_stmt_create(odbc, "DELETE FROM reservations");
	assert(stmt);
	assert(stmt->execute(stmt, NULL) == SM_OK);
	sm_odbc_stmt_free(&stmt);

	stmt = sm_odbc_stmt_create(odbc, "DELETE FROM reservations_acc");
	assert(stmt);
	assert(stmt->execute(stmt, NULL) == SM_OK);
	sm_odbc_stmt_free(&stmt);

	assert(odbc->commit(odbc) == SM_OK);
}

static char *
add_rsvn(const char *ardl)
{
	static int i = 0;
	char buf[64];
	sm_ret_code_t ret;
	smc_rsvn_t rsvn;
	char *id;

	SM_MANDATORY_P(ardl);

	SM_DEBUG_ENTER_STR(ardl);

	i++;
	snprintf(buf, sizeof(buf), "%d", i);

	rsvn = smc_rsvn_create();
	rsvn->set_drms_id(rsvn, buf);
	rsvn->set_subject(rsvn, "subject");
	rsvn->set_mapped_user(rsvn, "user");
	rsvn->set_submitted_ardl(rsvn, ardl);
	rsvn->set_ardl(rsvn, ardl);

	id = sm_strdup(rsvn->get_id(rsvn));
	assert(id);

	ret = rsvn->save(rsvn);
	if (ret.code)
		sm_error("%s", ret.string);
	assert(!ret.code);
	smc_rsvn_free(&rsvn);

	return id;
}

static smc_rsvn_t
get_rsvn(char *id)
{
	sm_ret_code_t ret;
	smc_rsvn_t rsvn;

	ret = smc_rsvn_get(id, &rsvn);
	if (ret.code)
		sm_error("%s", ret.string);
	assert(!ret.code);

	free(id);
	return rsvn;
}

#define CHECK_TIME(type, value) \
	do { \
		got = rsvn->get_##type##_time(rsvn); \
		strftime(tbuf, sizeof(tbuf), "%F %H:%M %Z", &got); \
		sm_debug("Got from db: %s", tbuf); \
		memset(&expected, 0, sizeof(expected)); \
		strptime(value, "%Y-%m-%d %T", &expected); \
		expected.tm_isdst = -1; \
		sm_debug("got = %d, expected = %d", (int) mktime(&got), (int) mktime(&expected)); \
		assert(mktime(&got) == mktime(&expected)); \
	} while (0)

#define CHECK_NOW_TIME(type, plus) \
	do { \
		time_t now = time(NULL); \
		int val; \
		got = rsvn->get_##type##_time(rsvn); \
		strftime(tbuf, sizeof(tbuf), "%F %H:%M %Z", &got); \
		sm_debug("Got from db: %s", tbuf); \
		sm_debug("got = %d, expected = %d", (int) mktime(&got), (int) now); \
		val = mktime(&got) - (now + RESERVATION_TIME_SHIFT + plus); \
		if (val < 0) val = -val; \
		assert(val <= 1); \
	} while (0)

static void
test_simple_times(void)
{
	char *id;
	struct tm got;
	struct tm expected;
	char tbuf[256];
	smc_rsvn_t rsvn;

	SM_DEBUG_ENTER();

	id = add_rsvn(ardl_st_end);
	rsvn = get_rsvn(id);
	CHECK_TIME(start, "2012-08-26 21:00:00");
	CHECK_TIME(end, "2012-08-26 23:00:00");
	smc_rsvn_free(&rsvn);

	id = add_rsvn(ardl_just_end);
	rsvn = get_rsvn(id);
	/*
	CHECK_NOW_TIME(start, 0);
	*/
	CHECK_TIME(end, "2012-08-26 23:00:00");
	smc_rsvn_free(&rsvn);

	id = add_rsvn(ardl_st_duration);
	rsvn = get_rsvn(id);
	CHECK_TIME(start, "2012-08-26 21:00:00");
	CHECK_TIME(end, "2012-08-26 23:00:00");
	smc_rsvn_free(&rsvn);

	id = add_rsvn(ardl_just_duration);
	rsvn = get_rsvn(id);
	/*
	CHECK_NOW_TIME(start, 0);
	CHECK_NOW_TIME(end, 7200);
	*/
	smc_rsvn_free(&rsvn);
}

static void
test_accounting(void)
{
	smc_rsvn_t rsvn;
	char drms_id[128];
	sm_ret_code_t ret;
	int i;

	SM_DEBUG_ENTER();

	for (i = 0; i < 10; i++) {
		rsvn = smc_rsvn_create();

		snprintf(drms_id, sizeof(drms_id), "%d", i);

		rsvn->set_drms_id(rsvn, drms_id);
		rsvn->set_submitted_ardl(rsvn, ardl_st_end);
		rsvn->set_ardl(rsvn, ardl_st_end);
		rsvn->set_subject(rsvn, "subject");
		rsvn->set_mapped_user(rsvn, "user");

		rsvn->save(rsvn);
		smc_rsvn_free(&rsvn);
	}

	for (i = 0; i < 10; i++) {
		sm_list_t result = NULL;

		snprintf(drms_id, sizeof(drms_id), "%d", i);

		assert(smc_rsvn_count() == 10 - i);
		smc_rsvn_get_active_readonly(&result);
		assert(sm_list_count(result) == (size_t) (10 - i));
		SM_LIST_FREE_USER(result, smc_rsvn_free);

		ret = smc_rsvn_get_by_drms_id(drms_id, &rsvn);
		assert(!ret.code);

		rsvn->archive(rsvn);

		rsvn->save(rsvn);
		smc_rsvn_free(&rsvn);

		/* Cannot get the rsvn for modification once it has been put into accounting table. */
		ret = smc_rsvn_get_by_drms_id(drms_id, &rsvn);
		assert(ret.code == SMC_RSVN_NOT_FOUND);

		/* Getting the rsvn object for reading should work on the other hand. */
		ret = smc_rsvn_get_by_drms_id_readonly(drms_id, &rsvn);
		assert(!ret.code);

		smc_rsvn_free(&rsvn);
	}
}

static void
test_duration(void)
{
	SM_DEBUG_ENTER();

	assert(smc_rsvn_parse_duration(sm_strdup("P0D")) == 0);
	assert(smc_rsvn_parse_duration(sm_strdup("P10D")) == 86400 * 10);
	assert(smc_rsvn_parse_duration(sm_strdup("PT2H")) == 3600 * 2);
	assert(smc_rsvn_parse_duration(sm_strdup("PT02H")) == 3600 * 2);
	assert(smc_rsvn_parse_duration(sm_strdup("PT00H01M10S")) == 70);
	assert(smc_rsvn_parse_duration(sm_strdup("PT00S")) == 0);
	assert(smc_rsvn_parse_duration(sm_strdup("PT70S")) == 70);
	assert(smc_rsvn_parse_duration(sm_strdup("PT1H70S")) == 3600 + 70);
	assert(smc_rsvn_parse_duration(sm_strdup("PT1M70S")) == 60 + 70);
	assert(smc_rsvn_parse_duration(sm_strdup("PT01M70S")) == 60 + 70);
}

int
main()
{
	sm_odbc_handle_t db;

	assert(sm_library_setup() == SM_OK);
	assert(sm_log_init_str(NULL, "DEBUG") == SM_OK);

	db = sm_odbc_connect("qcg_comp_test", "qcg_comp_test", "qcg_comp_test");
	if (!db) {
		sm_notice("Could not connect to database. Prepare the test database first.");
		sleep(1);
		return 0;
	}

	test_duration();

	smc_rsvn_setup(db);

	test_prepare();
	test_simple_times();

	test_prepare();
	test_accounting();

	smc_rsvn_cleanup();
	sm_odbc_disconnect(&db);
	sm_library_cleanup();

	return 0;
}
#endif
