/* $Id: staging.c 45 2009-12-30 16:24:35Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <unistd.h>

#include <qcg-core/sm.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/thread.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <job.h>
#include <jsdl.h>
#include <session_process.h>
#include <staging.h>
#include <include/staging_module.h>
#include <stub/qcgcompH.h>
#include <stub/qcgcompStub.h>

__SM_ID("$Id: staging.c 45 2009-12-30 16:24:35Z mamonski $");

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "staging"
        }
};

extern struct Namespace qcgcomp_namespaces[];

struct stager_data {
	const char *id;
	const char *reservation_id;
	unsigned int reservation_duration;
	sm_module_manager_t module_manager;
};

static void
set_staging_fault(smc_job_t job, staging_ret_code_t code, const char *descr)
{
	struct SOAP_ENV__Fault *f = NULL;
	char *buf = NULL;

	SM_MANDATORY_P(job);

	SM_DEBUG_ENTER();

	if (job->get_staging_fault(job)) {
		sm_debug("Staging fault already set");
		return;
	}

	SM_TRY_NOT(f = sm_malloc(sizeof(*f)));
	soap_default_SOAP_ENV__Fault(NULL, f);

	SM_TRY_NOT(f->detail = sm_malloc(sizeof(struct SOAP_ENV__Detail)));
	soap_default_SOAP_ENV__Detail(NULL, f->detail);

	switch (code) {
		case STAGING_INTERNAL_FAULT:
		case STAGING_DELETE_ON_TERMINATION_FAULT:
		case STAGING_LOCAL_STAGING_FAULT:
		case STAGING_REMOTE_STAGING_FAULT:
			f->faultcode = "SOAP-ENV:Server";
			break;
		case STAGING_NOT_AUTHORIZED_FAULT:
		case STAGING_FILE_NOT_FOUND_FAULT:
		case STAGING_UNSUPPORTED_PROTOCOL_FAULT:
			f->faultcode = "SOAP-ENV:Client";
			break;
		default:
			SM_ASSERT(0, "Wrong code %d in %s", code, __FUNCTION__);
			break;
	}

	f->faultstring = "HPC File Staging Profile fault";

	switch (code) {
		case STAGING_INTERNAL_FAULT:
			if (descr)
				f->faultstring = (char *) descr;
			else
				f->faultstring = "Internal fault";
			f->detail = NULL;
			break;
		case STAGING_NOT_AUTHORIZED_FAULT:
		{
			_bes_factory__NotAuthorizedFault *d;

			SM_TRY_NOT(d = sm_malloc(sizeof(*d)));

			f->detail->__type = SOAP_TYPE__bes_factory__NotAuthorizedFault;
			f->detail->fault = d;
			break;
		}
		case STAGING_DELETE_ON_TERMINATION_FAULT:
		{
			_hpcp_fs__DeleteOnTerminationFault *d;

			SM_TRY_NOT(d = sm_malloc(sizeof(*d)));

			SM_MANDATORY_P(descr);

			d->File = (char *) descr;
			f->detail->__type = SOAP_TYPE__hpcp_fs__DeleteOnTerminationFault;
			f->detail->fault = d;
			break;
		}
		case STAGING_FILE_NOT_FOUND_FAULT:
		{
			_hpcp_fs__FileNotFoundFault *d;

			SM_TRY_NOT(d = sm_malloc(sizeof(*d)));

			SM_MANDATORY_P(descr);

			d->__item = (char *) descr;
			f->detail->__type = SOAP_TYPE__hpcp_fs__FileNotFoundFault;
			f->detail->fault = d;
			break;
		}
		case STAGING_LOCAL_STAGING_FAULT:
		{
			_hpcp_fs__LocalStagingFault *d;

			SM_TRY_NOT(d = sm_malloc(sizeof(*d)));

			SM_MANDATORY_P(descr);

			d->__item = (char *) descr;
			f->detail->__type = SOAP_TYPE__hpcp_fs__LocalStagingFault;
			f->detail->fault = d;
			break;
		}
		case STAGING_REMOTE_STAGING_FAULT:
		{
			_hpcp_fs__RemoteStagingFault *d;

			SM_TRY_NOT(d = sm_malloc(sizeof(*d)));

			SM_MANDATORY_P(descr);

			d->__item = (char *) descr;
			f->detail->__type = SOAP_TYPE__hpcp_fs__RemoteStagingFault;
			f->detail->fault = d;
			break;
		}
		case STAGING_UNSUPPORTED_PROTOCOL_FAULT:
		{
			_hpcp_fs__UnsupportedProtocolFault *d;

			SM_TRY_NOT(d = sm_malloc(sizeof(*d)));

			SM_MANDATORY_P(descr);

			d->File = (char *) descr;
			f->detail->__type = SOAP_TYPE__hpcp_fs__UnsupportedProtocolFault;
			f->detail->fault = d;
			break;
		}
		default:
			SM_ASSERT(0, "Unknown error code (%d) in %s", code, __FUNCTION__);
	}

	buf = sm_gsoap_buffer_put(f, SOAP_ENV__Fault, "SOAP-ENV:Fault", qcgcomp_namespaces);
	if (buf)
		job->set_staging_fault(job, buf);

fault:
	SM_FREE(buf);

	if (f) {
		if (f->detail) {
			SM_FREE(f->detail->fault);
			SM_FREE(f->detail);
		}

		SM_FREE(f);
	}
}

sm_status_ret_code_t
smc_staging_delete_on_termination(smc_job_t job, bool quiet)
{
	int i;
	sm_status_ret_code_t ret;
	_jsdl__JobDefinition *jsdl;

	SM_MANDATORY_P(job);

	SM_DEBUG_ENTER();

	if (!(jsdl = job->get_jsdl_parsed(job)))
		return SM_FAULT;

	ret = SM_OK;

	for (i = 0; i < jsdl->jsdl__JobDescription->__sizeDataStaging; i++) {
		_jsdl__DataStaging *data_staging = jsdl->jsdl__JobDescription->jsdl__DataStaging + i;

		if (data_staging->jsdl__DeleteOnTermination && *data_staging->jsdl__DeleteOnTermination) {
			char buf[PATH_MAX];
			const char *wd = smc_jsdl_get_working_dir(jsdl, SM_FALSE);

			/* We always set WorkingDirectory but this may be a
			 * staging-only activity with no Application element. */
			if (wd)
				snprintf(buf, sizeof(buf), "%s/%s", wd, data_staging->jsdl__FileName);
			else
				snprintf(buf, sizeof(buf), "%s", data_staging->jsdl__FileName);

			if (unlink(buf) && !quiet) {
				if (errno != ENOENT) {
					char errbuf[1024];

					sm_strerror_r(errno, errbuf, sizeof(errbuf));
					sm_notice("Failed to delete file %s: %s", buf, errbuf);
					set_staging_fault(job, STAGING_DELETE_ON_TERMINATION_FAULT, buf);

					ret = SM_FAULT;

					/* Try to continue with deleting. */
				}
			}
		}
	}

	return ret;
}

static sm_status_ret_code_t
stager_in_helper(smc_job_t job, sm_module_manager_t module_manager)
{
	_jsdl__JobDefinition *jsdl = NULL;
	sm_list_t modules = NULL;
	sm_list_t l;
	sm_value_t *result = NULL;
	const char *working_dir;
	int i;

	SM_MANDATORY_P(job);

	SM_DEBUG_ENTER();

	modules = module_manager->find_by_api(module_manager, STAGING_MODULE_API);

	SM_TRY_NOT(jsdl = job->get_jsdl_parsed(job));
	working_dir = smc_jsdl_get_working_dir(jsdl, SM_TRUE);

	for (l = modules; l; l = l->next) {
		staging_module_t m = l->data;

		if (!(result = m->stage_in((sm_module_t) m, jsdl->jsdl__JobDescription->jsdl__DataStaging, jsdl->jsdl__JobDescription->__sizeDataStaging, working_dir)))
			goto fault;

		if (SM_VALUE_RET_CODE_C(result[0])) {
			set_staging_fault(job, SM_VALUE_RET_CODE_C(result[0]), SM_VALUE_RET_CODE_STRING(result[0]));
			goto fault;
		}

		sm_value_free(&result);
	}

	for (i = 0; i < jsdl->jsdl__JobDescription->__sizeDataStaging; i++) {
		_jsdl__DataStaging *ds;

		ds = jsdl->jsdl__JobDescription->jsdl__DataStaging + i;

		if (!ds->jsdl__Source || !ds->jsdl__Source->jsdl__URI)
			continue;

		if (ds->jsdl__FileName && (strstr(ds->jsdl__Source->jsdl__URI,"cid:") != ds->jsdl__Source->jsdl__URI)) {
			sm_notice("Unsupported protocol: %s", ds->jsdl__Source->jsdl__URI);
			set_staging_fault(job, STAGING_UNSUPPORTED_PROTOCOL_FAULT, ds->jsdl__Source->jsdl__URI);
			goto fault;
		}
	}

	sm_list_free(modules, SM_FALSE);
	return SM_OK;

fault:
	sm_value_free(&result);
	sm_list_free(modules, SM_FALSE);
	return SM_FAULT;
}

static sm_status_ret_code_t
stager_out_helper(smc_job_t job, sm_module_manager_t module_manager)
{
	_jsdl__JobDefinition *jsdl = NULL;
	sm_list_t modules = NULL;
	sm_list_t l;
	sm_value_t *result = NULL;
	const char *working_dir;
	int i;

	SM_MANDATORY_P(job);

	SM_DEBUG_ENTER();

	modules = module_manager->find_by_api(module_manager, STAGING_MODULE_API);

	SM_TRY_NOT(jsdl = job->get_jsdl_parsed(job));
	working_dir = smc_jsdl_get_working_dir(jsdl, SM_FALSE);

	for (l = modules; l; l = l->next) {
		staging_module_t m = l->data;

		if (!(result = m->stage_out((sm_module_t) m, jsdl->jsdl__JobDescription->jsdl__DataStaging, jsdl->jsdl__JobDescription->__sizeDataStaging, working_dir)))
			goto fault;

		if (SM_VALUE_RET_CODE_C(result[0])) {
			set_staging_fault(job, SM_VALUE_RET_CODE_C(result[0]), SM_VALUE_RET_CODE_STRING(result[0]));
			goto fault;
		}

		sm_value_free(&result);
	}

	for (i = 0; i < jsdl->jsdl__JobDescription->__sizeDataStaging; i++) {
		_jsdl__DataStaging *ds;

		ds = jsdl->jsdl__JobDescription->jsdl__DataStaging + i;

		if (!ds->jsdl__Target || !ds->jsdl__Target->jsdl__URI)
			continue;

		if (ds->jsdl__FileName && (strstr(ds->jsdl__Target->jsdl__URI,"cid:") != ds->jsdl__Target->jsdl__URI)) {
			sm_notice("Unsupported protocol: %s", ds->jsdl__Target->jsdl__URI);
			set_staging_fault(job, STAGING_UNSUPPORTED_PROTOCOL_FAULT, ds->jsdl__Target->jsdl__URI);
			goto fault;
		}
	}

	sm_list_free(modules, SM_FALSE);
	return SM_OK;

fault:
	sm_value_free(&result);
	sm_list_free(modules, SM_FALSE);
	return SM_FAULT;
}

#define GET_JOB() \
	do { \
		ret = smc_job_get(data->id, &jobs); \
		if (ret.code) { \
			sm_error("Couldn't retrieve job %s from database (%d:%s)", data->id, ret.code, ret.string); \
			goto fault; \
		} \
		SM_ASSERT(sm_list_count(jobs) <= 1, "Parameter sweep not implemented in %s", __FUNCTION__); \
	} while (0)

#define GET_JOB_READONLY() \
	do { \
		ret = smc_job_get_readonly(data->id, &jobs); \
		if (ret.code) { \
			sm_error("Couldn't retrieve job %s from database (%d:%s)", data->id, ret.code, ret.string); \
			goto fault; \
		} \
		SM_ASSERT(sm_list_count(jobs) <= 1, "Parameter sweep not implemented in %s", __FUNCTION__); \
	} while (0)

#define SAVE_JOB() \
	do { \
		ret = job->save(job); \
		if (ret.code) { \
			sm_error("Failed to save job in database: %d:%s", ret.code, ret.string); \
			goto fault; \
		} \
	} while (0)

#define FREE_STAGER_DATA() \
	do { \
		if (data) { \
			SM_FREE(data->reservation_id); \
			SM_FREE(data->id); \
			SM_FREE(data); \
		} \
	} while (0)

static void
try_mark_as_failed(smc_job_t job)
{
	sm_list_t jobs = NULL;
	sm_ret_code_t ret;

	if (!job)
		return;

	SM_DEBUG_ENTER_STR(job->get_id(job));

	if (!SMC_JOB_STATE_TERMINAL(job->get_state(job)))
		smc_staging_delete_on_termination(job, true);

	if (!job->is_readonly(job)) {
		job->set_state(job, SMC_FAILED);

		ret = job->save(job);
		if (!ret.code)
			return;

		/* Try on a read-write object. */
	}

	ret = smc_job_get(job->get_id(job), &jobs);
	SM_ASSERT(sm_list_count(jobs) <= 1, "Parameter sweep not implemented in %s", __FUNCTION__); \

	if (!ret.code) {
		smc_job_t job_rw = jobs->data;

		job_rw->set_state(job_rw, SMC_FAILED);

		/* Reuse staging fault set on the original object. */
		job_rw->set_staging_fault(job_rw, job->get_staging_fault(job));
		job_rw->save(job_rw);
	}

	SM_LIST_FREE_USER(jobs, smc_job_free);
}

static void *
stager_in(void *arg)
{
	struct stager_data *data = (struct stager_data *) arg;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	sm_list_t jobs = NULL;
	smc_job_t job = NULL;
	sm_ret_code_t ret;

	SM_DEBUG_ENTER_STR(data->id);

	GET_JOB_READONLY();
	job = jobs->data;
	SM_ASSERT(job->get_state(job) == SMC_STAGE_IN, "Job in stager but not in Stage-in state");

	/* This function might use job->set_staging_fault but we cannot save
	 * this object as it is read only. We copy the fault later. */
	SM_TRY_NOT_OK(stager_in_helper(job, data->module_manager));
	SM_LIST_FREE_USER(jobs, smc_job_free);

	GET_JOB();
	job = jobs->data;
	SM_ASSERT(job->get_state(job) == SMC_STAGE_IN, "Job in stager but not in Stage-in state");

	if (smc_jsdl_has_application(job->get_jsdl_parsed(job))) {

		if (data->reservation_id) {
				SM_TRY_NOT(params = sm_value_create(3, SM_VALUE_STRING, job->get_jsdl(job), SM_VALUE_STRING, data->reservation_id, SM_VALUE_UINT32, data->reservation_duration));
		} else {
				SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, job->get_jsdl(job)));
		}

		/* On success, it will set SMC_QUEUED state and save. */
		SM_TRY_NOT(result = smc_session_process_submit_job(job, params));

		if (SM_VALUE_RET_CODE_C(result[0])) {
			sm_notice("Failed to submit job %s after staging in: %d:%s",
					job->get_id(job), SM_VALUE_RET_CODE_C(result[0]),
					SM_VALUE_RET_CODE_STRING(result[0]));

			set_staging_fault(job, STAGING_INTERNAL_FAULT, SM_VALUE_RET_CODE_STRING(result[0]));
			goto fault;
		}

	} else {
		sm_debug("Activity has no application defined. Proceeding from stage in to stage out.");

		/* This will set appropriate state (SMC_STAGE_OUT or SMC_FAILED)
		 * and save the job object. */
		smc_staging_start_stage_out(job, data->module_manager);
	}

	sm_value_free(&params);
	sm_value_free(&result);

	SM_LIST_FREE_USER(jobs, smc_job_free);
	FREE_STAGER_DATA();
	smc_session_process_stager_done();
	return NULL;

fault:
	try_mark_as_failed(job);

	sm_value_free(&params);
	sm_value_free(&result);

	SM_LIST_FREE_USER(jobs, smc_job_free);
	FREE_STAGER_DATA();
	smc_session_process_stager_done();
	return NULL;
}

static void *
stager_out(void *arg)
{
	struct stager_data *data = (struct stager_data *) arg;
	sm_list_t jobs = NULL;
	smc_job_t job = NULL;
	sm_ret_code_t ret;

	SM_DEBUG_ENTER_STR(data->id);

	GET_JOB_READONLY();
	job = jobs->data;
	SM_ASSERT(job->get_state(job) == SMC_STAGE_OUT, "Job in stager but not in Stage-out state");

	/* These functions might use job->set_staging_fault but we cannot save
	 * this object as it is read only. We copy the fault later. */
	SM_TRY_NOT_OK(stager_out_helper(job, data->module_manager));
	SM_TRY_NOT_OK(smc_staging_delete_on_termination(job, false));
	SM_LIST_FREE_USER(jobs, smc_job_free);

	GET_JOB();
	job = jobs->data;
	SM_ASSERT(job->get_state(job) == SMC_STAGE_OUT, "Job in stager but not in Stage-out state");
	job->set_state(job, SMC_FINISHED);
	SAVE_JOB();

	SM_LIST_FREE_USER(jobs, smc_job_free);
	FREE_STAGER_DATA();
	smc_session_process_stager_done();
	return NULL;

fault:
	try_mark_as_failed(job);

	SM_LIST_FREE_USER(jobs, smc_job_free);
	FREE_STAGER_DATA();
	smc_session_process_stager_done();
	return NULL;
}

sm_status_ret_code_t
smc_staging_start_stage_in(smc_job_t job, smc_job_res_info_t res_info, sm_module_manager_t module_manager)
{
	sm_thread_t thread;
	struct stager_data *data = NULL;
	sm_ret_code_t ret;

	SM_MANDATORY_P(job);
	SM_MANDATORY_P(module_manager);

	SM_DEBUG_ENTER();

	smc_session_process_stager_new();

	SM_TRY_NOT(data = sm_malloc(sizeof(*data)));
	SM_TRY_NOT(data->id = sm_strdup(job->get_id(job)));

	if (res_info) {
		SM_TRY_NOT(data->reservation_id = sm_strdup(res_info->drms_id));
		SM_TRY_NOT(data->reservation_duration = res_info->duration);
	} else {
		data->reservation_id = NULL;
	}

	data->module_manager = module_manager;

	job->set_state(job, SMC_STAGE_IN);
	SAVE_JOB();

	SM_TRY_NOT_OK(sm_application_thread_create(&thread, NULL, stager_in, data));
	return SM_OK;

fault:
	try_mark_as_failed(job);

	FREE_STAGER_DATA();
	smc_session_process_stager_done();
	return SM_FAULT;
}

sm_status_ret_code_t
smc_staging_start_stage_out(smc_job_t job, sm_module_manager_t module_manager)
{
	sm_thread_t thread;
	struct stager_data *data = NULL;
	sm_ret_code_t ret;

	SM_MANDATORY_P(job);
	SM_MANDATORY_P(module_manager);

	SM_DEBUG_ENTER();

	smc_session_process_stager_new();

	SM_TRY_NOT(data = sm_malloc(sizeof(*data)));
	SM_TRY_NOT(data->id = sm_strdup(job->get_id(job)));
	data->module_manager = module_manager;

	job->set_state(job, SMC_STAGE_OUT);
	SAVE_JOB();

	SM_TRY_NOT_OK(sm_application_thread_create(&thread, NULL, stager_out, data));
	return SM_OK;

fault:
	try_mark_as_failed(job);

	FREE_STAGER_DATA();
	smc_session_process_stager_done();
	return SM_FAULT;
}
