/* $Id: user.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <stdbool.h>
#include <pwd.h>

#include <qcg-core/sm.h>
#include <qcg-core/access.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <common.h>
#include <user.h>
#include <stub/qcgcompStub.h>
#include <stub/qcgcompH.h>

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "user"
	}
};

extern struct Namespace qcgcomp_namespaces[];

sm_ret_code_t
smc_user_get_uid(const char *username, uid_t *uid)
{
	char buf[4096];
	struct passwd pwd, *ptr = NULL;
	int err;

	SM_MANDATORY_P(username);
	SM_MANDATORY_P(uid);

	SM_DEBUG_ENTER_STR(username);

	err = getpwnam_r(username, &pwd, buf, sizeof(buf), &ptr);

	if (err) {
		switch (err) {
			case EINTR:
			case EIO:
			case EMFILE:
			case ENFILE:
			case ENOMEM:
			case ERANGE:
				errno = err;
				sm_perror("getpwnam_r");
				return sm_ret_code_create(SMC_USER_INTERNAL_FAULT, "Failed to get user UID");
		}
	}

	if (!ptr)
		return sm_ret_code_create(SMC_USER_NOT_FOUND, "Requested user does not exist");

	*uid = ptr->pw_uid;

	return sm_ret_code_create_ok();
}

sm_ret_code_t
smc_user_get_home(const char *username, char **home)
{
	char buf[4096];
	struct passwd pwd, *ptr = NULL;
	int err;

	SM_MANDATORY_P(username);
	SM_MANDATORY_P(home);

	SM_DEBUG_ENTER_STR(username);

	*home = NULL;

	err = getpwnam_r(username, &pwd, buf, sizeof(buf), &ptr);

	if (err) {
		switch (err) {
			case EINTR:
			case EIO:
			case EMFILE:
			case ENFILE:
			case ENOMEM:
			case ERANGE:
				errno = err;
				sm_perror("getpwnam_r");
				return sm_ret_code_create(SMC_USER_INTERNAL_FAULT, "Failed to get user UID");
		}
	}

	if (!ptr)
		return sm_ret_code_create(SMC_USER_NOT_FOUND, "Requested user does not exist");

	if (!(*home = sm_strdup(ptr->pw_dir)))
		return sm_ret_code_create_oom();

	return sm_ret_code_create_ok();
}

sm_ret_code_t
smc_user_handle_submission(const char *requested_username, sm_atc_token_t token, char **user, uid_t *uid)
{
	const char *username = NULL;
	sm_ret_code_t ret;
	sm_list_t l;
	SM_BOOL have_access_any = SM_FALSE;

	SM_MANDATORY_P(token);
	SM_MANDATORY_P(user);
	SM_MANDATORY_P(uid);

	SM_DEBUG_ENTER();

	*user = NULL;

	username = NULL;

	/* Reservations policies are somehow different from submitting jobs.
	 * 1. by default only some set of users are allowed to create reservations
	 * 2. reservation created by one user may be used by another user if explicitly specified
	 */

	/* a) get the user from the sec context */
	for (l = token->mapped_users; l; l = l->next) {
		sm_user_t u = l->data;

		if (strcmp(u->username, SM_ACCESS_ANY)) {
			username = u->username;
			break;
		} else {
			have_access_any = SM_TRUE;
		}
	}

	/* b) if there is no security context */
	if (!username) {
		/* and anonymous atz module is in use then get the username from ARDL */
		if (have_access_any && requested_username) {
			username = requested_username;
		} else {
			/* TODO handle reservation administrator here? */
			return sm_ret_code_create(SMC_USER_NO_MAPPING, "You do not have any valid user mapping");
		}
	}

	sm_debug("Using local user: %s", username);

	ret = smc_user_get_uid(username, uid);
	if (ret.code)
		return ret;

	if (!(*user = sm_strdup(username)))
		return sm_ret_code_create_oom();

	return sm_ret_code_create_ok();
}

sm_status_ret_code_t
smc_user_authorize(smc_user_auth_method_t method, const char *owner, const char *subject)
{
#ifdef TEST
	(void) method; (void) owner; (void) subject;
	return SM_OK;
#else
	extern SM_BOOL smc_only_owner_can_terminate;
	extern SM_BOOL smc_only_owner_can_get_status;
	extern SM_BOOL smc_only_owner_can_get_document;

	SM_MANDATORY_P(owner);
	SM_MANDATORY_P(subject);

	SM_DEBUG_ENTER_STR(subject);

	/* If submitted as anonymous subject, then everyone can be considered
	 * the owner. */
	if (!strcmp(owner, SM_ACCESS_ANONYMOUS_SUBJECT))
		return SM_OK;

	switch (method) {
		case SMC_USER_AUTH_TERMINATE:
			if (!smc_only_owner_can_terminate)
				return SM_OK;
			break;
		case SMC_USER_AUTH_GET_DOCUMENT:
			if (!smc_only_owner_can_get_document)
				return SM_OK;
			break;
		case SMC_USER_AUTH_GET_STATUS:
			if (!smc_only_owner_can_get_status)
				return SM_OK;
			break;
	}

	if (strcmp(subject, owner)) {
		sm_notice("Owner (%s) != subject (%s), not authorizing", owner, subject);
		return SM_NEG;
	} else
		return SM_OK;
#endif
}
