/* $Id: ares.h 2 2009-10-05 10:09:52Z mamonski $ */

/*
 *  Copyright 2006-2009 QCG Project Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * @file ares.h
 * Advance Reservation API ver 0.3
 *
 * @author Pawel Lichocki
 * @author Lukasz Ciesnik <lukasz.ciesnik@gmail.com>
 * @author Mariusz Mamonski <mamonski@man.poznan.pl>
 *
 * For jobs submission use DRMAA libary - set drmaa_native_specification
 * with proper reservation id.
 */



#ifndef __ARES_H
#define __ARES_H

#include <stddef.h>

/** @defgroup ares  Advance Reservation API. */
/* @{ */

/**
 * @defgroup ares_bufsize Sufficient buffer sizes (including trailing '\0').
 * @ingroup ares
 */
/* @addtogroup ares_bufsize @{ */
#define ARES_ERROR_STRING_BUFFER    1024
#define ARES_ATTR_BUFFER            1024
#define ARES_RESERVATION_ID_BUFFER    64
#define ARES_IMPLEMENTATION_BUFFER  1024
#define ARES_DRM_SYSTEM_BUFFER      1024
/* @} */


/**
 * @defgroup ares_errno  Error codes.
 * @ingroup ares
 */
/* @addto ares_errno @{ */
#define ARES_ERRNO_SUCCESS                         0
#define ARES_ERRNO_INTERNAL_ERROR                101
#define ARES_ERRNO_DRM_COMMUNICATION_FAILURE     102
#define ARES_ERRNO_AUTH_FAILURE                  103
#define ARES_ERRNO_INVALID_ARGUMENT              104
#define ARES_ERRNO_NOT_INITIALIZED               105
#define ARES_ERRNO_NO_MEMORY                     106
#define ARES_ERRNO_INVALID_CONTACT_STRING        107
#define ARES_ERRNO_INVALID_ATTRIBUTE_FORMAT      108
#define ARES_ERRNO_INVALID_ATTRIBUTE_VALUE       109
#define ARES_ERRNO_CONFLICTING_ATTRIBUTE_VALUES  110
#define ARES_ERRNO_TRY_LATER                     111
#define ARES_ERRNO_NO_SLOTS_AVAILABLE            112
#define ARES_ERRNO_NO_MATCHING_RESOURCES         113
#define ARES_ERRNO_INVALID_RESERVATION_ID        114
#define ARES_ERRNO_NO_MORE_ELEMENTS              115
#define ARES_ERRNO_NOT_IMPLEMENTED               116
#define ARES_ERRNO_ATTRIBUTE_NOT_SET             117
#define ARES_ERRNO_NO_ACTIVE_RESERVATIONS        118
/* @} */


/**
 * @defgroup ares_generic_lists  Generic valeus and names types.
 * Used both by reservation template and reservation state.
 * @ingroup ares
 */
/* @addtogroup ares_generic_lists @{ */

/** List of attribute values (strings). */
typedef struct ares_attr_values_s ares_attr_values_t;
/** Get next attribute value.
 * @exception ARES_ERRNO_NO_MORE_ELEMENTS
 * */
int ares_get_next_attr_value(ares_attr_values_t *values, char *value, size_t value_len);
/** Get number of attribute values. */
int ares_get_num_attr_values(ares_attr_values_t *values, size_t *size);
/** Free memory and dispose of attribute values */
void ares_release_attr_values(ares_attr_values_t *values);

/** List of attribute names (strings). */
typedef struct ares_attr_names_s ares_attr_names_t;
/** Get next attribute name.
 * @exception ARES_ERRNO_NO_MORE_ELEMENTS
 * */
int ares_get_next_attr_name(ares_attr_names_t *names, char *value, size_t value_len);
/** Get number of attribute names. */
int ares_get_num_attr_names(ares_attr_names_t *names, size_t *size);
/** Free memory and dispose of attribute names. */
void ares_release_attr_names(ares_attr_names_t *names);

/* @} */



/**
 * @defgroup ares_rt  Reservation template.
 * Needed for adding new reservation.
 * @ingroup ares
 */
/* @addtogroup ares_rt @{ */

/* Opaque reservation template. */
typedef struct ares_reservation_template_s ares_reservation_template_t;

/** Create reservation template.
 * @exception ARES_ERRNO_NO_MEMORY
 * */
int ares_allocate_reservation_template(
		ares_reservation_template_t **art,
		char *error_diagnosis, size_t error_diag_len
		);

/** Delete reservation template. */
int ares_delete_reservation_template(
		ares_reservation_template_t *art,
		char *error_diagnosis, size_t error_diag_len
		);
/* @} */


/**
 * @defgroup ares_resv_attr  Reservation template attributes.
 * @ingroup ares_rt
 */
/* @addtogroup ares_resv_attr @{ */

/**
 * Human readable reservation name.
 */
#define ARES_RESERVATION_NAME "ares_reservation_name"

/**
 * Begin of time window in which resources should be reserved.
 * Date/time should be speciffied accroding to ISO 8601:
 * <tt> YYYY-MM-DD hh:mm:ss +/- hh:mm </tt>
 *
 * For compability with DRMAA implementations are advised
 * to accept DRMAA 1.0 date/time format:
 * <tt> YYYY/MM/DD hh:mm:ss +/- hh:mm </tt>
 *
 * @see ARES_RESERVATION_END_TIME
 * @see ARES_RESERVATION_STRICT_WINDOW
 */
#define ARES_RESERVATION_START_TIME "ares_reservation_start_time"

/**
 * End of time window in which resources should be reserved.
 * @see ARES_RESERVATION_START_TIME
 * @see ARES_RESERVATION_STRICT_WINDOW
 */
#define ARES_RESERVATION_END_TIME "ares_reservation_end_time"

/**
 * Whether jobs which use reservation can run
 * outside of reservation time window (1)
 * or they are terminated after end time is reached (0).
 *
 * Boolean attribute (1/0).  Defaults to false.
 *
 * @see ARES_RESERVATION_END_TIME
 */
#define ARES_RESERVATION_OPEN "ares_reservation_open"

/**
 * Whether reversed time window must exactly match
 * requested reservation start and end time (1)
 * or it is only enough to be not earlier than start time
 * and not later than end time (0).
 * If the ares_reservation_strict_window is set to "0"
 * the first possible reservation matching all the time
 * and resources constrains is created. Subsequent calls
 * with the same reservation template will result in
 * the set of reservations filling all free time slots
 * within requested time window.
 *
 * This attribute is mutual exclusive with ARES_RESERVATION_SHARE_NODES.
 * Boolean attribute (1/0). Defaults to true.
 *
 * @see ARES_RESERVATION_SHARE_NODES
 * @see ARES_RESERVATION_START_TIME
 * @see ARES_RESERVATION_END_TIME
 */
#define ARES_RESERVATION_STRICT_WINDOW "ares_reservation_strict_window"

/**
 * Whether jobs not submitted within reservation
 * can be run on reserved nodes (when node has spare cpus).
 *
 * Boolean attribute (1/0).  Default is to share nodes.
 */
#define ARES_RESERVATION_SHARE_NODES "ares_reservation_share_nodes"

/**
 * Specification of user for which reservation
 * is made (eligible to submit jobs within reservation).
 * @see ARES_RESERVATION_GROUP
 */
#define ARES_RESERVATION_USER "ares_reservation_user"

/**
 * Specification of DRMS group for which reservation is made.
 * @see ARES_RESERVATION_USER
 */
#define ARES_RESERVATION_GROUP "ares_reservation_group"

/**
 * Requested architecture of reserved nodes
 * according to JSDL specification.
 */
#define ARES_ARCHITECTURE "ares_architecture"

/**
 * Requested operating system platform of reserved nodes
 * according to JSDL specification.
 */
#define ARES_OPERATING_SYSTEM "ares_operating_system"

/** Number of slots reserved. */
#define ARES_N_SLOTS "ares_n_slots"

/** Number of additional slots reserved. */
#define ARES_ADDITIONAL_N_SLOTS "ares_additional_n_slots"

/**
 * Amount of reserved physical memory (RSS) which may be
 * used simultaneously by all jobs from reservation.
 */
#define ARES_MEMORY "ares_memory"

/**
 * Amount of reserved swap space.
 */
#define ARES_SWAP "ares_swap"

/**
 * Amount of reserved file system space within home directory
 * used simultaneously by all processes.
 */
#define ARES_HOME_SPACE "ares_home_space"

/**
 * Amount of reserved file system space within temporary directory/partition
 * used simultaneously by all processes.  This file system
 * may not survive after job completes.
 */
#define ARES_TMP_SPACE "ares_tmp_space"

/** Names of reserved nodes */
#define ARES_V_NODES "ares_v_nodes"

/**
 * Names and values of DRM specific resources given in format
 * res=value
 */
#define ARES_V_NATIVE_RESOURCES "ares_v_native_resources"

/* @} */

/** Get scalar attributes names. */
int ares_get_reservation_attribute_names(
		ares_attr_names_t **names,
		char *error_diagnosis, size_t error_diag_len
		);

/** Get vector attributes names. */
int ares_get_vector_reservation_attribute_names(
		ares_attr_names_t **names,
		char *error_diagnosis, size_t error_diag_len
		);

/** Set scalar reservation attributes. */
int ares_set_reservation_attribute(
		ares_reservation_template_t *art,
		const char *name, const char *value,
		char *error_diagnosis, size_t error_diag_len
		);
/** Get scalar reservation attributes. */
int ares_get_reservation_attribute(
		ares_reservation_template_t *art,
		const char *name, char *value, size_t value_len,
		char *error_diagnosis, size_t error_diag_len
		);

/** Set vector reservation attributes. */
int ares_set_vector_reservation_attribute(
		ares_reservation_template_t *art,
		const char *name, const char *value[],
		char *error_diagnosis, size_t error_diag_len
		);
/** Get vector reservation attributes. */
int ares_get_vector_reservation_attribute(
		ares_reservation_template_t *art,
		const char *name, ares_attr_values_t **values,
		char *error_diagnosis, size_t error_diag_len
		);


/**
 * Reservation state as returned by ares_reservation_state().
 */
typedef struct ares_reservation_state_s ares_reservation_state_t;

/**
 * Delete and dispose reservation state structure.
 */
int ares_delete_reservation_state(
		ares_reservation_state_t *arit,
		char *error_diagnosis, size_t error_diag_len
		);

/*
 * ARES_STATE_RESERVATION_START_TIME
 * ARES_STATE_RESERVATION_END_TIME
 * ARES_STATE_ERROR
 * ARES_STATE_RESERVATION_USER
 * ARES_STATE_RESERVATION_GROUP
 * ARES_STATE_STATE_V_JOBS
 * ARES_STATE_V_NODES_INFO
 */

/*
 * Information template may have following attributes:
 */

/**
 * If set contains diagnostic message explaining why the reservation turned into error state
 */

#define ARES_STATE_ERROR_DIAG "ares_state_error"

/**
 * Begin of time window in which resources should be reserved.
 * Date/time should be speciffied accroding to ISO 8601:
 * <tt> YYYY-MM-DD hh:mm:ss +/- hh:mm </tt>
 *
 * For compability with DRMAA implementations are advised
 * to accept DRMAA 1.0 date/time format:
 * <tt> YYYY/MM/DD hh:mm:ss +/- hh:mm </tt>
 *
 * @see ARES_STATE_RESERVATION_END_TIME
 */
#define ARES_STATE_RESERVATION_START_TIME "ares_state_reservation_start_time"

/**
 * End of time window in which resources should be reserved.
 * @see ARES_STATE_RESERVATION_START_TIME
 */
#define ARES_STATE_RESERVATION_END_TIME "ares_state_reservation_end_time"

/**
 * Specification of DRMS user/group for which reservation
 * was made (eligible to submit jobs within reservation).
 * @see ARES_STATE_RESERVATION_GROUP
 */
#define ARES_STATE_RESERVATION_USER "ares_state_reservation_user"

/**
 * Specification of DRMS group for which reservation was made.
 * @see ARES_STATE_RESERVATION_USER
 */
#define ARES_STATE_RESERVATION_GROUP "ares_state_reservation_group"


/**
 * List of job identifiers currently assigned to given reservation.
 */
#define ARES_STATE_V_JOBS "ares_state_jobs"

/**
 * List of colon separated:
 *  - node name
 *  - number of slots on node,
 *  - number of slots assigned to given reservation,
 *  - number of slots used by jobs submitted within reservation.
 */
#define ARES_STATE_V_NODES_INFO "ares_state_v_nodes_info"

/* get non-vector info element (set not possible, information is read-only) */
int ares_get_state_attribute(
		ares_reservation_state_t *ait,
		const char *name, char *value, size_t value_len,
		char *error_diagnosis, size_t error_diag_len
		);

/* get vector info element (set not possible, information is read-only) */
int ares_get_vector_state_attribute(
		ares_reservation_state_t *ait,
		const char *name, ares_attr_values_t **values,
		char *error_diagnosis, size_t error_diag_len
		);


/**
 * @defgroup ares_main  Main Advance Reservation API.
 */
/* @{ */

/**
 * Initialize the library.
 */
int ares_init(
		const char *contact,
		char *error_diagnosis, size_t error_diag_len
		);

/**
 * Clean up the library.
 */
int ares_exit(char *error_diagnosis, size_t error_diag_len);

/**
 * Add reservation.
 */
int ares_add_reservation(char *ares_id, size_t ares_id_len, const ares_reservation_template_t *art, char *error_diagnosis, size_t error_diag_len);

/**
 * Retrieve reservation state.
 */
int ares_reservation_state(const char *ares_id, ares_reservation_state_t **ari, char *error_diagnosis, size_t error_diag_len);

/**
 * Get list of all submitted reservation identifiers
 * within the system.
 */
int ares_get_reservation_ids(ares_attr_values_t **rsv_ids, char *error_diagnosis, size_t error_diag_len);

/**
 * Delete reservation.  Might fail when reservation outdates.
 */
int ares_remove_reservation(const char *ares_id,char *erro_diagnosis, size_t error_diag_len);

/* @} */


/**
 * Get information about this ARes API implementation.
 */
int ares_get_ARES_implementation(char *ares_impl, size_t ares_impl_len, char *error_diagnosis, size_t error_diag_len);

/**
 * Get ARes API version implemented by this library.
 */
int ares_version(unsigned int *major, unsigned int *minor,char *error_diagnosis, size_t error_diag_len);

/**
 * Get information about underlying DRM system.
 */
int ares_get_DRM_system(char *drm_system, size_t drm_system_len, char *error_diagnosis, size_t error_diag_len);

/**
 * Get error string for given error code.
 */
const char *ares_strerror(int ares_errno);

/* @} */

#endif /* __ARES_H */
