/* $Id: reservation_ares.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <dlfcn.h>
#include <time.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <ares.h>
#include <include/reservation_module.h>
#include <xmlstructs/xmlstructs.h>

__SM_ID("$Id: reservation_ares.c 2 2009-10-05 10:09:52Z mamonski $");

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *reservation_ares_create(sm_module_t, sm_value_t *);
static sm_value_t *reservation_ares_terminate(sm_module_t, sm_value_t *);
static sm_value_t *reservation_ares_get_status(sm_module_t, sm_value_t *);
static sm_value_t *reservation_ares_get_active(sm_module_t);

static sm_status_ret_code_t reservation_ares_get_time_window(const char *arid,time_t *start_time, time_t *end_time);

RESERVATION_MODULE = {

	.header = {
		.entry_size = RESERVATION_MODULE_ENTRY_SIZE,
		.api_type = RESERVATION_MODULE_API,

		.name = "reservation_ares",
		.descr = "ARES reservation",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},

	.create = reservation_ares_create,
	.terminate = reservation_ares_terminate,
	.get_status = reservation_ares_get_status,
	.get_active = reservation_ares_get_active
};

#define SET_ERRBUF() \
	do { \
		char __buf[64]; \
		if (!strlen(errbuf)) \
			strlcpy(errbuf, ares_strerror(ret), sizeof(errbuf)); \
		snprintf(__buf, sizeof(__buf), " (ARES errno = %d)", ret); \
		strlcat(errbuf, __buf, sizeof(errbuf)); \
	} while(0)

static int
setup(sm_module_t this)
{
	const char *path = NULL;
	void *handle = NULL;
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	char attr[ARES_ATTR_BUFFER] = "";
	char impl[ARES_IMPLEMENTATION_BUFFER] = "";
	char drm[ARES_DRM_SYSTEM_BUFFER] = "";
	ares_attr_names_t *attributes = NULL;
	int ret = ARES_ERRNO_SUCCESS;
	unsigned int major, minor;
	int initialized = 0;

	SM_DEBUG_ENTER();

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/@path", &path) != SM_OK)
		goto fault;

	handle = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
	if (!handle) {
		const char *msg = dlerror();
		if (!msg)
			msg = path;
		sm_error("Could not load ARES library: %s", msg);
		goto fault;
	}

	sm_debug("-> ares_init()");

	if ((ret = ares_init(NULL, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS)
		goto fault;

	initialized = 1;

	if ((ret = ares_get_reservation_attribute_names(&attributes, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS)
		goto fault;

	while ((ret = ares_get_next_attr_name(attributes, attr, sizeof(attr))) == ARES_ERRNO_SUCCESS) {
		sm_debug("ARES supported attribute: %s", attr);
	}

	ares_release_attr_names(attributes);
	attributes = NULL;

	if ((ret = ares_version(&major, &minor, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS)
		goto fault;

#if 0
	/* TODO after the API gets stable */
	if (major != 1 || minor != 0) {
		sm_error("Unsupported ARES version: %d.%d", major, minor);
		goto fault;
	}
#endif

	if ((ret = ares_get_ARES_implementation(impl, sizeof(impl), errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS)
		goto fault;

	if ((ret = ares_get_DRM_system(drm, sizeof(drm), errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS)
		goto fault;

	sm_debug("DRM system: %s, ARES implementation: %s", drm, impl);

	this->ctxt = handle;
	return SM_OK;

fault:
	if (ret != ARES_ERRNO_SUCCESS) {
		SET_ERRBUF();
		sm_error("Failed to set up ARES: %s", errbuf);
	}
	if (attributes)
		ares_release_attr_names(attributes);
	if (initialized)
		ares_exit(errbuf, sizeof(errbuf));
	if (handle)
		dlclose(handle);
	return SM_FAULT;
}

static void
cleanup(sm_module_t this)
{
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";

	SM_DEBUG_ENTER();

	if (this->ctxt) {
		ares_exit(errbuf, sizeof(errbuf));
		dlclose(this->ctxt);
		this->ctxt = NULL;
	}
}

static sm_status_ret_code_t
reservation_ares_datetime_parse(const char *str, time_t *val)
{
	struct tm tm;
	char *ptr = NULL;

	SM_MANDATORY_P(str);
	SM_MANDATORY_P(val);

	SM_DEBUG_ENTER_STR(str);

	memset(&tm, 0, sizeof(tm));

	/* 2010-01-03 17:00 */
	/* 2010-01-03 17:00Z */
	/* 2010-01-03 17:00+01:00 */
	/* 2010-01-03 17:00-01:00 */

	/* Consider using soap_s2dateTime here? */

	if (((ptr = strptime(str, "%Y-%m-%d %H:%M:%S", &tm)) == NULL) && ((ptr = strptime(str, "%Y-%m-%d %H:%M", &tm)) == NULL))
		return SM_FAULT;

	tm.tm_isdst = -1; /* and negative if the information is not available. */

	if (*ptr == 'Z') {
		*val = timegm(&tm);
	} else if (*ptr == '+' || *ptr == '-') {
		*val = mktime(&tm);
		ptr++;
		if ((ptr = strptime(ptr, "%H:%M", &tm)) == NULL)
			return SM_FAULT;

		if (*ptr != '\0')
			return SM_FAULT;

	} else if (*ptr == '\0') {
		*val = mktime(&tm);
	} else {
		return SM_FAULT;
	}

	return SM_OK;
}

static sm_status_ret_code_t
reservation_ares_get_time_window(const char *arid, time_t *start_time, time_t *end_time)
{
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	char start_time_s[64] = "";
	char end_time_s[64] = "";
	ares_reservation_state_t *ari = NULL;

	SM_MANDATORY_PARAM(arid, SM_FAULT);

	SM_DEBUG_ENTER_STR(arid);

	if (ares_reservation_state(arid, &ari, errbuf, sizeof(errbuf)) != ARES_ERRNO_SUCCESS) {
		sm_error("ares_reservation_state(%s) failed: %s", arid, errbuf);
		goto fault;
	}

	if (ares_get_state_attribute(ari, ARES_STATE_RESERVATION_START_TIME, start_time_s, sizeof(start_time_s), errbuf, sizeof(errbuf)) != ARES_ERRNO_SUCCESS) {
		sm_error("ares_get_state_attribute(%s) failed: %s", arid, errbuf);
		goto fault;
	}

	if (ares_get_state_attribute(ari, ARES_STATE_RESERVATION_END_TIME, end_time_s, sizeof(end_time_s), errbuf, sizeof(errbuf)) != ARES_ERRNO_SUCCESS) {
		sm_error("ares_get_state_attribute(%s) failed: %s", arid, errbuf);
		goto fault;
	}

	ares_delete_reservation_state(ari, errbuf, sizeof(errbuf));
	ari = NULL;

	sm_debug("Reservation start time = %s", start_time_s);
	sm_debug("Reservation end time = %s", end_time_s);

	if (reservation_ares_datetime_parse(start_time_s, start_time) != SM_OK) {
		/* The time comes from DRMS so this is an internal fault. */
		sm_error("Parsing reservation start time failed: %s", start_time_s);
		goto fault;
	}

	if (reservation_ares_datetime_parse(end_time_s, end_time) != SM_OK) {
		sm_error("Parsing reservation end time failed: %s", end_time_s);
		goto fault;
	}

	return SM_OK;

fault:
	ares_remove_reservation(arid, errbuf, sizeof(errbuf));

	if (ari)
		ares_delete_reservation_state(ari, errbuf, sizeof(errbuf));

	return SM_FAULT;
}

static ares_reservation_template_t *
ardl_to_ares_reservation_template(const char *ardlstr, sm_value_t **result)
{
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	ares_reservation_template_t *at = NULL;
	sm_alloc_tracked_t tracked = NULL;
	_ardl__ReservationDefinition ardl;
	_ardl__ReservationDescription *description;
	_ardl__Resources *resources;
	struct ardl__ARDLTimeWindow_USCOREType *timewindow;
	int ret = ARES_ERRNO_SUCCESS;
	time_t start = 0;
	time_t end = 0;
	struct tm *starttm;
	struct tm *endtm;
	char buf[256];
	SM_BOOL recheck_rsv_time = SM_FALSE;

	SM_MANDATORY_PARAM(ardlstr, NULL);
	SM_MANDATORY_PARAM(result, NULL);

	SM_DEBUG_ENTER_STR(ardlstr);

	if (!(tracked = sm_gsoap_buffer_get(ardlstr, &ardl, ardl__ARDLReservationDefinition_USCOREType, "ardl:ReservationDefinition", xmlstructs_namespaces))) {
		sm_error("Failed to parse ARDL");
		goto fault;
	}

	description = ardl.ardl__ReservationDescription;
	timewindow = description->TimeWindow;

	if ((ret = ares_allocate_reservation_template(&at, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS)
		goto fault;

	if (description->ReservationIdentification && description->ReservationIdentification->ReservationName)
		if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_NAME, description->ReservationIdentification->ReservationName, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;

	start = *timewindow->StartTime;
	end = timewindow->union_ARDLTimeWindow_USCOREType.EndTime;
	assert(timewindow->__union_ARDLTimeWindow_USCOREType == SOAP_UNION__ardl__union_ARDLTimeWindow_USCOREType_EndTime);

	/*
	 * - We're not multi-threaded so don't bother with gmtime_r and the like.
	 * - The time is in UTC, because gSOAP does all the conversions.
	 */
	starttm = gmtime(&start);
	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%SZ", starttm);
	if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_START_TIME, buf, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;

	endtm = gmtime(&end);
	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%SZ", endtm);
	if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_END_TIME, buf, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;

	if (timewindow->open)
		if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_OPEN, "1", errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;

	if (!timewindow->strict) {
		recheck_rsv_time = SM_TRUE;
		if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_STRICT_WINDOW, "0", errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
	}

	if ((resources = description->ardl__Resources)) {
		struct jsdl__CPUArchitecture_USCOREType *cpu;

		if (resources->ReservedSlotsCount) {
			snprintf(buf, sizeof(buf), "%llu", *resources->ReservedSlotsCount);
			if ((ret = ares_set_reservation_attribute(at, ARES_N_SLOTS, buf, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
		}

		if (resources->AdditionalReservedSlotsCount) {
			snprintf(buf, sizeof(buf), "%llu", *resources->AdditionalReservedSlotsCount);
			if ((ret = ares_set_reservation_attribute(at, ARES_ADDITIONAL_N_SLOTS, buf, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
		}

		if (resources->ExclusiveExecution) {
			snprintf(buf, sizeof(buf), "0");
			if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_SHARE_NODES, buf, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
		}

		if (resources->UserName)
			if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_USER, resources->UserName, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;

		if (resources->GroupName)
			if ((ret = ares_set_reservation_attribute(at, ARES_RESERVATION_GROUP, resources->GroupName, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;

		if ((cpu = resources->CPUArchitecture)) {
			const char *name = soap__jsdl__CPUArchitectureName2s(NULL, cpu->jsdl__CPUArchitectureName);
			if ((ret = ares_set_reservation_attribute(at, ARES_ARCHITECTURE, name, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
		}

		if (resources->CandidateHosts && resources->CandidateHosts->__sizeHostName) {
			int count = resources->CandidateHosts->__sizeHostName;
			char **hosts = sm_alloc_tracked_calloc(tracked, count + 1, sizeof(char *));
			int i;

			if (!hosts)
				goto fault;

			for (i = 0; i < count; i++)
				hosts[i] = resources->CandidateHosts->HostName[i];

			hosts[count] = NULL;

			if ((ret = ares_set_vector_reservation_attribute(at, ARES_V_NODES, (const char**) hosts, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
		}

		if (resources->NativeResource && resources->__sizeNativeResource) {
			int count = resources->__sizeNativeResource;
			char **native = sm_alloc_tracked_calloc(tracked, count + 1, sizeof(char *));
			int i;

			if (!native)
				goto fault;

			for (i = 0; i < count; i++) {
				 if (!resources->NativeResource[i].__item || strcmp("", resources->NativeResource[i].__item) == 0)
					 native[i] = resources->NativeResource[i].name; /* boolean resource (e.g.: "openmpi") */
				 else {
					 /* other resources (e.g.: "mem=1000") */
					 SM_TRY_NOT(native[i] = sm_alloc_tracked_asprintf(tracked, "%s=%s",resources->NativeResource[i].name, resources->NativeResource[i].__item));
				 }
			}

			native[count] = NULL;

			if ((ret = ares_set_vector_reservation_attribute(at, ARES_V_NATIVE_RESOURCES, (const char**) native, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) goto fault;
		}

		/* TODO: FileSystem, OperatingSystem, IndividualPhysicalMemory,
		 * IndividualSwapMemory */
	}

	sm_alloc_tracked_free(&tracked);
	SM_TRY_NOT(*result = sm_value_create(1, SM_VALUE_BOOL, recheck_rsv_time));
	return at;

fault:
	if (ret != ARES_ERRNO_SUCCESS) {
		SET_ERRBUF();

		switch (ret) {
			case ARES_ERRNO_INVALID_ATTRIBUTE_FORMAT:
			case ARES_ERRNO_INVALID_ARGUMENT:
			case ARES_ERRNO_INVALID_ATTRIBUTE_VALUE:
			case ARES_ERRNO_CONFLICTING_ATTRIBUTE_VALUES:
				*result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, errbuf);
				break;
			case ARES_ERRNO_NO_SLOTS_AVAILABLE:
				*result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_NO_SLOTS_AVAILABLE_FAULT, errbuf);
				break;
			case ARES_ERRNO_NO_MATCHING_RESOURCES:
				*result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_NO_MATCHING_RESOURCES_FAULT, errbuf);
				break;
			default:
				*result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, errbuf);
				break;
		}
	}

	if (at)
		ares_delete_reservation_template(at, errbuf, sizeof(errbuf));
	sm_alloc_tracked_free(&tracked);
	return NULL;
}

static sm_value_t *
reservation_ares_create(sm_module_t this, sm_value_t *ardl)
{
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	char arid[ARES_RESERVATION_ID_BUFFER] = "";
	sm_value_t *result = NULL;
	ares_reservation_template_t *at = NULL;
	SM_BOOL must_check_rsv_time = SM_TRUE;
	int ret;

	SM_MANDATORY_PARAM(this, NULL);
	SM_MANDATORY_PARAM(ardl, NULL);

	assert(sm_value_count(ardl) == 1);
	assert(SM_VALUE_IS_STRING(ardl[0]));

	SM_DEBUG_ENTER();

	if (!(at = ardl_to_ares_reservation_template(ardl[0]->vstring, &result)))
		goto fault;

	assert(SM_VALUE_IS_BOOL(result[0]));
	must_check_rsv_time = result[0]->vbool;
	sm_value_free(&result);

	sm_debug("-> ares_add_reservation()");

	ret = ares_add_reservation(arid, sizeof(arid), at, errbuf, sizeof(errbuf));

	if (ret != ARES_ERRNO_SUCCESS)
		SET_ERRBUF();

	switch (ret) {
		case ARES_ERRNO_SUCCESS:
			sm_debug("Created reservation: %s", arid);

			if (!must_check_rsv_time) {
				result = sm_value_create(2, SM_VALUE_RET_CODE, RESERVATION_OK, SM_VALUE_STRING, arid);
			} else {
				time_t start_time = 0;
				time_t end_time = 0;

				sm_debug("Checking reservation start/end time");

				if (reservation_ares_get_time_window(arid, &start_time, &end_time) == SM_OK)
					result = sm_value_create(4, SM_VALUE_RET_CODE, RESERVATION_OK, SM_VALUE_STRING, arid, SM_VALUE_UINT32, (unsigned int)start_time, SM_VALUE_UINT32, (unsigned int)end_time);
				else
					result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Unable to get reservation time window");
			}
			break;
		case ARES_ERRNO_TRY_LATER:
			sm_notice("DRM system didn't accept new reservation, probably saturated");
			result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Try again later");
			break;
		case ARES_ERRNO_INVALID_ATTRIBUTE_VALUE:
		case ARES_ERRNO_INVALID_ATTRIBUTE_FORMAT:
		case ARES_ERRNO_INVALID_ARGUMENT:
		case ARES_ERRNO_CONFLICTING_ATTRIBUTE_VALUES:
			result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, errbuf);
			break;
		case ARES_ERRNO_AUTH_FAILURE:
			result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_NOT_AUTHORIZED_FAULT, errbuf);
			break;
		case ARES_ERRNO_NO_SLOTS_AVAILABLE:
			result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_NO_SLOTS_AVAILABLE_FAULT, errbuf);
			break;
		case ARES_ERRNO_NO_MATCHING_RESOURCES:
			result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_NO_MATCHING_RESOURCES_FAULT, errbuf);
			break;
		default:
			result = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, errbuf);
			break;
	}

	ares_delete_reservation_template(at, errbuf, sizeof(errbuf));
	return result;

fault:
	if (at)
		ares_delete_reservation_template(at, errbuf, sizeof(errbuf));
	if (result && result[0]->vint32 != RESERVATION_OK)
		return result;
	sm_value_free(&result);
	return sm_value_create(1, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, "Internal fault in reservation_ares module");
}

static sm_value_t *
reservation_ares_terminate(sm_module_t this, sm_value_t *identifier)
{
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	sm_value_t *result = NULL;
	int ret;

	SM_MANDATORY_PARAM(this, NULL);
	SM_MANDATORY_PARAM(identifier, NULL);

	assert(sm_value_count(identifier) == 1);
	assert(SM_VALUE_IS_STRING(identifier[0]));

	SM_DEBUG_ENTER_STR(identifier[0]->vstring);

	sm_debug("-> ares_remove_reservation(%s)", identifier[0]->vstring);

	if ((ret = ares_remove_reservation(identifier[0]->vstring, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) {

		SET_ERRBUF();

		switch (ret) {
			case ARES_ERRNO_AUTH_FAILURE:
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_NOT_AUTHORIZED_FAULT, errbuf));
				break;
			case ARES_ERRNO_INVALID_ARGUMENT:
			case ARES_ERRNO_INVALID_RESERVATION_ID:
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, identifier[0]->vstring));
				break;
			default:
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, errbuf));
				break;
		}

	} else
		SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_OK));

	return result;

fault:
	sm_value_free(&result);
	return sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in reservation_ares module");
}

static sm_value_t *
reservation_ares_get_status(sm_module_t this, sm_value_t *identifier)
{
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	sm_value_t *result = NULL;
	ares_reservation_state_t *ari = NULL;
	sm_value_t *vnodes_info = NULL;
	int ret;

	SM_MANDATORY_PARAM(this, NULL);
	SM_MANDATORY_PARAM(identifier, NULL);

	assert(sm_value_count(identifier) == 1);
	assert(SM_VALUE_IS_STRING(identifier[0]));

	SM_DEBUG_ENTER_STR(identifier[0]->vstring);

	sm_debug("-> ares_reservation_state(%s)", identifier[0]->vstring);

	if ((ret = ares_reservation_state(identifier[0]->vstring, &ari, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) {
		SET_ERRBUF();

		switch (ret) {
			case ARES_ERRNO_AUTH_FAILURE:
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_NOT_AUTHORIZED_FAULT, errbuf));
				break;
			case ARES_ERRNO_INVALID_ARGUMENT:
			case ARES_ERRNO_INVALID_RESERVATION_ID:
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, identifier[0]->vstring));
				break;
			default:
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, errbuf));
				break;
		}
	} else {
		char tmp[512] = "";
		ares_attr_values_t *ares_nodes_info = NULL;

		if (ares_get_vector_state_attribute(ari, ARES_STATE_V_NODES_INFO, &ares_nodes_info, errbuf, sizeof(errbuf)) != ARES_ERRNO_SUCCESS) {
			sm_error("Failed to get ARES_STATE_V_NODES_INFO attribute");
			goto fault;
		} else {
			while (ares_get_next_attr_value(ares_nodes_info, tmp, sizeof(tmp)) == ARES_ERRNO_SUCCESS) {
				sm_debug("ARES_STATE_V_NODES_INFO: %s", tmp);
				SM_TRY_NOT(sm_value_append(&vnodes_info, 1, SM_VALUE_STRING, tmp));
			}
		}

		/* Prepare the final result. */
		SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_OK));
		SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_ARRAY, vnodes_info));
		vnodes_info = NULL;
	}

	if (ari)
		ares_delete_reservation_state(ari, errbuf, sizeof(errbuf));

	return result;

fault:
	if (ari)
		ares_delete_reservation_state(ari, errbuf, sizeof(errbuf));

	sm_value_free(&vnodes_info);
	sm_value_free(&result);
	return sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in reservation_ares module");
}

static sm_value_t *
reservation_ares_get_active(sm_module_t this)
{
	int ret;
	char errbuf[ARES_ERROR_STRING_BUFFER] = "";
	char rsv_id[ARES_RESERVATION_ID_BUFFER] = "";
	sm_value_t *result = NULL;
	ares_attr_values_t *rsv_ids = NULL;

	SM_MANDATORY_PARAM(this, NULL);

	SM_DEBUG_ENTER();

	sm_debug("-> ares_get_reservation_ids");

	if (((ret = ares_get_reservation_ids(&rsv_ids, errbuf, sizeof(errbuf))) != ARES_ERRNO_SUCCESS) && (ret != ARES_ERRNO_NO_ACTIVE_RESERVATIONS)) {
		SET_ERRBUF();
		SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, errbuf));
	} else {
		SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_RET_CODE, RESERVATION_OK));

		if (ret != ARES_ERRNO_NO_ACTIVE_RESERVATIONS) {
			while (ares_get_next_attr_value(rsv_ids, rsv_id, sizeof(rsv_id)) == ARES_ERRNO_SUCCESS) {
				sm_debug("rsv_id: %s", rsv_id);
				SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_STRING, rsv_id));
			}
		}
	}

	if (rsv_ids)
		ares_release_attr_values(rsv_ids);

	return result;

fault:
	if (rsv_ids)
		ares_release_attr_values(rsv_ids);

	sm_value_free(&result);
	return sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in reservation_ares module");
}
